// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: CodeAxe.cc 803 2007-09-09 23:25:28Z ctsa $

/// \file

#include "cat_info.h"
#include "rate_gtor_nscodon.h"
#include "rate_gtor_nuc.h"
#include "process_fsa.h"
#include "site_data_report.h"
#include "site_data_state_remap.h"
#include "site_data_util.h"
#include "sim.h"
#include "subs_ml.h"
#include "subs_ml_confidence.h"
#include "substk_exception.h"
#include "subs_ml_print_util.h"
#include "subs_ml_model_util.h"
#include "tang_ei.h"
#include "util/bio/bioseq_util.h"
#include "util/general/die.h"
#include "util/general/io_util.h"
#include "util/general/metatags.h"
#include "util/math/random_util.h"

#include <cstdlib>

#include <fstream>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>

using namespace std;


char* progname;

const unsigned DEFAULT_MIN_MATCH(5);
const char stdstream_tag[] = "-";

const char prog_version[] = "1.0";



namespace MODE {
  enum index_t {
    NONE,
    PROCESS,
    TRAIN,
    LHOOD,
    CAT_POSTP,
    SIM,
    DENOVO,
    NORM,
    REPORT_MODEL,
    REPORT_SEQ,
    REPORT_DATA,
    CONVERT_DATA,
    CONFIDENCE,
    EI
  };
}



void
usage_error(const string& xmessage = "")  NORETURN_TAG;

void
usage_error(const string& xmessage) {

  static const char emphasis_tag[] = "-----------";

  std::ostream& os(std::cerr);

  os << "\n"
     << progname << " :  phylogenetic analysis and simulation tools\n"
     << "\n  version : " << prog_version << "\n"
     << "\n\n"
     << emphasis_tag << " MODES: " << emphasis_tag << "\n"
     << "\n"
     << "-process-seq - process sequence files into CodeAxe site data format\n"
     << "-report-seq  - report summary from sequence files\n"
     << "  -site-model             - (see below)\n"
     << "  -in-seq {dir,file}      - file or dir of files for alignment in fasta format (.fsa)\n"
     << "  [-out {file,-}]         - write site data to file (default stdout)\n"
     << "  [-max-align-count N]    - read only the first N alignments in directory\n"
     << "  [-4fold]                - if site-model is \"nuc\", filter for 4-fold synonymous positions\n"
     << "  [-gc-max x]             - filter any sequence with %gc > x\n"
     << "  [-gc-min x]             - filter any sequence with %gc <= x\n"
     << "  [-no-window-filter]     - turn off window quality filter\n"
     << "  [-no-ambig-gap-filter]  - allow ambiguous/gap positions to be treated as a single ambiguous state type\n"
     << "  [-no-filter]            - no sequence window, ambiguous/gap, or coding sequence filtration\n"
     << "  [-min-window-match N]   - minimum of N/20 matches in quality window (default << " << DEFAULT_MIN_MATCH << " <<)\n"
     << "  [-no-codon-border]      - use all possible codons (default syncs site count for codon/c4/c5 sites)\n"
     << "  [-in-cat {dir,file}]    - category assignments for positions from -in-seq (s/.fsa/.cat/) \n"
     << "  [-cat-labels file]      - category labels\n"
     << "  [-no-adjacent-nuc-diff] - filter sites to have no adjacent differences\n"
     << "  [-only-single-nuc-diff] - filter sites to have only single nuc differences\n"
     << "\n"
     << "-create-model - generate a new model\n"
     << "  [-in-data {file,-}]   - read site data from file\n"
     << "  [-out {file,-}]       - write model to file (default stdout)\n"
     << "  [-random-param]       - randomize starting parameters\n"
     << "  [-seed N]             - set random number seed\n"
     << "  (+ any model flags)\n"
     << "\n"
     << "-ml - search for maximum likelihood model parameters\n"
     << "  -in-model {file,-}    - read model from file\n"
     << "  -in-data {file,-}     - read site data from file\n"
     << "  [-out {file,-}]       - write model to file (default: stdout)\n"
     << "  (+ any model scoring flags)\n"
     << "\n"
     << "-lhood - report log likelihood of the model, log(P(data|model))\n"
     << "  -in-model {file,-}    - read model from file\n"
     << "  -in-data {file,-}     - read site data from file\n"
     << "  (+ any model scoring flags)\n"
     << "\n"
     << "-confidence - report (1.-alpha) confidence intervals on model parameters (alpha fixed @ " << DEFAULT_ALPHA <<  ")\n"
     << "  -in-model {file,-}  - read model from file (assumed to be MLE's wrt data)\n"
     << "  -in-data {file,-}   - read site data from file\n"
     << "  [-ci-search]        - solve ci's by an inversion of the likelihood ratio statistic (default)\n"
     << "  [-ci-fisher]        - solve ci's by the inverse fisher information matrix\n"
//     << "  [-ci-alpha x]       - 100*(1-x)% confidence range (default: x=" << DEFAULT_ALPHA << ")\n"
     << "  [-ci-mutation-only] - evaluate confidence intervals for mutation parameters only (search mode only)\n"
     << "  [-ci-selection-only]- evaluate confidence intervals for selection parameters only (search mode only)\n"
     << "  [-ci-single]        - evaluate the confidence interval for the next unsolved parameter only (search mode only)\n"
     << "  -out file           - write confidence intervals to file. If multiple processes are given the same filename,\n"
     << "                        they will partition the job to find the confidence intervals in parallel\n"
     << "\n"
     << "-norm - set the root and background distribution to the model's stationary distro\n"
     << "  -in-model {file,-}  - read model from file\n"
     << "  [-out {file,-}]     - write equilibrated model to file (default stdout)\n"
     << "\n"
     << "-cat-postp - get site category posterior probability for each site\n"
     << "  -in-model {file,-}           - read model from file\n"
     << "  -in-data {file,-}            - read site data from file\n"
     << "\n"
     << "-sim - simulate model data\n"
     << "  -in-model {file,-}         - read model from file\n"
     << "  -sim-size N                - size of simulated data, in sites\n"
     << "  [-sim-model]  (continuous) - simulator type\n"
     << "   continuous = continuous time simulator\n"
     << "   discrete   = discrete time simulator\n"
     << "   iss        = independent site sampler: alternate continuous time method; indy mutation models only\n"
     << "  [-sim-report-time]         - use simulator to estimate branch time. units: attempted substitutions/nuc (non-iss only)\n"
     << "  [-sim-assigned-cat-prob p] - probability of prior category knowledge at each site (default: 0.)\n"
     << "  [-sim-assigned-cat]        - category at each site is known (shortcut for -sim-assigned-cat-prob 1)\n"
     << "  [-site-model x]            - site model for simulated data, if different from model (default: in-model site type) (non-iss only)\n"
    //     << "  [-fsa-dir dir]             - write simulated data as a set of continuous sequence files in dir\n"
     << "  [-out {file,-}]            - write simulated site data to file (default stdout)\n"
     << "  [-seed N]                  - set random number seed\n"
     << "\n"
     << "-report-model - summarize model parameters\n"
     << "  -in-model {file,-}        - read model from file\n"
     << "  [-in-confidence {file,-}] - read free parameter confidence intervals from file\n"
     << "  [-out {file,-}]           - write report to file (default stdout)\n"
     << "\n"
     << "-report-data  - summarize model data\n"
     << "  -in-data {file,-}    - read site data from file\n"
     << "  [-pretty-print-data] - decode data and pretty print\n"
     << "  [-out {file,-}]      - write report to file (default stdout)\n"
     << "\n"
     << "-convert-data - convert data to new basic model type\n"
     << "  -site-model        - (see below)\n"
     << "  -in-data {file,-}  - read site data from file\n"
     << "  [-out {file,-}]    - write site data to file (default stdout)\n"
     << "\n"
     << "-ei - calculate (an approximation of) the evolutionary index matrix of Tang et al.\n"
     << "  -in-data {file,-}  - read site data from file, assumed to be only two species\n"
     << "  -org1\n"
     << "  -org2\n"
     << "\n\n";
  os << emphasis_tag << " MODEL FLAGS (used in create-model mode): " << emphasis_tag << "\n"
     << "\n"
     << "-tree-file file\n"
     << " file    = newick format tree, organism names must match those in data file\n"
     << "\n"
     << "-reversible-tree - lock one time parameter so that reversible models\n"
     << "                   are not overdetermined\n"
     << "\n"
     << "-site-model (nuc|dinuc|trinuc|codon|c4-pre|c4-post|c5|binary) - model site type\n"
     << " nuc     = nucleotide\n"
     << " dinuc   = overlapping dinucleotide conditioned to single nuc\n"
     << " trinuc  = overlapping trinucleotide conditioned to single nuc\n"
     << " codon   = codon\n"
     << " c4-pre  = (C4-) codon+preceding nucleotide conditioned to single codon\n"
     << " c4-post = (C4+) codon+following nucleotide conditioned to single codon\n"
     << " c5      = codon+preceding and following nucleotide conditioned to single codon\n"
     << " c5-pre  = codon+2 preceding nucleotides conditioned to single codon\n"
     << " c5-post = codon+2 following nucleotides conditioned to single codon\n"
     << " binary  = 2-state abstract model (for debugging)\n"
     << "\n"
     << "[-rate-model] (nonrev) - mutation rate type\n"
     << " jc69   = jukes-cantor\n"
     << " k80    = kimura-80\n"
     << " f81    = felsenstein-81\n"
     << " hky85  = hasegawa-kishino-yano-85\n"
     << " rev    = general time reversible\n"
     << " nonrev = unrestricted\n"
     << "\n"
     << "[-site-rate-cats] (1)  - mutation rate site categories\n"
     << "[-site-rate-cats-from-data]  - mutation rate categories from data-cat-map\n"
     << "\n"
     << "[-site-mutation-model-cats] (1)       - mutation model categories\n"
     << "[-site-mutation-model-cats-from-data] - mutation model categories from data-cat-map\n"
     << "\n"
     << "[-context-model] (indy) - mutation context\n"
     << " indy             = independent nucleotide parameters\n"
     << " pre-doublet      = f(N-1) context modifier\n"
     << " post-doublet     = g(N+1) context modifier\n"
     << " factored-triplet = f(N-1)*g(N+1) context modifier\n"
     << " triplet          = h(N-1,N+1) context modifier\n"
     << " cpg-only         = add two rates for CpG transitions and tranversions\n"
     << " tpa-only         = add two rates for TpA transitions and tranversions\n"
     << " cpg-1ti          = add one rate for CpG transition\n"
     << " cpg-2ti          = add two rates; one for each CpG transition\n"
     << " cpg-nonrev       = add 8 rates, one for every CpG context\n"
     << "\n"
     << "[-no-edge-correction] - turn off context edge correction\n"
     << "[-free-edge-correction] - free edge-correction strength parameters (default: locked at 1)\n"
     << "\n"
     << "[-select-model] (asymmetric) - aa selection (codon models only)\n"
     << " none              = no selection\n"
     << " single            = one aa selection parameter\n"
     << " hp                = hydrophobic-polar selection\n"
     << " from              = one selection parameter per source aa (20 param)\n"
     << " to                = one selection parameter per sink aa (20 param)\n"
     << " from-to           = product of source and sink aa parameters (40 param)\n"
     << " symmetric         = general aa symmetric selection (75 param)\n"
     << " from-symmetric    = product of source aa and symmetric exchange parameters (95 param)\n"
     << " symmetric-to      = product of sink aa and symmetric exchange parameters (95 param)\n"
     << " from-symmetric-to = product of source aa, sink aa, and symmetric exchange parameters (115 param)\n"
     << " asymmetric        = general aa selection (150 param)\n"
     << "\n"
     << "[-site-select-cats] (1) - number of aa selection categories\n"
     << "[-site-select-cats-from-data] - use categories from data file for site selection\n"
     << "\n"
     << "[-site-select-matrix-cats] (1) - number of aa exchange matrix categories\n"
     << "[-site-select-matrix-cats-from-data] - use categories from data file for site selection matrix\n"
     << "\n"
     << "[-data-cat-map] (default_map) - specify the mapping between model cats and data cats\n"
     << " map_syntax:\n"
     << "  data_cat_map = modelcat_label_rule1[,modelcat_label_rule2,...]\n"
     << "  modelcat_label_rule = modelcat_label:datacat_label1[+datacat+label2...]\n"
     << "   modelcat_label 'unassigned' is reserved to indicate an unassigned category\n"
     << " default_map = unassigned:datacat_label1+datacat_label2+...+datacat_labelN\n"
     << "\n"
#ifdef SHOW_EXPERIMENTAL_OPTIONS
     << "[-codon-bias-model] (none) - model selection on amino acid & codon usage\n"
     << " none = \n"
     << " synon-ratio = ratio for synonymous changes only (41 free param)\n"
     << "\n"
#endif
     << "[-c5-approx-model] (none) - approximation option for c5 model only\n"
     << " none     =\n"
     << " c4-joint = approx c5 lhood by conditioning 2 c4 blocks\n"
     << " c4-gmean = use geometric mean of 2 c4 blocks\n"
     << "\n"
     << "[-root-model] (full) - root state distro\n"
     << " full         = root from 1 parameter per state\n"
     << " codon        = root from nscodon distro (c4/c5 site-models)\n"
     << " codon-dinuc  = root from nscodon & codon edge dinuc distro (c4/c5 site-models)\n"
     << " nuc          = root from nuc distro (dinuc/trinuc/codon/c4 site-models)\n"
     << " nuc-pos      = root from codon position nuc distro (codon/c4 site-models)\n"
     << " obs-time-avg = root from time averaged observed leaf distros\n"
    //     << " post-prob    = root posterior probability\n"
     << " least-sq     = per-branch least-squares error (BLS) root\n"
     << "\n"
     << "[-lock-root]        - lock root distribution(s)\n"
     << "\n\n";
  os << emphasis_tag << " PERSISTENT MODEL SCORING FLAGS (used in ml/lhood modes, saved in model file): " << emphasis_tag << "\n"
     << "\n"
     << "[-tol]                - absolute convergence tolerance in log lhood units (default: " << DEFAULT_CONVERGE_TOLERANCE << ")\n"
     << "[-prefer-diag-expm]   - prefer matrix diagonalization to series for exp(M)\n"
     << "\n"
     << "[-min] (conj-hybrid)  - minimizer type\n"
     << " conj-dir      = conjugate direction set (powell 1964)\n"
     << " conj-grad     = conjugate gradient (polak-ribiere)\n"
     << " conj-hybrid   = " << DEFAULT_CONJ_GRAD_START_ITER << " iterations of conj-grad, followed by conj-dir\n"
#ifdef SHOW_EXPERIMENTAL_OPTIONS
     << " praxis        = praxis conjugate direction set method (brent 1973)\n"
     << " praxis-hybrid = " << DEFAULT_CONJ_GRAD_START_ITER << " iterations of conj-grad, followed by praxis\n"
     << " em            = use EM (using conj-dir for m-step min)\n"
#endif
     << "\n\n";
  os << emphasis_tag << " NONPERSISTENT MODEL SCORING FLAGS (used in ml/lhood modes, not saved in model file): " << emphasis_tag << "\n"
     << "\n"
     << "[-refine]             - start minimization at full tolerance\n"
//       << "[-start-tol]          - scale tol from start-tol to tol during minimization (default: 10)\n"
#ifdef INCLUDE_PPIT_MODE
     << "[-ppit-mode]          - use edge posterior prob training\n"
#endif
     << "[-max-steps N]        - stop minimization and report result after N iterations\n"
     << "[-pingpong file]      - periodically backup model state to file.{ping,pong}\n"
     << "\n";

  if(xmessage != "" ) {
    cerr << "\n"
         << "******** COMMAND-LINE ERROR:: " << xmessage << " ********\n"
         << "\n";
  }

  exit(EXIT_FAILURE);
}





static
void
get_tree_buffer(const std::string& treefile,
                std::string& tree_buffer){
  if(treefile.empty()) {
    pass_away("empty treefile");
  }
  ifstream tdata_is(treefile.c_str());
  check_nonempty_istream(tdata_is,treefile.c_str());

  string line_buffer;
  while(getline(tdata_is,line_buffer)){
    tree_buffer += line_buffer;
  }
}



/// \brief read in site data and convert directly to fast-index form
///
static
void
codeaxe_sdf_init(site_data_fastlup& sdf,
                 subs_ml_model& mdl,
                 std::istream& is){

  site_data sd;

  sd.load_state(is);
  const SITE_MODEL::index_t msm(mdl.get_rate_gtor().site_model());
  if(msm != sd.sm){
    std::cerr << "mdlsm/datasm: " << msm << " " << sd.sm << "\n";
    pass_away("Data file and model file site types conflict!");
  }

  // set untrained parameters, such as the background distribution, from data:
  init_subs_ml_model_untrained_data_dependencies(mdl,sd);

  mdl_site_data_fastlup_init(sd,mdl,sdf);
}




/// \brief (hacky) parser of data_catmap_str into catmap and labels
///
/// catmap is f(data_cat_label) -> model_assigned_cat_num
///
static
void
read_data_catmap(const std::string data_catmap_str,
                 data_catmap_type& data_catmap,
                 std::vector<std::string>& assigned_cat_labels){

  data_catmap.clear();

  if( data_catmap_str.empty() ) return;

  static const char rule_delimit(',');
  unsigned start(0);
  unsigned current_assigned_cat_no(UNASSIGNED_CAT_ID);
  while(true){
    const std::size_t end(data_catmap_str.find(rule_delimit,start));
    const std::string model_rule(data_catmap_str.substr(start,end-start));
    static const char section_delimit(':');
    {
      const std::size_t end2(model_rule.find(section_delimit));
      const std::string assigned_cat_label(model_rule.substr(0,end2));

      unsigned assigned_cat_no(UNASSIGNED_CAT_ID);
      if(assigned_cat_label != UNASSIGNED_CAT_LABEL){
        current_assigned_cat_no += 1;
        assigned_cat_no = current_assigned_cat_no;

        if(assigned_cat_labels.size() < (assigned_cat_no)) assigned_cat_labels.resize(assigned_cat_no);
        assigned_cat_labels[assigned_cat_no-1] = assigned_cat_label;
      }
      const std::string data_cats(model_rule.substr(end2+1));

      static const char data_delimit('+');
      unsigned start3(0);
      while(true){
        const std::size_t end3(data_cats.find(data_delimit,start3));
        const std::string data_cat_label(data_cats.substr(start3,end3-start3));
        data_catmap[data_cat_label] = assigned_cat_no;

        if(end3==std::string::npos) break;
        start3=end3+1;
      }
    }
    if(end==std::string::npos) break;
    start=end+1;
  }
}




static
void
try_main(int argc,char* argv[]){

  // mode independent setup::
  //
  progname=argv[0];

  long int seed(-1);

  std::ios_base::sync_with_stdio(false);

  // convert args to strings
  vector<string> argstr(argc);
  for(int i(0);i<argc;++i) { argstr[i]=argv[i]; }

  // save full command line for audit trail services
  string full_cmdline;
  for(int i(0);i<argc;++i){
    full_cmdline += ' ';
    full_cmdline += argv[i];
  }

  // mark arguments which have been processed:
  vector<bool> argmark(argc);
  for(int i(0);i<argc;++i){
    if(i==0) argmark[i]=true;
    else argmark[i]=false;
  }


  /////////////////////////////////////
  // setup other modeless flags:
  {
    bool seed_set(false);
    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;

      if       (argstr[i]=="-seed"){
        if(++i>=argc) { usage_error("No seed arg"); }
        if(seed_set) { usage_error("Seed already set"); }
	std::istringstream ia(argstr[i]);
	ia >> seed;
        seed_set=true;
      } else { continue; }

      argmark[i-1] = true;
      argmark[i] = true;
    }

    if(seed_set) {
      random_init(seed);
    } else {
      seed=random_init();
    }
  }

  //////////////////////////////////////////
  // determine program mode:
  MODE::index_t mode(MODE::NONE);

  for(int i(0);i<argc;++i){
    if(argmark[i]) continue;
    MODE::index_t new_mode(MODE::NONE);
    if     (argstr[i]=="-process-seq")  { new_mode = MODE::PROCESS; }
    else if(argstr[i]=="-ml")           { new_mode = MODE::TRAIN; }
    else if(argstr[i]=="-ei")           { new_mode = MODE::EI; }
    else if(argstr[i]=="-create-model")    { new_mode = MODE::DENOVO; }
    else if(argstr[i]=="-norm")         { new_mode = MODE::NORM; }
    else if(argstr[i]=="-lhood")        { new_mode = MODE::LHOOD; }
    else if(argstr[i]=="-cat-postp")    { new_mode = MODE::CAT_POSTP; }
    else if(argstr[i]=="-sim")          { new_mode = MODE::SIM; }
    else if(argstr[i]=="-report-model") { new_mode = MODE::REPORT_MODEL; }
    else if(argstr[i]=="-report-data")  { new_mode = MODE::REPORT_DATA; }
    else if(argstr[i]=="-report-seq")   { new_mode = MODE::REPORT_SEQ; }
    else if(argstr[i]=="-convert-data") { new_mode = MODE::CONVERT_DATA; }
    else if(argstr[i]=="-confidence")   { new_mode = MODE::CONFIDENCE; }
    else { continue; }

    if(mode != MODE::NONE) usage_error("multiple modes selected");
    mode = new_mode;
    argmark[i]=true;
  }
  if(mode == MODE::NONE) usage_error("No mode selected");



  ///////////////////////////////////////
  // setup primary file stream info:
  string outfile;
  string indatafile,inmodelfile;

  for(int i(0);i<argc;++i){
    if(argmark[i]) continue;
    string* str_ptr(0);
    if     (argstr[i]=="-out")     { str_ptr=&outfile; }
    else if(argstr[i]=="-in-data") { str_ptr=&indatafile; }
    else if(argstr[i]=="-in-model"){ str_ptr=&inmodelfile; }
    else { continue; }

    if(++i>=argc || (argstr[i].empty() || (argstr[i].size()>1 && argstr[i][0]=='-'))){
      usage_error("no filename following "+argstr[i-1]);
    }
    if( ! str_ptr->empty() ) usage_error(argstr[i-1]+" assigned multiple times");
    str_ptr->operator=(argstr[i]);
    argmark[i-1] = true;
    argmark[i] = true;
  }

  if(! inmodelfile.empty() && inmodelfile == indatafile){
    usage_error("-in-data and -in-model args match");
  }

  // setup primary output file streams::
  //
  //  ostream& log_os(std::cerr);

  // setup input-model stream:
  istream* model_isp(0);
  ifstream model_isf;
  if(! inmodelfile.empty() ){
    if( inmodelfile == stdstream_tag ){
      model_isp = &cin;
    } else {
      model_isf.open(inmodelfile.c_str());
      model_isp = &model_isf;
    }
    check_nonempty_istream(*model_isp,inmodelfile.c_str());
  }

  // setup input-data stream
  istream* data_isp(0);
  ifstream data_isf;
  if(! indatafile.empty() ){
    if( indatafile == stdstream_tag ){
      data_isp = &cin;
    } else {
      data_isf.open(indatafile.c_str());
      data_isp = &data_isf;
    }
    check_nonempty_istream(*data_isp,indatafile.c_str());
  }


  const bool is_no_stream_out_mode(mode == MODE::CONFIDENCE);

  // setup output stream
  ostream* main_osp(&cout);
  ofstream main_osf;
  if( ! (outfile.empty() || outfile == stdstream_tag || is_no_stream_out_mode) ){
    main_osf.open(outfile.c_str());
    main_osp = &main_osf;
  }
  ostream& main_os(*main_osp);



  // setup site model type:
  //
  RATE_GTOR_MODEL::index_t rate_gtor_model(RATE_GTOR_MODEL::NONE);
  if(mode == MODE::DENOVO || mode == MODE::PROCESS ||
     mode == MODE::CONVERT_DATA || mode == MODE::SIM){
    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;

      if       (argstr[i]=="-site-model"){
        if(rate_gtor_model != RATE_GTOR_MODEL::NONE) usage_error("Multiple site-model args");
        if(++i>=argc) { usage_error("No site-model arg"); }
        if       (argstr[i]=="nuc"){
          rate_gtor_model = RATE_GTOR_MODEL::NUC;
        } else if(argstr[i]=="dinuc"){
          rate_gtor_model = RATE_GTOR_MODEL::DINUC;
        } else if(argstr[i]=="trinuc"){
          rate_gtor_model = RATE_GTOR_MODEL::TRINUC;
        } else if(argstr[i]=="codon"){
          rate_gtor_model = RATE_GTOR_MODEL::CODON;
        } else if(argstr[i]=="c4-pre"){
          rate_gtor_model = RATE_GTOR_MODEL::C4PRE;
        } else if(argstr[i]=="c4-post"){
          rate_gtor_model = RATE_GTOR_MODEL::C4POST;
        } else if(argstr[i]=="c5"){
          rate_gtor_model = RATE_GTOR_MODEL::C5;
        } else if(argstr[i]=="c5-pre"){
          rate_gtor_model = RATE_GTOR_MODEL::C5PRE;
        } else if(argstr[i]=="c5-post"){
          rate_gtor_model = RATE_GTOR_MODEL::C5POST;
        } else if(argstr[i]=="binary"){
          rate_gtor_model = RATE_GTOR_MODEL::BINARY;
        } else { usage_error("Unknown site-model arg"); }
      } else { continue; }

      argmark[i-1] = true;
      argmark[i] = true;
    }
  }



  ///////////////////////////////////////
  if(mode == MODE::NORM){
    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(! model_isp) usage_error("mode requires model file");
    if(data_isp) usage_error("mode does not use data file");

    subs_ml_model mdl(*model_isp,full_cmdline);

    norm_model(mdl);

    mdl.store_state(main_os);
  }



  //////////////////////////////////////////
  else if(mode == MODE::CAT_POSTP){
    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(! model_isp) usage_error("mode requires model file");
    if(! data_isp) usage_error("mode requires data file");

    subs_ml_model mdl(*model_isp,full_cmdline);

    site_data_fastlup sdf;
    codeaxe_sdf_init(sdf,mdl,*data_isp);

    smlfloat lnp,lnp_norm;
    score_subs_ml_model(sdf,mdl,lnp,lnp_norm,true);
  }



  //////////////////////////////////////////////
  else if(mode == MODE::CONVERT_DATA){
    if(rate_gtor_model == RATE_GTOR_MODEL::NONE) {
      usage_error("no destination site-model type selected in convert-data mode");
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(model_isp) usage_error("mode does not use model file");
    if(! data_isp) usage_error("mode requires data file");

    site_data sd;
    sd.load_state(*data_isp);

    const SITE_MODEL::index_t r_in(sd.sm);
    const SITE_MODEL::index_t r_out(RATE_GTOR_MODEL::convert_to_site_model(rate_gtor_model));

    if(r_in == r_out){
      // trivial_case -- do nothing
    } else if(r_in == SITE_MODEL::NSC5) {
      if(r_out == SITE_MODEL::NSC4PRE) {
        site_data_nsc5_to_nsc4pre(sd);
      } else if (r_out == SITE_MODEL::NSC4POST) {
        site_data_nsc5_to_nsc4post(sd);
      } else if (r_out == SITE_MODEL::NSCODON) {
        site_data_nsc5_to_nscodon(sd);
      } else {
        pass_away("Invalid data conversion");
      }
    } else if((r_in == SITE_MODEL::NSC4PRE || r_in == SITE_MODEL::NSC4POST)
              && r_out == SITE_MODEL::NSCODON){
      site_data_nsc4_to_nscodon(sd);
    } else {
      pass_away("Invalid data conversion");
    }

    sd.sm = r_out;
    sd.store_state(main_os,full_cmdline.c_str());
  }



  //////////////////////////////////////////
  else if(mode==MODE::REPORT_MODEL){

    std::string ci_file;

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      if(argstr[i]=="-in-confidence"){
        if(++i>=argc) usage_error();
        ci_file = argstr[i];
        argmark[i-1] = true;
        argmark[i] = true;
        break;
      }
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(! model_isp) usage_error("mode requires model file");
    if(data_isp) usage_error("mode does not use data file");

    subs_ml_model mdl(*model_isp,full_cmdline);

    if(!ci_file.empty()){
      subs_ml_model_ci_info mdl_ci;

      std::ifstream infp(ci_file.c_str());
      mdl_ci.load_state(infp);
      mdl.attach_ci(mdl_ci.val.begin());
    }

    mdl.report(main_os);
  }


  //////////////////////////////////////////
  else if(mode == MODE::REPORT_DATA) {

    bool is_pretty_print_data(false);

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      if(argstr[i]=="-pretty-print-data"){
        is_pretty_print_data = true;
        argmark[i] = true;
        break;
      }
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(model_isp) usage_error("mode does not use model file");
    if(! data_isp) usage_error("mode requires data file");


    site_data sd;
    sd.load_state(*data_isp);
    if(sd.sm != SITE_MODEL::NONE){
      site_data_report(sd,is_pretty_print_data,main_os);
    }
  }


  //////////////////////////////////////////
  else if(mode == MODE::EI) {

    string org1,org2;

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      const int starti(i);
      if       (argstr[i]=="-org1"){
        if(++i>=argc) usage_error("No org1 argument");
        org1 = argstr[i];
      } else if(argstr[i]=="-org2"){
        if(++i>=argc) usage_error("No org2 argument");
        org2 = argstr[i];
      } else { continue; }

      argmark[starti] = true;
      argmark[i] = true;
    }

    if(org1.empty() || org2.empty()) usage_error("org1 and org2 required in ei mode");

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(model_isp) usage_error("mode does not use model file");
    if(! data_isp) usage_error("mode requires data file");


    site_data sd;
    sd.load_state(*data_isp);

    get_ei(sd,org1,org2);
  }



  /////////////////////////////////////////
  else if(mode == MODE::SIM){
    sim_options sim_opt;
#if 0
    std::string seq_file_node;
#endif

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      int starti(i);

      if       (argstr[i]=="-sim-size"){
        if(++i>=argc) usage_error();
	std::istringstream ia(argstr[i]);
	ia >> sim_opt.size;
        if(sim_opt.size < 0 ) usage_error("Invalid sim size");

      } else if(argstr[i]=="-sim-model"){
        if(++i>=argc) { usage_error("No site-model arg"); }
        if       (argstr[i]=="continuous"){
          sim_opt.method = SIM_MODEL::CONTINUOUS;
        } else if(argstr[i]=="discrete"){
          sim_opt.method = SIM_MODEL::DISCRETE;
        } else if(argstr[i]=="iss"){
          sim_opt.method = SIM_MODEL::ISS;
        } else { usage_error((string("Unknown sim-model arg: ")+argstr[i]).c_str()); }

      } else if(argstr[i]=="-sim-assigned-cat-prob"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> sim_opt.assigned_cat_prob;
        if(sim_opt.assigned_cat_prob < 0.) usage_error("Invalid sim-assigned-cat-prob");

      } else if(argstr[i]=="-sim-assigned-cat"){
        sim_opt.assigned_cat_prob = 2.;

      } else if(argstr[i]=="-sim-report-time"){
        sim_opt.is_report_time = true;

#if 0
      } else if(argstr[i]=="-out-seq"){
        if(++i>=argc) usage_error();
        seq_file_node = argv[i];
#endif
      } else { continue; }

      argmark[starti] = true;
      argmark[i] = true;
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(! model_isp) usage_error("mode requires model file");
    if(data_isp) usage_error("mode does not use data file");

    site_data sd;
    subs_ml_model mdl(*model_isp,full_cmdline);

    const RATE_GTOR_MODEL::index_t rgm(mdl.rate_gtor_model());

    if(rate_gtor_model != RATE_GTOR_MODEL::NONE) {
      if(sim_opt.method == SIM_MODEL::ISS && rate_gtor_model != rgm){
        usage_error("cannot change site model when using iss simulator");
      }

      sim_opt.output_site_model = rate_gtor_model;
    } else {
      sim_opt.output_site_model = rgm;
    }

    simulate_data(sim_opt,mdl,sd);

    sd.store_state(main_os,full_cmdline.c_str(),seed);
  }



  ////////////////////////////////////////////
  else if(mode == MODE::PROCESS || mode == MODE::REPORT_SEQ){

    if(model_isp) usage_error("mode does not use  model file");
    if(data_isp) usage_error("mode does not use data file");

    string seq_file_node,cat_file_node,cat_label_file;

    double gc_max(1.0);
    double gc_min(0.0);

    bool is_nuc_4fold_filter(false);
    bool is_codon_border(true);

    bool is_no_adjacent_nuc_diff(false);
    bool is_single_nuc_diff(false);

    seq_filter_options fopt;

    fopt.min_match = DEFAULT_MIN_MATCH;

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      const int starti(i);

      if(argstr[i]=="-max-align-count"){
        if(++i>=argc) usage_error();
	std::istringstream ia(argstr[i]);
	ia >> fopt.max_align_count;

      } else if(argstr[i]=="-4fold"){
        is_nuc_4fold_filter = true;

      } else if(argstr[i]=="-gc-max"){
        if(++i>=argc) usage_error();
	std::istringstream ia(argstr[i]);
	ia >> gc_max;

      } else if(argstr[i]=="-gc-min"){
        if(++i>=argc) usage_error();
	std::istringstream ia(argstr[i]);
	ia >> gc_min;

      } else if(argstr[i]=="-no-window-filter"){
        fopt.is_use_window_filter = false;

      } else if(argstr[i]=="-no-ambig-gap-filter"){
        fopt.is_use_ambi_filter = false;
        fopt.is_use_gap_filter = false;

      } else if(argstr[i]=="-no-filter"){
        fopt.is_skip_filters = true;

      } else if(argstr[i]=="-min-window-match"){
        if(++i>=argc) usage_error();
	std::istringstream ia(argstr[i]);
	ia >> fopt.min_match;

      } else if(argstr[i]=="-no-codon-border"){
        is_codon_border = false;

      } else if(argstr[i]=="-in-seq"){
        if(++i>=argc) usage_error();
        seq_file_node = argv[i];

      } else if(argstr[i]=="-in-cat"){
        if(++i>=argc) usage_error();
        cat_file_node = argv[i];

      } else if(argstr[i]=="-no-adjacent-nuc-diff"){
        is_no_adjacent_nuc_diff = true;

      } else if(argstr[i]=="-only-single-nuc-diff"){
        is_single_nuc_diff = true;

      } else if(argstr[i]=="-cat-labels"){
        if(++i>=argc) usage_error();
        cat_label_file = argv[i];

      } else { continue; }

      argmark[starti] = true;
      argmark[i] = true;
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    bool is_readseq_ready(! seq_file_node.empty());
    if( ! is_readseq_ready ) usage_error("no -in-seq argument");

    if( mode == MODE::PROCESS && rate_gtor_model == RATE_GTOR_MODEL::NONE ){
      usage_error("no site-model selected in process-seq mode");
    }

    nuc_seq_data in_seq(&full_cmdline);
    process_seq_data(in_seq,fopt,seq_file_node.c_str(),
                     cat_file_node.c_str(),cat_label_file.c_str());

    if(mode == MODE::REPORT_SEQ){
      nuc_seq_data_report(in_seq,main_os);
    } else {
      site_data sd;
      // convert filtered nucleotide data into site-model specific data for further analysis
      nuc_seq_to_site_data(sd,in_seq,rate_gtor_model,is_codon_border,
                           is_nuc_4fold_filter,gc_max,gc_min,false,is_no_adjacent_nuc_diff,is_single_nuc_diff);
      sd.store_state(main_os,full_cmdline.c_str());
    }
  }



  //////////////////////////////////////////////
  else if(mode == MODE::CONFIDENCE){

    conf_options co;
    bool is_method_set(false);
    bool is_selection_only(false);
    bool is_mutation_only(false);

    // set any new model training options:
    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      const int starti(i);
      if       (argstr[i]=="-ci-alpha"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> co.alpha;

      } else if(argstr[i]=="-ci-fisher"){
        if(is_method_set) usage_error("multiple ci methods defined");
        co.is_fisher_method=true;
        is_method_set=true;

      } else if(argstr[i]=="-ci-search"){
        if(is_method_set) usage_error("multiple ci methods defined");
        co.is_fisher_method=false;
        is_method_set=true;

      } else if(argstr[i]=="-ci-mutation-only"){
        is_mutation_only=true;

      } else if(argstr[i]=="-ci-selection-only"){
        is_selection_only=true;

      } else if(argstr[i]=="-ci-single"){
        co.is_single_param=true;

      } else { continue; }

      argmark[starti] = true;
      argmark[i] = true;
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    if(! model_isp) usage_error("mode requires model file");
    if(! data_isp) usage_error("mode requires data file");
    if(outfile.empty()) usage_error("mode requires explicit filename for output");

    if(is_mutation_only && is_selection_only){
      usage_error("only one ci subset flag allowed");
    }

    if(co.is_fisher_method){
      if(co.is_single_param){
        usage_error("single param option is meaningless for fisher conf interval calculation");
      }
      if(is_mutation_only){
       usage_error("mutation param only option is meaningless for fisher conf interval calculation");
      }
      if(is_selection_only){
       usage_error("selection param only option is meaningless for fisher conf interval calculation");
      }
    }

    // read model:
    subs_ml_model mdl(*model_isp,full_cmdline);

    // read site data, store in fastlup format required by lhood-calling functions:
    site_data_fastlup sdf;
    codeaxe_sdf_init(sdf,mdl,*data_isp);

    // setup mask on ci calculations:
    const unsigned ps(mdl.param_size());
    std::vector<bool> ci_mask(ps,false);

    /// \todo improve brittle hacks used to set ci subsets
    ///
    if(is_mutation_only){
      std::fill(ci_mask.begin(),ci_mask.end(),true);

      const rate_gtor_nuc_base* rgns(dynamic_cast<const rate_gtor_nuc_base*>(&mdl.get_rate_gtor()));
      if(rgns){
        const unsigned st(mdl.get_time_gtor().param_size()+rgns->param_start_mut());
        const unsigned s(rgns->param_size_mut());
        for(unsigned i(st);i<(st+s);++i){
          ci_mask[i] = false;
        }
      }
    } else if(is_selection_only){
      std::fill(ci_mask.begin(),ci_mask.end(),true);

      const rate_gtor_nscodon_base* rgns(dynamic_cast<const rate_gtor_nscodon_base*>(&mdl.get_rate_gtor()));
      if(rgns){
        const unsigned st(mdl.get_time_gtor().param_size()+rgns->param_start_aa());
        const unsigned s(rgns->param_size_aa());
        for(unsigned i(st);i<(st+s);++i){
          ci_mask[i] = false;
        }
      }
    }

    get_subs_ml_model_conf_interval(sdf,mdl,co,ci_mask,outfile);
  }



  //////////////////////////////////////////////
  else if(mode == MODE::TRAIN || mode == MODE::LHOOD){

    if(! model_isp) usage_error("mode requires model file");
    if(! data_isp) usage_error("mode requires data file");

    // read model:
    bool is_refine(false);
    subs_ml_model mdl(*model_isp,full_cmdline);

    // set any new model training options:
    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      const int starti(i);
      if(argstr[i]=="-tol"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> mdl.opt().converge_value;

      } else if(argstr[i]=="-max-steps"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> mdl.opt().max_steps;

      } else if(argstr[i]=="-pingpong"){
        if(++i>=argc) usage_error();
        mdl.opt().is_pingpong = true;
        mdl.opt().pingpongfile = argv[i];

      } else if(argstr[i]=="-min"){
        if(++i>=argc) usage_error();
        if       (argstr[i]=="conj-dir"){
          mdl.opt().min = MIN::CONJ_DIR;
        } else if(argstr[i]=="conj-grad"){
          mdl.opt().min = MIN::CONJ_GRAD;
        } else if(argstr[i]=="conj-hybrid"){
          mdl.opt().min = MIN::CG_CD_HYBRID;
        } else if(argstr[i]=="praxis"){
          mdl.opt().min = MIN::PRAXIS;
        } else if(argstr[i]=="praxis-hybrid"){
          mdl.opt().min = MIN::CG_PRAXIS_HYBRID;
        } else if(argstr[i]=="em"){
          mdl.opt().min = MIN::EM;
        }

      } else if(argstr[i]=="-prefer-diag-expm"){
        mdl.opt().is_prefer_diag_expm = true;

#ifdef INCLUDE_PPIT_MODE
      } else if(argstr[i]=="-ppit-mode"){
        mdl.opt().is_ppit = true;
#endif

      } else if(argstr[i]=="-refine"){
        is_refine = true;

      } else { continue; }

      argmark[starti] = true;
      argmark[i] = true;
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    // read site data, store in fastlup format required by lhood-calling functions:
    site_data_fastlup sdf;
    codeaxe_sdf_init(sdf,mdl,*data_isp);

    if(mode == MODE::TRAIN){
      try {
        train_subs_ml_model(sdf,mdl,is_refine);
      } catch (substk_exception& s) {
        std::cerr << "SubsTK EXCEPTION:: " << s.what() << "\n";
        mdl.store_state(std::cerr);
        abort();
      } catch(math_util_exception& e) {
        std::cerr << "FATAL:: MATH UTIL EXCEPTION: " << e.what() << "\n";
        mdl.store_state(std::cerr);
        abort();
      }

      mdl.store_state(main_os);
    } else if(mode == MODE::LHOOD){
      smlfloat lnp,lnp_norm;
      score_subs_ml_model(sdf,mdl,lnp,lnp_norm);

      const unsigned site_count(sdf.total_count);
      lnp_line_report(lnp,lnp_norm,site_count,main_os);
    }
  }



  /////////////////////////////////////////////
  else if(mode == MODE::DENOVO){

    if(model_isp) usage_error("mode does not use model file");

    string treefile;
    time_gtor_options tgo;

    DENOVO_MODEL::index_t denovo_model(DENOVO_MODEL::START);

    ROOT_GTOR_MODEL::index_t root_gtor_model(ROOT_GTOR_MODEL::FULL);

    rate_gtor_nuc_options rgt_nuc_opt;
    rate_gtor_nscodon_options rgt_nscodon_opt;

    C5_APPROX::index_t c5_approx_model(C5_APPROX::NONE);

    std::vector<bool> is_assigned_cats(SITE_CAT_TYPE::SIZE,false);
    std::vector<bool> is_user_n_site_cats(SITE_CAT_TYPE::SIZE,false);
    cat_manager_options catman_opt;
    bool is_lock_root(false);
    bool is_free_edge_correction(false);
    string data_cat_map_str;

    for(int i(0);i<argc;++i){
      if(argmark[i]) continue;
      const int starti(i);

      if       (argstr[i]=="-tree-file"){
        if(++i>=argc) { usage_error("No tree-file arg"); }
        treefile = argv[i];

      } else if(argstr[i]=="-reversible-tree"){
        tgo.is_reversible=true;

      } else if(argstr[i]=="-rate-model"){
        if(++i>=argc) usage_error("No rate-model arg");
        if       (argstr[i]=="jc69"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::JC69);
        } else if(argstr[i]=="k80"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::K80);
        } else if(argstr[i]=="f81"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::F81);
        } else if(argstr[i]=="hky85"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::HKY85);
        } else if(argstr[i]=="rev"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::REV);
        } else if(argstr[i]=="nonrev"){
          rgt_nuc_opt.set_all_rate_model(RATE_MODEL_NUC::NONREV);
        } else { usage_error((string("unknown rate-model arg: ")+argstr[i]).c_str()); }

      } else if(argstr[i]=="-site-rate-cats"){
        if(++i>=argc) usage_error("Missing site-rate-cats arg");
        std::istringstream ia(argstr[i]);
        ia >> catman_opt.n_site_cats[SITE_CAT_TYPE::MUT_RATE];
        is_user_n_site_cats[SITE_CAT_TYPE::MUT_RATE]=true;
        if(catman_opt.n_site_cats[SITE_CAT_TYPE::MUT_RATE] < 1) usage_error("Invalid site-rate-cats arg");

      } else if(argstr[i]=="-site-mutation-model-cats"){
        if(++i>=argc) usage_error("Missing site-mutation-model-cats arg");
        std::istringstream ia(argstr[i]);
        ia >> catman_opt.n_site_cats[SITE_CAT_TYPE::MUT_MODEL];
        is_user_n_site_cats[SITE_CAT_TYPE::MUT_MODEL] = true;
        if(catman_opt.n_site_cats[SITE_CAT_TYPE::MUT_MODEL] < 1) usage_error("Invalid site-mutation-model-cats arg");

      } else if(argstr[i]=="-context-model"){
        if(++i>=argc) usage_error("no context-model arg");
        if       (argstr[i]=="indy"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::INDY);
        } else if(argstr[i]=="pre-doublet"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::PRE_DOUBLET);
        } else if(argstr[i]=="post-doublet"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::POST_DOUBLET);
        } else if(argstr[i]=="factored-triplet"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::FACTORED_TRIPLET);
        } else if(argstr[i]=="triplet"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::TRIPLET);
        } else if(argstr[i]=="cpg-only"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::CPG_ONLY);
        } else if(argstr[i]=="tpa-only"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::TPA_ONLY);
        } else if(argstr[i]=="cpg-1ti"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::CPG_1TI);
        } else if(argstr[i]=="cpg-2ti"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::CPG_2TI);
        } else if(argstr[i]=="cpg-nonrev"){
          rgt_nuc_opt.set_all_context_model(CONTEXT_MODEL_NUC::CPG_NONREV);
        } else { usage_error((string("unknown context-model arg: ")+argstr[i]).c_str()); }

      } else if(argstr[i]=="-no-edge-correction"){
        rgt_nuc_opt.is_use_edge_correction=false;

      } else if(argstr[i]=="-free-edge-correction"){
        is_free_edge_correction=true;

      } else if(argstr[i]=="-select-model"){
        if(++i>=argc) usage_error("no select-model arg");
        if       (argstr[i]=="none"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::NONE);
        } else if(argstr[i]=="single"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::SINGLE);
        } else if(argstr[i]=="hp"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::HP);
        } else if(argstr[i]=="from"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::FROM);
        } else if(argstr[i]=="to"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::TO);
        } else if(argstr[i]=="from-to"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::FROM_TO);
        } else if(argstr[i]=="symmetric"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::SYMM);
        } else if(argstr[i]=="from-symmetric"){
          /// \todo -- check from-symmetric for a normalization trap!!
          usage_error("from-symmetric selection suspected of minimization trap due to incorrect normalization.");
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::FROM_SYMM);
        } else if(argstr[i]=="symmetric-to"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::SYMM_TO);
        } else if(argstr[i]=="from-symmetric-to"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::FROM_SYMM_TO);
        } else if(argstr[i]=="asymmetric"){
          rgt_nscodon_opt.set_all_select_model(SELECT_MODEL::ASYMM);
        } else { usage_error((string("unknown select-model arg: ")+argstr[i]).c_str()); }

      } else if(argstr[i]=="-site-select-cats"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> catman_opt.n_site_cats[SITE_CAT_TYPE::SEL_STRENGTH];
        is_user_n_site_cats[SITE_CAT_TYPE::SEL_STRENGTH] = true;
        if(catman_opt.n_site_cats[SITE_CAT_TYPE::SEL_STRENGTH] < 1) usage_error();

      } else if(argstr[i]=="-site-select-matrix-cats"){
        if(++i>=argc) usage_error();
        std::istringstream ia(argstr[i]);
        ia >> catman_opt.n_site_cats[SITE_CAT_TYPE::SEL_MATRIX];
        is_user_n_site_cats[SITE_CAT_TYPE::SEL_MATRIX] = true;
        if(catman_opt.n_site_cats[SITE_CAT_TYPE::SEL_MATRIX] < 1) usage_error();

      } else if(argstr[i]=="-codon-bias-model"){
        if(++i>=argc) usage_error();
        if       (argstr[i]=="none"){
          rgt_nscodon_opt.codon_bias_model = CODON_BIAS_MODEL::NONE;
        } else if(argstr[i]=="synon-ratio"){
          rgt_nscodon_opt.codon_bias_model = CODON_BIAS_MODEL::SYNON_RATIO;
        } else if(argstr[i]=="synon-norm-ratio"){
          rgt_nscodon_opt.codon_bias_model = CODON_BIAS_MODEL::SYNON_NORM_RATIO;
        } else { usage_error("unknown -codon-bias-model arg"); }

      } else if(argstr[i]=="-c5-approx-model"){
        if(++i>=argc) usage_error();
        if       (argstr[i]=="none"){
          c5_approx_model = C5_APPROX::NONE;
        } else if(argstr[i]=="c4-joint"){
          c5_approx_model = C5_APPROX::C4_JOINT;
        } else if(argstr[i]=="c4-gmean"){
          c5_approx_model = C5_APPROX::C4_GMEAN;
        } else { usage_error("unknown c5-approx arg"); }

      } else if(argstr[i]=="-root-model"){
        if(++i>=argc) usage_error();
        if       (argstr[i]=="full"){
          root_gtor_model = ROOT_GTOR_MODEL::FULL;
        } else if(argstr[i]=="codon"){
          root_gtor_model = ROOT_GTOR_MODEL::CODON;
        } else if(argstr[i]=="codon-dinuc"){
          root_gtor_model = ROOT_GTOR_MODEL::CODON_DINUC;
        } else if(argstr[i]=="nuc"){
          root_gtor_model = ROOT_GTOR_MODEL::NUC;
        } else if(argstr[i]=="nuc-pos"){
          root_gtor_model = ROOT_GTOR_MODEL::NUC_POS;
        } else if(argstr[i]=="obs-time-avg"){
          root_gtor_model = ROOT_GTOR_MODEL::OBS_TIME_AVG;
        } else if(argstr[i]=="post-prob"){
          root_gtor_model = ROOT_GTOR_MODEL::PPROB;
        } else if(argstr[i]=="least-sq"){
          root_gtor_model = ROOT_GTOR_MODEL::LSPROB;
        } else { usage_error("unknown root-model arg"); }

      } else if(argstr[i]=="-random-param"){
        denovo_model = DENOVO_MODEL::RANDOM;

      } else if(argstr[i]=="-site-rate-cats-from-data"){
        is_assigned_cats[SITE_CAT_TYPE::MUT_RATE] = true;

      } else if(argstr[i]=="-site-mutation-model-cats-from-data"){
        is_assigned_cats[SITE_CAT_TYPE::MUT_MODEL] = true;

      } else if(argstr[i]=="-site-select-cats-from-data"){
        is_assigned_cats[SITE_CAT_TYPE::SEL_STRENGTH] = true;

      } else if(argstr[i]=="-site-select-matrix-cats-from-data"){
        is_assigned_cats[SITE_CAT_TYPE::SEL_MATRIX] = true;

      } else if(argstr[i]=="-data-cat-map"){
        if(++i>=argc || (argstr[i].empty() || argstr[i][0]=='-')) usage_error("no arg to -data-cat-map");
        data_cat_map_str=argstr[i];

      } else if(argstr[i]=="-lock-root"){
        is_lock_root=true;

      } else { continue;}

      argmark[starti] = true;
      argmark[i] = true;
    }

    for(int i(0);i<argc;++i){ if(! argmark[i]) usage_error(string("Invalid argument: ")+argstr[i]); }

    // check input flags:
    if( rate_gtor_model == RATE_GTOR_MODEL::NONE ){
      usage_error("no site-model selected in create-model mode");
    }

    bool is_any_assigned_cats(false);
    {
      unsigned sc_count(0);
      for(unsigned i(0);i<SITE_CAT_TYPE::SIZE;++i){
        if(is_assigned_cats[i]) sc_count++;
      }
      if( sc_count > 1){
        pass_away("can only assign data cats to one cat type");
      } else if( sc_count == 1){
        is_any_assigned_cats=true;
      }
    }

    for(unsigned i(0);i<SITE_CAT_TYPE::SIZE;++i){
      if(is_assigned_cats[i]){
        for(unsigned j(0);j<SITE_CAT_TYPE::SIZE;++j){
          if(i==j) continue;
          if( catman_opt.n_site_cats[j] != 1 ){
            pass_away("can't mix assigned & standard site categories");
          }
        }
      }
    }

    if( data_cat_map_str.empty() && is_any_assigned_cats ) {
      pass_away("data-cat-map must be specified when any site cat is linked to data.");
    }

    if( rate_gtor_model == RATE_GTOR_MODEL::BINARY &&
        is_any_assigned_cats ){
      pass_away("wrong site-model for categories");
    }

    if(is_assigned_cats[SITE_CAT_TYPE::SEL_STRENGTH] ||
       is_assigned_cats[SITE_CAT_TYPE::SEL_MATRIX]){
      if(RATE_GTOR_MODEL::base_size_conditioned(rate_gtor_model) != CODON::BASE_SIZE){
        pass_away("wrong site-model for categories");
      }
    }

    if(is_free_edge_correction && (! rgt_nuc_opt.is_use_edge_correction)){
      usage_error("incompatible edge correction options");
    }

    // process data_cat_map_str:
    //
    std::vector<std::string> assigned_cat_labels;
    read_data_catmap(data_cat_map_str,catman_opt.data_catmap,assigned_cat_labels);

    const unsigned n_assigned_cats(assigned_cat_size(catman_opt.data_catmap));

    if(is_any_assigned_cats){
      // set the site cat count from data_cat map:
      for(unsigned i(0);i<SITE_CAT_TYPE::SIZE;++i){
        if(is_assigned_cats[i]){
          if(is_user_n_site_cats[i])
            pass_away("Invalid flag combination: number of site categories already assigned from data-cat-map.");
          if(n_assigned_cats == 0)
            pass_away("Assigning site cats from data without any assigned data categories.");

          catman_opt.n_site_cats[i] = n_assigned_cats;
          break;
        }
      }

    } else {
      if(n_assigned_cats > 0) pass_away("data-cat-map does not agree with assigned data flags");
    }

    // initialize model
    get_tree_buffer(treefile,tgo.tree_buffer);
    subs_ml_model mdl(tgo,rate_gtor_model,rgt_nuc_opt,rgt_nscodon_opt,
                      c5_approx_model,root_gtor_model,catman_opt,
                      full_cmdline,denovo_model);

    /////////////////////////////////////////////////
    /// after ctor mdl mods:
    ///
    if(is_lock_root) mdl.get_root_gtor().lock_root();

    /// \todo this is embarrassing... find a way to make this happen in
    /// the ctor instead
    if(! is_free_edge_correction){
      rate_gtor_nuc_base* rgn(dynamic_cast<rate_gtor_nuc_base*>(&mdl.get_rate_gtor()));
      if(rgn) rgn->lock_full_edge_correction();
    }

    /// \todo move labels to cat_manager:
    if(assigned_cat_labels.size()){
      if       (is_assigned_cats[SITE_CAT_TYPE::MUT_RATE]){
        static_cast<rate_gtor_nuc_base&>(mdl.get_rate_gtor()).site_rate_cats_labels() = assigned_cat_labels;
      } else if(is_assigned_cats[SITE_CAT_TYPE::MUT_MODEL]){
        static_cast<rate_gtor_nuc_base&>(mdl.get_rate_gtor()).site_mutation_model_cats_labels() = assigned_cat_labels;
      } else if(is_assigned_cats[SITE_CAT_TYPE::SEL_STRENGTH]){
        static_cast<rate_gtor_nscodon_base&>(mdl.get_rate_gtor()).site_select_cats_labels() = assigned_cat_labels;
      } else if(is_assigned_cats[SITE_CAT_TYPE::SEL_MATRIX]){
        static_cast<rate_gtor_nscodon_base&>(mdl.get_rate_gtor()).site_select_matrix_cats_labels() = assigned_cat_labels;
      }
    }


    // read (optional) site data:
    if(data_isp){

      site_data sd;
      sd.load_state(*data_isp);
      if(mdl.get_rate_gtor().site_model() != sd.sm){
        std::cerr << "rgm: " << mdl.get_rate_gtor().site_model() << " sd: " << sd.sm << "\n";
        pass_away("Data file and model file site types conflict!");
      }


      // check data_catmap labels, and set priors to model_cats
      data_catmap_type::const_iterator i=mdl.get_cat_manager().data_catmap().begin(),i_end=mdl.get_cat_manager().data_catmap().end();
      for(;i!=i_end;++i){
        if( i->first != UNASSIGNED_CAT_LABEL &&
            sd.catlabel_2_id.find(i->first) == sd.catlabel_2_id.end() ){
          pass_away("data-cat-map references unknown data cat label");
        }
      }

      // set trained model parameters from data, such as the starting
      // root distribution, and data category priors:
      //
      init_subs_ml_model_trained_data_dependencies(mdl,sd,is_assigned_cats);

      // set untrained parameters from data (data parameters), such as
      // the background distribution. These will be reset for a new data
      // file associated with any model
      //
      init_subs_ml_model_untrained_data_dependencies(mdl,sd);
    }

    mdl.store_state(main_os);
  }
}




int
main(int argc,char* argv[]){

  // last chance to catch any exceptions that make it down here...
  //
  try{
    try_main(argc,argv);

  } catch (substk_exception& e) {
    std::cerr << "FATAL:: SubsTK EXCEPTION: " << e.what() << "\n";
    throw;

  } catch(std::exception& e) {
    std::cerr << "FATAL:: STD LIBRARY EXCEPTION: " << e.what() << "\n";
    throw;

  } catch(...) {
    std::cerr << "FATAL:: UNKNOWN EXCEPTION\n";
    throw;
  }
}
