// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: bi_tree_type.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __BI_TREE_TYPE_H
#define __BI_TREE_TYPE_H

#include "substk_exception.h"

#include <cassert>

#include <iosfwd>
#include <map>
#include <string>
#include <vector>
#include <utility>

struct bi_tree_node {

  bi_tree_node()
    : parent(0),child1(0),child2(0),label(),length(0.),
      is_length_set(false),
      node_index(-1), leaf_index(-1){}

  void
  clear(){
    if(child1){ child1->clear();}
    if(child2){ child2->clear();}
    delete this;
  }

  void
  print(std::ostream& os) const;

  bi_tree_node* parent;
  bi_tree_node* child1;
  bi_tree_node* child2;
  std::string label;
  double length;
  bool is_length_set;
  short int node_index;
  short int leaf_index;
};


std::ostream& operator<<(std::ostream& os,
                         const bi_tree_node* b);




// copys a bi_tree from the root down
//
bi_tree_node*
copy_node(const bi_tree_node* b,
          bi_tree_node* copy_parent=0);




struct bi_tree_type {
private:
  bi_tree_type& operator=(const bi_tree_type& rhs);

public:
  bi_tree_type(const char* tree_string=0)
    : root(0) { if(tree_string) read_tree(tree_string); }

  bi_tree_type(const bi_tree_type& s){
    root = copy_node(s.root);
  }

  virtual
  ~bi_tree_type() { clear(); }

  virtual
  void
  init(const char* tree_string) { if(tree_string) read_tree(tree_string); }

  void
  read_tree(const char* s);

  void clear(){
    if(root) root->clear();
    root=0;
  }

  bi_tree_node* root;
};


std::ostream& operator<<(std::ostream& os,
                         const bi_tree_type& b);




struct bi_tree_sml_type : public bi_tree_type {

  bi_tree_sml_type(const char* tree_string=0)
    : bi_tree_type(tree_string) { if(tree_string) setup_tree_index(); }

  bi_tree_sml_type(const bi_tree_sml_type& s) : bi_tree_type(s) {
    setup_tree_index();
  }

  virtual
  void
  init(const char* tree_string){
    read_tree(tree_string);
    setup_tree_index();
  }

  void setup_tree_index();

  unsigned n_leaves() const { return leaf.size(); }

  unsigned n_branches() const {
    if(tree.size()) return tree.size()-1;
    else return 0;
  }

  unsigned n_nodes() const { return tree.size(); }

  unsigned get_node_index(const bi_tree_node* b) const {
    assert(b->node_index>=0);
    return b->node_index;
  }

  unsigned get_node_index_from_leaf_index(unsigned n) const {
    return get_node_index(leaf[n]);
  }

  unsigned get_parent_index(unsigned n) const {
    const bi_tree_node* parent(tree[n]->parent);
    if(parent==0) parent=tree[n];
    return get_node_index(parent);
  }

  const std::string& branch_label(unsigned n) const { return tree[n+1]->label; }

  double branch_length(unsigned n) const { return tree[n+1]->length; }

  bool is_branch_length_set(unsigned n) const { return tree[n+1]->is_length_set; }

  unsigned get_branch_index(const std::string& s) const {
    std::map<std::string,unsigned>::const_iterator i(tree_index_map.find(s));
    if(i==tree_index_map.end() || i->second==0) throw substk_exception("invalid branch name");
    return i->second-1;
  }

  unsigned get_branch_index(const bi_tree_node* b) const {
    assert(get_node_index(b)>0);
    return get_node_index(b)-1;
  }

  const std::string& leaf_label(unsigned n) const { return leaf[n]->label; }

  unsigned get_leaf_index(const std::string& s) const {
    std::map<std::string,unsigned>::const_iterator i(leaf_index_map.find(s));
    if(i==leaf_index_map.end()) throw substk_exception("invalid leaf name");
    return i->second;
  }

  unsigned get_leaf_index(const bi_tree_node* b) const {
    assert(b->leaf_index>=0);
    return b->leaf_index;
  }

  void clear_lengths(const bool is_warn = true);

  std::vector<bi_tree_node*> leaf;
  std::vector<bi_tree_node*> tree;
  std::map<std::string,unsigned> leaf_index_map;
  std::map<std::string,unsigned> tree_index_map;
};


#endif
