// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: lhood_model.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "cat_post_prob.h"
#include "lhood_model.h"
#include "lhood_site_cached.h"
#include "site_data_fastlup.h"
#include "site_data_fastlup_core_util.h"
#include "site_data_fastlup_util.h"
#include "subs_ml_model.h"
#include "substk_exception.h"
#include "tree_probs.h"
#include "util/math/matrix_util.h"
#include "util/math/test_float.h"
#include "util/general/die.h"
#include "util/general/log.h"

#include <ostream>

#ifdef INCLUDE_PPIT_MODE
/// ppit-mode stuff:
#include "lhood_site.h"
#include "pprob_cache.h"
#include "pprob_site.h"
#include "util/bio/bioseq_util_pdf_conversions.h"

#ifdef PPIT_UGLY_HACK
#include "rate_gtor_nuc_base.h"
#endif

#endif



/// \brief calc array of site probabilities for all site categories
///
static
void
get_cached_site_prob_array(const bi_tree_sml_type& tree,
                           const tree_probs& tprob,
                           const site_data_fastlup_core& sdf,
                           const int n_states,
                           const condition_func& cf,
                           const std::vector<bool>& site_prob_mask,
                           prob_t* site_prob){

  const tree_lhood_info ti(tree,tprob,n_states);

  lhood_site_cached(ti,sdf,site_prob_mask,site_prob);

  cf.condition_site_prob(tprob,site_prob_mask,site_prob);
}


#ifdef INCLUDE_PPIT_MODE
/// \brief calc array of site probabilities for all site categories
///
static
void
get_site_prob_array(const bi_tree_sml_type& tree,
                    const tree_probs& tprob,
                    const site_data_fastlup_core& sdf,
                    const int n_states,
                    const condition_func&,  // condition func ignored!
                    const std::vector<bool>& site_prob_mask,
                    prob_t** down_prob_tmp,
                    prob_t* site_prob){

  const unsigned n_sites(sdf.len);

  for(unsigned i(0);i<n_sites;++i){
    if(! site_prob_mask[i]){
      site_prob[i] = get_site_prob_single_cat_prep(sdf.data_core[i].index,tree,tprob,
                                                   n_states,down_prob_tmp,i);
    }
  }
}
#endif


static
smlfloat
get_lnprob_from_site_prob(const site_data_fastlup& full_sdf,
                          const prob_t* cat_site_prob,
                          const prob_t* unassigned_cat_site_prob,
                          const int n_site_cats){

  smlfloat lnp(0.);

  if(full_sdf.is_use_unassigned_cats){
    for(unsigned i(0);i<full_sdf.len;++i){
      lnp += unassigned_cat_site_prob[i]*static_cast<smlfloat>(full_sdf.data[i].unassigned_cat_count);
    }
  }
  if(full_sdf.is_use_assigned_cats){
    for(unsigned i(0);i<full_sdf.len;++i){
      for(int sc(0);sc<n_site_cats;++sc){
        lnp += cat_site_prob[i+sc*full_sdf.len]*static_cast<smlfloat>(full_sdf.data[i].assigned_cat_count[sc]);
      }
    }
  }
  return lnp;
}




struct submodel_data {

  submodel_data() : n_states(0), full_sdf_index_map(0) {}

  ~submodel_data() {
    if(full_sdf_index_map) delete [] full_sdf_index_map;
  }

  void
  init(const bi_tree_sml_type& tree,
       const int init_n_states,
       const site_data_fastlup& full_sdf,
       const unsigned* index_translation){

    if(full_sdf_index_map){ die("submodel_data bad init"); }

    n_states = init_n_states;
    full_sdf_index_map = new unsigned[full_sdf.len];

    {
      site_data_fastlup tmp_sdf;
      site_data_fastlup_state_reduction(tree,n_states,full_sdf,index_translation,tmp_sdf);
      sdf = tmp_sdf;
      cat_site_mask = tmp_sdf.cat_site_mask;
    }
    map_translated_index_superset_data(full_sdf,sdf,index_translation,full_sdf_index_map);
  }

  int n_states;
  site_data_fastlup_core sdf;
  std::vector<std::vector<bool> > cat_site_mask;
  unsigned* full_sdf_index_map;
};




struct submodel_manager{

  submodel_manager(const subs_ml_model& mdl,
                   const site_data_fastlup& full_sdf)
    : n_submodels(static_cast<int>(mdl.submodel_size())),
      _submodel(n_submodels),_tprob(n_submodels),
      site_prob(new smlfloat*[n_submodels]),
      _cf(new std::auto_ptr<condition_func>[n_submodels]){

    const int n_states(static_cast<int>(mdl.state_size()));
    simple_array<unsigned> index_translation(n_states+1);

    for(int s(0);s<n_submodels;++s){
      mdl.get_rate_gtor().submodel_state_reduction_map(s,index_translation.val);

      const int submodel_n_states(static_cast<int>(mdl.submodel_state_size(s)));
      submodel(s).init(mdl.tree(),submodel_n_states,full_sdf,index_translation.val);
      site_prob[s] = new smlfloat[submodel(s).sdf.len];
      _cf[s].reset(mdl.get_rate_gtor().submodel_condition_func_factory(s));
      cf(s).data_init(mdl.tree(),submodel(s).sdf);
    }
  }

  condition_func& cf(unsigned i) { return *_cf[i]; }

  submodel_data& submodel(unsigned i) { return _submodel.val[i]; }
  tree_probs& tprob(unsigned i) { return _tprob.val[i]; }

  ~submodel_manager() {
    for(int s(0);s<n_submodels;++s) {
      delete [] site_prob[s];
    }
    delete [] _cf;
    delete [] site_prob;
  }


  int n_submodels;
private:
  simple_array<submodel_data> _submodel;
  simple_array<tree_probs> _tprob;
public:
  smlfloat** site_prob;
private:
  std::auto_ptr<condition_func>* _cf;
};




smlfloat
get_lnp_norm(const subs_ml_model& mdl,
             const site_data_fastlup& sdf){

  const unsigned datasize(sdf.total_count);
  return static_cast<smlfloat>(datasize)*
    std::log(static_cast<smlfloat>(mdl.get_rate_gtor().state_size_conditioned()));
}



nonparam_lnprob_data::
nonparam_lnprob_data(const subs_ml_model& mdl,
                     const site_data_fastlup& sdf)
  : _tprob(new tree_probs), cat_site_prob(0), unassigned_cat_site_prob(0),
    site_cat_prior_pdistro(),tprob_ws(new workspace[3]) {

  const unsigned full_count_len(sdf.len);
  const bool is_use_submodels(mdl.submodel_size()>1);

  const int n_site_cats(static_cast<int>(mdl.get_rate_gtor().category_size_site()));

  const bool is_ppit(mdl.opt().is_ppit);

  if(is_use_submodels){
    _sm.reset(new submodel_manager(mdl,sdf));
  } else {
    _cf.reset(mdl.get_rate_gtor().condition_func_factory());
    cf().data_init(mdl.tree(),sdf);
  }

  cat_site_prob = new prob_t[full_count_len*n_site_cats];
  if(sdf.is_use_unassigned_cats) unassigned_cat_site_prob = new prob_t[full_count_len];

  site_cat_prior_pdistro.resize(n_site_cats);

  if(is_ppit){
    const unsigned n_states(mdl.state_size());
    const unsigned n_branches(mdl.tree().n_branches());
    up_branch_prob.init(n_branches,n_states);
    down_branch_prob.init(n_branches,n_states);
    tmp_state_array.init(n_states);
  }
}


nonparam_lnprob_data::
~nonparam_lnprob_data(){
  delete [] cat_site_prob;
  if(unassigned_cat_site_prob) delete [] unassigned_cat_site_prob;

  delete [] tprob_ws;
}


#ifdef INCLUDE_PPIT_MODE
static
void
get_edge_pprob_average(const std::vector<neighbor_count>& nb,
                       const pprob_cache& edge,
                       const unsigned n_branches,
                       const site_data_fastlup_cell* sdfdata,
                       pprob_cache& average){

  static const std::vector<prob_t> bg_nuc_pdf(NUC::SIZE,(1./static_cast<prob_t>(NUC::SIZE)));

  unsigned baseid_count(0);
  std::vector<prob_t> nuc_pdf(n_branches*NUC::SIZE,0.);

  const unsigned nbs(nb.size());
  for(unsigned j(0);j<nbs;++j){
    const unsigned baseid(nb[j].id1);
    const unsigned id2(nb[j].id2);
    const unsigned count(nb[j].count);

    for(unsigned k(0);k<n_branches;++k){
      for(unsigned n(0);n<NUC::SIZE;++n) nuc_pdf[k*NUC::SIZE+n] += static_cast<prob_t>(count)*edge.val[id2][k][n];
    }

    baseid_count += count;

    if((j+1)==nbs || nb[j+1].id1 != baseid){
      const unsigned baseid_total(sdfdata[baseid].unassigned_cat_count);

      assert(baseid_total>=baseid_count);
      const unsigned bg_count(baseid_total-baseid_count);
      for(unsigned k(0);k<n_branches;++k){
        for(unsigned n(0);n<NUC::SIZE;++n) nuc_pdf[k*NUC::SIZE+n] += static_cast<prob_t>(bg_count)*bg_nuc_pdf[n];
        pdistro_norm(nuc_pdf.begin()+k*NUC::SIZE,nuc_pdf.begin()+(k+1)*NUC::SIZE);
      }

      for(unsigned k(0);k<n_branches;++k){
        for(unsigned n(0);n<NUC::SIZE;++n){
          average.val[baseid][k][n] = nuc_pdf[k*NUC::SIZE+n];
        }
      }

      baseid_count=0;
      std::fill(nuc_pdf.begin(),nuc_pdf.end(),0.);
    }
  }
}



static
void
get_cat_site_prob_ppit(const subs_ml_model& mdl,
                       const site_data_fastlup& full_sdf,
                       nonparam_lnprob_data& nlp,
                       const int sc,
                       const int gc){

  const unsigned full_count_len(full_sdf.len);
  const unsigned n_sites(full_count_len);
  const unsigned n_branches(mdl.tree().n_branches());
  pprob_cache left_average(n_sites,n_branches,NUC::SIZE);
  pprob_cache right_average(n_sites,n_branches,NUC::SIZE);

#ifdef PPIT_ITER_DEBUG
  pprob_cache last_left_average(n_sites,n_branches,NUC::SIZE);
  pprob_cache last_right_average(n_sites,n_branches,NUC::SIZE);
#endif


  static const unsigned ppit_iter(1);

  const unsigned n_states(mdl.state_size());
  const unsigned n_nodes(mdl.tree().n_nodes());

#ifdef PPIT_DEBUG
  const unsigned n_seqs(full_sdf.n_orgs);
#endif

  // for now, abandon state generality in ppit mode unless it's easy:
  /// \tmp
  assert(n_states==NSCODON::SIZE || n_states==NUC::SIZE);

  pprob_cache ppc(n_sites,n_nodes,n_states);

  simple_matrix<prob_t> left_node_edge(n_nodes,NUC::SIZE);
  simple_matrix<prob_t> right_node_edge(n_nodes,NUC::SIZE);

  pprob_cache left_edge(n_sites,n_branches,NUC::SIZE);
  pprob_cache right_edge(n_sites,n_branches,NUC::SIZE);


  for(unsigned i(0);i<ppit_iter;++i){
    if(i==0){
      nlp.tprob().init_probs(mdl,sc,gc,false,0,nlp.tprob_ws);
    } else {
      // left_average and right average must be set from first iteration:
      nlp.tprob().init_site_probs(mdl,sc,gc,false,0,left_average,right_average,n_sites,nlp.tprob_ws);
    }
    for(unsigned s(0);s<n_sites;++s){
      get_site_pprob_single_cat_prep(full_sdf.data_core[s].index,mdl.tree(),nlp.tprob(),n_states,
                                     nlp.up_branch_prob.val,nlp.down_branch_prob.val,
                                     nlp.tmp_state_array.val,
                                     ppc.val[s],s);
    }

#ifdef PPIT_DEBUG
    // dump ppc results:
    for(unsigned k(0);k<n_sites;++k){
      std::cerr << "site state seq: " << k;
      for(unsigned j(0);j<n_seqs;++j){
        std::cerr << " " << full_sdf.data_core[k].index[j];
      }
      std::cerr << "\n";

      for(unsigned b(0);b<n_nodes;++b){
        std::cerr << "GREP_PPC s n: " << k << " " << b;
        for(unsigned n(0);n<4;++n) std::cerr << " "  << ppc.val[k][b][n];
        std::cerr << "\n";
      }
    }
#endif

    // have iteration pprob - now process this into per-site
    // left and right nucleotide contexts -- assume that any
    // missing counts are drawn from the naive background --
    //
    // substep for later -- acquire the naive background called in the rate matrix function from
    // an independent entry point -- for now use simple position based backgrounds

    // 1) go through site list and make left and right nuc distros,
    // 2) go through site list:
    //   - combine neighbor distro's and naive distros

    static const std::vector<prob_t> bg_nuc_pdf(NUC::SIZE,(1./static_cast<prob_t>(NUC::SIZE)));

#ifdef PPIT_UGLY_HACK
    const prob_t* bgn(dynamic_cast<const rate_gtor_nuc_base*>(&mdl.get_rate_gtor())->bg_pdistro_nuc());
    const smlfloat pfrac(0.);
#endif

    prob_t nuc_pdf[NUC::SIZE*CODON::BASE_SIZE];
    for(unsigned j(0);j<n_sites;++j){
      for(unsigned k(0);k<n_nodes;++k){
        if(n_states==NSCODON::SIZE){
          nscodon_pdf_2_nuc_pdf_all_pos(ppc.val[j][k],nuc_pdf);
          for(unsigned n(0);n<NUC::SIZE;++n){
            left_node_edge.val[k][n] = nuc_pdf[n];
            right_node_edge.val[k][n] = nuc_pdf[2*NUC::SIZE+n];
          }
        } else {
          for(unsigned n(0);n<NUC::SIZE;++n){
#ifdef PPIT_UGLY_HACK
            //            left_node_edge.val[k][n] = bgn[n]*pfrac+ppc.val[j][k][n]*(1.-pfrac);
            //            right_node_edge.val[k][n] = bgn[n]*pfrac+ppc.val[j][k][n]*(1.-pfrac);
            left_node_edge.val[k][n] = 0.;
            right_node_edge.val[k][n] = 0.;
            for(unsigned k2(0);k2<n_nodes;k2++){
              left_node_edge.val[k][n] += ppc.val[j][k2][n];
              right_node_edge.val[k][n] += ppc.val[j][k2][n];
            }
            left_node_edge.val[k][n] /= static_cast<smlfloat>(n_nodes);
            right_node_edge.val[k][n] /= static_cast<smlfloat>(n_nodes);
#else
            left_node_edge.val[k][n] = ppc.val[j][k][n];
            right_node_edge.val[k][n] = ppc.val[j][k][n];
#endif
          }
        }
      }

      //    - average nodes-to-branches here too
      for(unsigned k(1);k<n_nodes;++k){ // assume node 0 is parent
        const unsigned kpa(mdl.tree().get_parent_index(k));
        if(kpa==k) die("can't interpret tree");

        for(unsigned n(0);n<NUC::SIZE;++n){
          left_edge.val[j][k-1][n]  = (left_node_edge.val[k][n] + left_node_edge.val[kpa][n])/2.;
          right_edge.val[j][k-1][n] = (right_node_edge.val[k][n] + right_node_edge.val[kpa][n])/2.;
        }
      }

      // initialize averages:
      for(unsigned k(0);k<n_branches;++k){
        for(unsigned n(0);n<NUC::SIZE;++n){
          left_average.val[j][k][n] = bg_nuc_pdf[n];
          right_average.val[j][k][n] = bg_nuc_pdf[n];
        }
      }
    }

    // walk through nb_left/right structure:
    get_edge_pprob_average(full_sdf.nb_left,right_edge,n_branches,full_sdf.data,left_average);
    get_edge_pprob_average(full_sdf.nb_right,left_edge,n_branches,full_sdf.data,right_average);

#ifdef PPIT_DEBUG
    // dump left/right average results:
    for(unsigned k(0);k<n_sites;++k){
      std::cerr << "site state seq: " << k;
      for(unsigned j(0);j<n_seqs;++j){
        std::cerr << " " << full_sdf.data_core[k].index[j];
      }
      std::cerr << "\n";

      for(unsigned b(0);b<n_branches;++b){
        std::cerr << "GREP_LEFT s b: " << k << " " << b;
        for(unsigned n(0);n<4;++n) std::cerr << " "  << left_average.val[k][b][n];
        std::cerr << "\n";
        std::cerr << "GREP_RIGH s b: " << k << " " << b;
        for(unsigned n(0);n<4;++n) std::cerr << " "  << right_average.val[k][b][n];
        std::cerr << "\n";
      }
    }
#endif

#ifdef PPIT_ITER_DEBUG
    if(i>0){
      smlfloat maxleftdiff=0;
      smlfloat maxrightdiff=0;
      for(unsigned j(0);j<n_sites;++j){
        for(unsigned b(0);b<n_branches;++b){
          for(unsigned n(0);n<NUC::SIZE;++n){
            const smlfloat leftdiff=std::abs(last_left_average.val[j][b][n]-left_average.val[j][b][n]);
            const smlfloat rightdiff=std::abs(last_right_average.val[j][b][n]-right_average.val[j][b][n]);
            maxleftdiff=std::max(maxleftdiff,leftdiff);
            maxrightdiff=std::max(maxrightdiff,rightdiff);
          }
        }
      }
      std::cerr << "IT MAXL MAXR: " << i << " " << maxleftdiff << " " << maxrightdiff << "\n";
    }

    for(unsigned j(0);j<n_sites;++j){
      for(unsigned b(0);b<n_branches;++b){
        for(unsigned n(0);n<NUC::SIZE;++n){
          last_left_average.val[j][b][n]=left_average.val[j][b][n];
          last_right_average.val[j][b][n]=right_average.val[j][b][n];
        }
      }
    }
#endif

#if 0
    // peek at left edge distros, to check for differentiation by site
    for(unsigned j(0);j<std::min(n_sites,10);++j){
      std::cerr << "GREP_ITER" << i << "_LEDGE" << j;
      for(unsigned n(0);n<NUC::SIZE;++n)
        std::cerr << " " << left_average.val[j][0][n];
      std::cerr << "\n";
    }
#endif
  }

#if 0
  // 0 iterations= a regular lhood calculation, just much, much
  // slower, this loop is here to fill in naive nuc values only
  // for sanity checking...  against the regular lhood
  // procedure:
  //
  if(ppit_iter==0){
    nlp.tprob().init_probs(mdl,sc,gc,0,nlp.tprob_ws);
    //          const prob_t* bgn_left(dynamic_cast<const rate_gtor_nscodon_base*>(&mdl.get_rate_gtor())->bg_pdistro_nuc_pos(2));
    //          const prob_t* bgn_right(dynamic_cast<const rate_gtor_nscodon_base*>(&mdl.get_rate_gtor())->bg_pdistro_nuc_pos(0));
    for(unsigned i(0);i<n_sites;++i){
      for(unsigned j(0);j<n_branches;++j){
        for(unsigned k(0);k<NUC::SIZE;++k){
          left_average.val[i][j][k] = bgn_left[k];
          right_average.val[i][j][k] = bgn_right[k];
        }
      }
    }
  }
#endif

  // final update of site dependent transition probs for lhood calculation
  nlp.tprob().init_site_probs(mdl,sc,gc,false,0,left_average,right_average,n_sites,nlp.tprob_ws);

  // can't use site caching to get lhood when the rate matrix changes for each site!!
  get_site_prob_array(mdl.tree(),nlp.tprob(),full_sdf,
                      mdl.state_size(),nlp.cf(),full_sdf.get_cat_site_mask(sc),
                      nlp.down_branch_prob.val,
                      nlp.cat_site_prob+sc*full_count_len);
}
#endif // INCLUDE_PPIT_MODE



// all tprobs transposed!
void
get_lnprob_from_param_prep(const subs_ml_model& mdl,
                           const site_data_fastlup& full_sdf,
                           nonparam_lnprob_data& nlp,
                           smlfloat& lnp,
                           smlfloat& lnp_norm,
                           const bool is_cat_post_prob_mode){

  const unsigned full_count_len(full_sdf.len);
  const bool is_use_submodels(mdl.submodel_size()>1);

  if(is_use_submodels && mdl.opt().is_ppit){
    die("Cannot use submodel approximations in ppit mode");
  }

  const int n_site_cats(static_cast<int>(mdl.get_rate_gtor().category_size_site()));

  // get category distros
  mdl.get_rate_gtor().get_site_category_pdistro(nlp.site_cat_prior_pdistro.begin());

  std::fill(nlp.cat_site_prob,nlp.cat_site_prob+full_count_len*n_site_cats,1.);
  for(int sc(0);sc<n_site_cats;++sc){

    if(! is_use_submodels) {
      if(! mdl.opt().is_ppit){
        nlp.tprob().init_probs(mdl,sc,0,is_use_submodels,0,nlp.tprob_ws);
        get_cached_site_prob_array(mdl.tree(),nlp.tprob(),full_sdf,
                                   mdl.state_size(),nlp.cf(),full_sdf.get_cat_site_mask(sc),
                                   nlp.cat_site_prob+sc*full_count_len);
      } else {
#ifdef INCLUDE_PPIT_MODE
        get_cat_site_prob_ppit(mdl,full_sdf,nlp,sc,0);
#else
        pass_away("No ppit mode in this build");
#endif
      }
    } else {

      for(int s(0);s<nlp.sm().n_submodels;++s){
        std::fill(nlp.sm().site_prob[s],nlp.sm().site_prob[s]+nlp.sm().submodel(s).sdf.len,0.);
        nlp.sm().tprob(s).init_probs(mdl,sc,0,is_use_submodels,s,nlp.tprob_ws);

        const submodel_data& sd(nlp.sm().submodel(s));
        get_cached_site_prob_array(mdl.tree(),nlp.sm().tprob(s),sd.sdf,
                                   sd.n_states,nlp.sm().cf(s),sd.cat_site_mask[sc],
                                   nlp.sm().site_prob[s]);

        for(unsigned i(0);i<full_count_len;++i){
          const int s_index(sd.full_sdf_index_map[i]);
          nlp.cat_site_prob[i+sc*full_count_len] *= nlp.sm().site_prob[s][s_index];
        }
      }
    }
  }
  mdl.get_rate_gtor().submodel_adjust_p(nlp.cat_site_prob,full_count_len*n_site_cats);

  if(is_cat_post_prob_mode){ cat_post_prob_report(mdl.rate_gtor_model(),mdl.tree(),n_site_cats,full_sdf,nlp.cat_site_prob,nlp.site_cat_prior_pdistro); }

  if(full_sdf.is_use_unassigned_cats){
    for(unsigned i(0);i<full_count_len;++i){
      nlp.unassigned_cat_site_prob[i] = 0.;
      for(int sc(0);sc<n_site_cats;++sc){
        nlp.unassigned_cat_site_prob[i] += nlp.cat_site_prob[i+sc*full_count_len]*nlp.site_cat_prior_pdistro[sc];
      }
    }
    array_log(nlp.unassigned_cat_site_prob,full_sdf.len);
  }
  if(full_sdf.is_use_assigned_cats){
    array_log(nlp.cat_site_prob,full_count_len*n_site_cats);
  }

  lnp = get_lnprob_from_site_prob(full_sdf,nlp.cat_site_prob,nlp.unassigned_cat_site_prob,n_site_cats);


  const smlfloat norm(get_lnp_norm(mdl,full_sdf));

  // apply penalties for out-of-range parameters
  const smlfloat pen(mdl.param_penalty());
  if(pen>0.){
    log_os << "WARNING:: Applying parameter out-of-range penalty: " << pen << "\n";
    lnp -= pen*norm;
  }

  lnp_norm = lnp/norm;

  if( is_float_nan(lnp_norm)){
    log_os << "EXCEPTION:: Invalid lnp_norm: " << lnp_norm << "\n";
    log_os << "Dumping model:\n";
    mdl.store_state(log_os);
#ifdef DEBUG
    die("Invalid lnp_norm");
#else
    throw substk_exception("Invalid lnp_norm");
#endif
  } else if( is_float_inf(lnp_norm)){
    log_os << "WARNING:: get_lnprob returning lnp_norm: " << lnp_norm << "\n";
    log_os << "          attempting to proceed...\n";
    mdl.store_state(log_os);
  } else if( lnp > 0. ){
    log_os << "VERY FATAL:: Positive lnp!!!: " << lnp << "\n";
    log_os << "Dumping model:\n";
    mdl.store_state(log_os);
    abort();
  }
}




void
get_lnprob_from_param(const subs_ml_model& mdl,
                      const site_data_fastlup& sdf,
                      smlfloat& lnp,
                      smlfloat& lnp_norm,
                      const bool is_cat_post_prob_mode){

  // parameter independent setup:
  nonparam_lnprob_data nlp(mdl,sdf);

  get_lnprob_from_param_prep(mdl,sdf,nlp,lnp,lnp_norm,is_cat_post_prob_mode);
}
