// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: pprob_site.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "pprob_site.h"
#include "substk_exception.h"
#include "util/math/matrix_util.h"
#include "util/math/prob_util.h"

#ifdef XTRA_DEBUG
#include "lhood_site.h"
#include "util/general/log.h"
#include <ostream>
#endif



static
void
tree_branch_down_prob(const bi_tree_node* b,
                      const bi_tree_sml_type& tree,
                      const tree_probs& tprob,
                      const unsigned n_states,
                      const unsigned site_id,
                      const prob_t* const * up_branch_prob,
                      const prob_t* root_prob,
                      prob_t** down_branch_prob,
                      prob_t* const tmp_state_array){

  if(b==0 || b->parent==0){
    throw substk_exception("x_up");

  } else if((b->child1==0 && b->child2!=0) ||
            (b->child2==0 && b->child1!=0)){
    throw substk_exception("Invalid tree graph");

  } else {
    const int branch_id(tree.get_node_index(b)-1);
    prob_t* branch_p(down_branch_prob[branch_id]);

    const bi_tree_node* parent(b->parent);
    const bi_tree_node* sister(parent->child2==b ? parent->child1 : parent->child2);

    const int parent_id(tree.get_node_index(parent)-1);
    const int sister_id(tree.get_node_index(sister)-1);

    const prob_t* parent_p;
    if(parent->parent==0){ // use root as parent down branch
      parent_p=root_prob;
    } else {
      parent_p=down_branch_prob[parent_id];
    }

    for(unsigned i_node(0);i_node<n_states;++i_node){
      const prob_t* const tprob_vec(tprob.branch_site_prob(branch_id,site_id)+i_node*n_states);

      prob_t* parent_node_combined_p(tmp_state_array);

      for(unsigned i_parent(0);i_parent<n_states;++i_parent){
        parent_node_combined_p[i_parent] = up_branch_prob[sister_id][i_parent]*parent_p[i_parent];
      }

      // dot product sums over parent node states...
      branch_p[i_node] = array_dot(tprob_vec,parent_node_combined_p,n_states);
    }

    if(b->child1!=0 && b->child2!=0){
      tree_branch_down_prob(b->child1,tree,tprob,n_states,site_id,up_branch_prob,
                            root_prob,down_branch_prob,tmp_state_array);
      tree_branch_down_prob(b->child2,tree,tprob,n_states,site_id,up_branch_prob,
                            root_prob,down_branch_prob,tmp_state_array);
    }
  }
}




// probability of subtree rooted at node b, for all NSTATE values
// at node b:
//
static
void
tree_branch_up_prob(const bi_tree_node* b,
                    const bi_tree_sml_type& tree,
                    const tree_probs& tprob,
                    const site_code::index_type* seq_state,
                    const unsigned n_states,
                    const unsigned site_id,
                    prob_t** up_branch_prob){

  const int branch_id(tree.get_node_index(b)-1);
  prob_t* branch_p(up_branch_prob[branch_id]);

  if(b==0 || b->parent==0){
    throw substk_exception("x_down");

  } else if(b->child1==0 && b->child2==0){  // leaf node
    const unsigned seq_val(seq_state[tree.get_leaf_index(b)]);
    for(unsigned i(0);i<n_states;++i){
      branch_p[i] = tprob.branch_site_prob(branch_id,site_id)[seq_val*n_states+i];
    }

  } else  if(b->child1==0 || b->child2==0){
    throw substk_exception("Invalid tree graph");

  } else {  // assume child2 exists
    const int child1_id(tree.get_node_index(b->child1)-1);
    const int child2_id(tree.get_node_index(b->child2)-1);

    tree_branch_up_prob(b->child1,tree,tprob,seq_state,n_states,site_id,up_branch_prob);
    tree_branch_up_prob(b->child2,tree,tprob,seq_state,n_states,site_id,up_branch_prob);

    std::fill(branch_p,branch_p+n_states,0.);

    // loop over node state:
    for(unsigned i_node(0);i_node<n_states;++i_node){
      const prob_t node_down_combined_p(up_branch_prob[child1_id][i_node]*
                                        up_branch_prob[child2_id][i_node]);
      const prob_t* const tprob_vec(tprob.branch_site_prob(branch_id,site_id)+i_node*n_states);
      // loop over state of parent node:
      for(unsigned i_parent(0);i_parent<n_states;++i_parent){
        branch_p[i_parent] += tprob_vec[i_parent]*node_down_combined_p;
      }
    }
  }
}




void
get_site_pprob_from_branch_prob(const bi_tree_sml_type& tree,
                                const unsigned n_states,
                                const prob_t* root_prob,
                                const prob_t* const * up_branch_prob,
                                const prob_t* const * down_branch_prob,
                                const site_code::index_type* seq_state,
                                prob_t** pprob){

#ifdef DEBUG
    const unsigned n_branches(tree.n_branches());
    for(unsigned j(0);j<n_branches;++j){
      for(unsigned i(0);i<n_states;++i){
        assert(up_branch_prob[j][i] >= 0.);
        assert(down_branch_prob[j][i] >= 0.);
      }
    }
#endif

  const unsigned n_nodes(tree.n_nodes());

  for(unsigned j(0);j<n_nodes;++j){
    const bi_tree_node* jn(tree.tree[j]);
    prob_t* pn(pprob[j]);

    if(jn->parent == 0){ // root node:
      for(unsigned i(0);i<n_states;++i){
        pn[i] =
          root_prob[i]*
          up_branch_prob[jn->child1->node_index-1][i]*
          up_branch_prob[jn->child2->node_index-1][i];
      }
#ifdef XTRA_DEBUG
      prob_t sum(0.);
      for(unsigned i(0);i<n_states;++i) sum += pn[i];
      log_os << "node pp(r): " << j << " " << sum << "\n";
#endif
    } else if(jn->child1 == 0){ // leaf node:
      std::fill(pn,pn+n_states,0.);
      pn[seq_state[jn->leaf_index]] = 1.;

    } else {
      for(unsigned i(0);i<n_states;++i){
        pn[i] =
          down_branch_prob[jn->node_index-1][i]*
          up_branch_prob[jn->child1->node_index-1][i]*
          up_branch_prob[jn->child2->node_index-1][i];
      }
#ifdef XTRA_DEBUG
      prob_t sum(0.);
      for(unsigned i(0);i<n_states;++i) sum += pn[i];
      log_os << "node ppsum: " << j << " " << sum << "\n";
#endif
    }

    pdistro_norm(pn,pn+n_states);
  }
}




void
get_site_pprob_single_cat_prep(const site_code::index_type* seq_state,
                               const bi_tree_sml_type& t,
                               const tree_probs& tprob,
                               const unsigned n_states,
                               prob_t** up_branch_prob, // n_branches*n_states
                               prob_t** down_branch_prob, // n_branches*n_states
                               prob_t* tmp_state_array,  // n_states
                               prob_t** pprob, // n_nodes*n_states
                               const unsigned site_id){

  // temporary structures should match what's found in twe in the cached version of this routine:
  // basically just need an up and down prob distro for every branch

#ifdef DEBUG
  // fill with bogus vals to check against branch skip...
  const unsigned n_branches(t.n_branches());
  const unsigned vi(n_branches*n_states);
  std::fill(up_branch_prob[0],up_branch_prob[0]+vi,-1);
  std::fill(up_branch_prob[0],up_branch_prob[0]+vi,-1);
#endif

  tree_branch_up_prob(t.root->child1,t,tprob,seq_state,n_states,site_id,up_branch_prob);
  tree_branch_up_prob(t.root->child2,t,tprob,seq_state,n_states,site_id,up_branch_prob);

  tree_branch_down_prob(t.root->child1,t,tprob,n_states,site_id,up_branch_prob,
                        tprob.root_prob(),down_branch_prob,tmp_state_array);
  tree_branch_down_prob(t.root->child2,t,tprob,n_states,site_id,up_branch_prob,
                        tprob.root_prob(),down_branch_prob,tmp_state_array);

  // share pprob calculation step from cached_site version....
  get_site_pprob_from_branch_prob(t,n_states,tprob.root_prob(),
                                  up_branch_prob,down_branch_prob,
                                  seq_state,pprob);

#ifdef XTRA_DEBUG
  const prob_t lhood(get_site_prob_single_cat_prep(seq_state,t,tprob,n_states,down_branch_prob,site_id));
  log_os << "site/lhood: " << site_id << " " << lhood << "\n";
#endif
}




void
get_site_pprob_single_cat(const site_code::index_type* seq_state,
                          const bi_tree_sml_type& t,
                          const tree_probs& tprob,
                          const unsigned n_states,
                          prob_t** pprob, // n_nodes*n_states
                          const unsigned site_id){

  // temporary structures should match what's found in twe in the cached version of this routine:
  const unsigned n_branches(t.n_branches());
  simple_matrix<prob_t> up_branch_prob(n_branches,n_states);
  simple_matrix<prob_t> down_branch_prob(n_branches,n_states);
  simple_array<prob_t> tmp_state_array(n_states);

  get_site_pprob_single_cat_prep(seq_state,t,tprob,n_states,
                                 up_branch_prob.val,down_branch_prob.val,tmp_state_array.val,
                                 pprob,site_id);
}
