// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: prob_gtor.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __PROB_GTOR_H
#define __PROB_GTOR_H

#include "param_util.h"
#include "param_view.h"
#include "subs_ml_model_comex.h"
#include "util/math/prob_util.h"
#include "util/general/die.h"

//#include <iostream>
#include <vector>

/// \brief probability generator base class
///
/// this was intended to be a generally parameterizable prob distro
/// class, it's become a bit specialized as the (category) model root
/// distro
///
template <typename FloatType>
struct prob_gtor {

  typedef prob_gtor<FloatType> self_t;

  prob_gtor(const unsigned init_state_size = 0,
            const unsigned init_param_size = 0)
    : _distro(0), _param(0), _is_train_param(init_param_size,true),
      _is_free_block_start(init_param_size,false),
      _is_free_block_stop(init_param_size,false),
      is_valid(false),
      _param_variance() {
    init(init_state_size,init_param_size);
    if(_is_free_block_start.size()) _is_free_block_start.front() = true;
    if(_is_free_block_stop.size()) _is_free_block_stop.back() = true;
  }

  prob_gtor(const self_t& s)
    : _distro(0), _param(0), _is_train_param(s._is_train_param),
      _is_free_block_start(s._is_free_block_start), _is_free_block_stop(s._is_free_block_stop),
      is_valid(s.is_valid),
      _param_variance() {
    init(s._state_size,s._param_size);
    for(unsigned i(0);i<state_size();++i) _distro[i] = s._distro[i];
    for(unsigned i(0);i<param_size();++i) _param[i] = s._param[i];
  }

  virtual ~prob_gtor(){ clear(); }

private:
  self_t& operator=(const self_t& rhs); // disable assignment copy
public:

  virtual self_t* clone(const root_gtor_sml_share&) const = 0;

  const FloatType* distro() const {
    if(! is_valid) die("prob_gtor not set");
    return _distro;
  }

  unsigned state_size() const { return _state_size; }

  unsigned param_size(const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const {
    if       (pv == PARAM_VIEW::ALL){
      return _param_size;
    } else if(pv == PARAM_VIEW::MIN_PARAM ||
              pv == PARAM_VIEW::TRAINABLE){
      unsigned pst(0);
      const unsigned ps(param_size());
      for(unsigned i(0);i<ps;++i) if(_is_train_param[i]) pst++;
      return pst;
    } else if(pv == PARAM_VIEW::INDY_MIN_PARAM){
      unsigned pst(0);
      bool is_in_block(false);
      bool is_first_in_block(false);
      const unsigned ps(param_size());
      for(unsigned i(0);i<ps;++i) {
        if(_is_free_block_start[i]) {
          is_in_block=true;
          is_first_in_block=true;
        }

        if(_is_train_param[i]){
          if(is_in_block && is_first_in_block){
            is_first_in_block=false;
          } else { pst++; }
        }

        if(_is_free_block_stop[i]) is_in_block=false;
      }
      return pst;
    } else { die("prob_gtor: unknown param view"); }
  }

  void
  lock_root() {
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) _is_train_param[i] = false;
  }

  template <typename OutputIterator>
  OutputIterator
  param_state(OutputIterator x,
              const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const {
    if(! is_valid) die("prob_gtor not set");
    const unsigned ps(param_size());
    if       (pv == PARAM_VIEW::ALL){
      for(unsigned i(0);i<ps;++i) {*x =_param[i]; ++x;}

    } else if(pv == PARAM_VIEW::MIN_PARAM ||
              pv == PARAM_VIEW::TRAINABLE){
      for(unsigned i(0);i<ps;++i) {
        if(_is_train_param[i]) { *x =_param[i]; ++x; }
      }

    } else if(pv == PARAM_VIEW::INDY_MIN_PARAM){
      bool is_in_block(false);
      for(unsigned i(0);i<ps;++i){
        if(_is_free_block_start[i]) is_in_block=true;

        if(_is_train_param[i] && ! is_in_block){
          *x =  _param[i];;
          ++x;
        }

        if(_is_free_block_stop[i]) {
          is_in_block=false;

          x=dependent_to_independent_param_block(_is_train_param,
                                                 _is_free_block_start,
                                                 i,
                                                 _param,
                                                 x);
        }
      }

    } else { die("prob_gtor: unknown param view"); }

    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_param_state(InputIterator x,
                  const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL,
                  const bool is_secondary_update = true) {
    const unsigned ps(param_size());

    if       (pv == PARAM_VIEW::ALL){
      is_valid=true;
      for(unsigned i(0);i<ps;++i) {
        _param[i]=*x;
        if(_param[i] < 0.) _param[i] *= -1.;
        ++x;
      }

    } else if(pv == PARAM_VIEW::MIN_PARAM ||
              pv == PARAM_VIEW::TRAINABLE){
      if(! is_valid) die("prob_gtor not set");
      for(unsigned i(0);i<ps;++i){
        if(_is_train_param[i]) {
          _param[i]=*x;
          if(_param[i] < 0.) _param[i] *= -1.;
          ++x;
        }
      }

    } else if(pv == PARAM_VIEW::INDY_MIN_PARAM){
      if(! is_valid) die("prob_gtor not set");

      x=set_indy_min_param(_is_train_param,
                           _is_free_block_start,
                           _is_free_block_stop,
                           ps,
                           _param,
                           x);

    } else { die("prob_gtor: unknown param view"); }

    param_update(is_secondary_update);
    return x;
  }

  template <typename OutputIterator>
  OutputIterator
  is_train_param(OutputIterator x) const {
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) {
      *x = _is_train_param[i];
      ++x;
    }
    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_is_train_param(InputIterator x) {
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) {
      _is_train_param[i] = *x;
      ++x;
    }
    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_distro(InputIterator x) {
    is_valid=true;
    const unsigned ss(state_size());
    for(unsigned i(0);i<ss;++i) { _distro[i]=*x; ++x;}
    distro_update();
    return x;
  }

  template <typename InputIterator>
  InputIterator
  attach_ci(InputIterator ci){
    const unsigned ps(param_size());
    _param_variance.clear();
    _param_variance.resize(ps,0.);

    for(unsigned i(0);i<ps;++i){
      if(_is_train_param[i]){
        _param_variance[i] = ci->variance;
        ci++;
      }
    }
    return ci;
  }

  FloatType
  param_variance(const unsigned i) const {
    if(_param_variance.size()>i){
      return _param_variance[i];
    } else {
      return FloatType(0);
    }
  }

protected:
  void
  distro_update() {
    // recycle distro into distro set from parameters:
    get_param_from_distro();
    param_update();
  }

  virtual
  void get_param_from_distro() const = 0;

  virtual
  void param_update(const bool is_secondary_update = true)  = 0;

private:
  void
  clear() {
    if(_distro) {
      delete [] _distro;
      _distro = 0;
    }
    if(_param) {
      delete [] _param;
      _param = 0;
    }
  }

  void
  init(const unsigned init_state_size,
       const unsigned init_param_size){
    clear();
    _state_size = init_state_size;
    _param_size = init_param_size;
    if(state_size()) _distro = new FloatType[state_size()];
    if(param_size()) _param = new FloatType[param_size()];
  }

  unsigned _state_size;
  unsigned _param_size;
protected:
  FloatType* _distro;
  FloatType* _param;
  std::vector<bool> _is_train_param;
  std::vector<bool> _is_free_block_start;
  std::vector<bool> _is_free_block_stop;
private:
  bool is_valid;
  std::vector<FloatType> _param_variance;
};


#endif
