// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: process_fsa.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "process_fsa.h"
#include "simple_util.h"
#include "util/bio/fasta_seq.h"
#include "util/bio/seq_util.h"
#include "util/general/die.h"
#include "util/general/io_util.h"

#include <sys/types.h>
#include <dirent.h>

#include <cassert>
#include <cctype>
#include <cstdlib>

#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

using namespace std;


const unsigned UNASSIGNED_CAT_ID(0);
const char UNASSIGNED_CAT_LABEL[] = "unassigned";

const unsigned CODON_BASE_SIZE(3);

typedef std::vector<simple_init_array<char> > tmp_seq_type;


// convert input sequence to align_dat_t format, no
// questions asked:
//
static
void
no_filter(align_dat_t& filter_align,
          const align_dat_t::info_type& info,
          const tmp_seq_type& seq){

  const unsigned n_seqs(seq.size());
  if(n_seqs==0) return;

  const unsigned n_bases(seq[0].size());

  filter_align.info.resize(n_bases);
  filter_align.seq.init(n_seqs,n_bases);

  for(unsigned i(0);i<n_bases;++i){
    filter_align.info[i] = info[i];
    for(unsigned s(0);s<n_seqs;++s){
      filter_align.seq.val[s][i] = NUC::remap(seq[s].val[i]);
    }
  }
}



// check for 'check-val' in any seq at codon site::
static
bool
check_codon_site(const tmp_seq_type& seq,
                 const unsigned codonno,
                 const char check_val){

  const unsigned baseno(codonno*CODON_BASE_SIZE);
  const unsigned n_seqs(seq.size());

  for(unsigned t(0);t<n_seqs;++t){
    const char aa_site(codon_trans_known(seq[t].val+baseno));
    if (aa_site == check_val) { return true; }
  }
  return false;
}


// check for stop in current position:
static
bool
check_stop_codon(const tmp_seq_type& seq,
                 const unsigned codonno){

  return check_codon_site(seq,codonno,'*');
}


// check for gap/ambiguity/stop in current position::
static
bool
check_gap_codon(const tmp_seq_type& seq,
                const unsigned codonno){

  return check_codon_site(seq,codonno,'-');
}


static
bool
check_ambi_codon(const tmp_seq_type& seq,
                 const unsigned codonno){

  return check_codon_site(seq,codonno,'X');
}




// assume this is aligned coding sequence, skip gaps and codons with
// ambiguities, previous and next codons must be clean as well,
// gapless filter currently applies to the entire column only, which
// only works as long as we allow no missing sequences/codons
//
static
void
codon_filter(align_dat_t& filter_align,
             const align_dat_t::info_type& info,
             const tmp_seq_type& seq,
             const seq_filter_options& fopt){

  static const unsigned default_seqwindow(10);
  static const unsigned min_gapless(20);

  const unsigned n_seqs(seq.size());
  if(n_seqs==0) return;

  const unsigned n_bases(seq[0].size());

  const unsigned seqwindow((fopt.is_use_window_filter?default_seqwindow:0));

  if( n_bases%CODON_BASE_SIZE ) {
    die("process_alignment codon_filter: fsa sequences not divisible to codons");
  }

  const unsigned n_codons(n_bases/CODON_BASE_SIZE);

  std::vector<bool> pass_list(n_codons,false);

  // pre-calc seq aa's
  simple_matrix<char> aa(n_seqs,n_codons);
  for(unsigned t(0);t<n_seqs;++t){
    for(unsigned codonno=0;codonno<n_codons;codonno++){
      const unsigned baseno(codonno*CODON_BASE_SIZE);
      aa.val[t][codonno] = codon_trans(seq[t].val+baseno);
    }
  }

  // go thru entire sequence, bounded by window-ends...
  //
  for(unsigned codonno(seqwindow); (codonno+seqwindow) < n_codons; ++codonno){

    if(check_stop_codon(seq,codonno)) continue;
    if(fopt.is_use_gap_filter && check_gap_codon(seq,codonno)) continue;
    if(fopt.is_use_ambi_filter && check_ambi_codon(seq,codonno)) continue;

    // run windowed quality filter:
    if(fopt.is_use_window_filter){

      unsigned n_gapless(0);
      unsigned n_match(0);

      for(int i=-static_cast<int>(seqwindow);i<(static_cast<int>(seqwindow)+1);++i){
        if (i==0) continue;

        const unsigned window_codonno(codonno+i);

        bool is_gap(false);
        for(unsigned t(0);t<n_seqs;++t){
          if(aa.val[t][window_codonno] == '-'){
            is_gap=true;
            break;
          }
        }

        if(! is_gap) n_gapless++;

        bool is_match(true);
        char maa(0);
        for(unsigned t(0);t<n_seqs;++t){
          if(t==0){
            maa = aa.val[t][window_codonno];
          } else {
            if( maa != aa.val[t][window_codonno] ){
              is_match=false;
              break;
            }
          }
        }

        if(is_match) n_match++;
      }

      // filter from window results
      if(n_gapless < min_gapless) continue;
      if(n_match < fopt.min_match) continue;
    }

    // made it! record codon:
    pass_list[codonno] = true;
  }

  unsigned n_pass_bases(0);
  for(unsigned i(0);i<n_codons;++i) if(pass_list[i]) n_pass_bases+=3;

  filter_align.info.resize(n_pass_bases);
  filter_align.seq.init(n_seqs,n_pass_bases);

  bool is_last_accepted_codonno_set(false);
  unsigned last_accepted_codonno(0);
  for(unsigned i(0),pi(0);i<n_codons;++i){
    const bool is_continuous(is_last_accepted_codonno_set && (i == last_accepted_codonno+1));

    if(pass_list[i]){
      for(unsigned j(0);j<CODON_BASE_SIZE;++j){
        const unsigned base(i*CODON_BASE_SIZE+j);
        const unsigned pbase(pi*CODON_BASE_SIZE+j);
        filter_align.info[pbase] = info[base];
        filter_align.info[pbase].is_continuous = (j!=0 || (filter_align.info[pbase].is_continuous && is_continuous));
        for(unsigned s(0);s<n_seqs;++s){
          filter_align.seq.val[s][pbase] = NUC::remap(seq[s].val[base]);
        }
      }
      pi++;

      is_last_accepted_codonno_set = true;
      last_accepted_codonno = i;
    }
  }
}



static
void
harvest_fsa_file(align_dat_t::info_type& info,
                 tmp_seq_type& seq,
                 name_id_lup& taxid_lup,
                 const char* fsa_file){

  // just assume it's the file for now
  ifstream infp(fsa_file);
  check_nonempty_istream(infp,fsa_file);

  info.clear();
  seq.clear();

#if 0
  // faster code to replace FastaSeq functions
  //
  int c=infp.peek();
  if(c!='>') die("can't parse fasta");

  // read seq label from the 3rd column:
  string seq_label;
  infp >> seq_label >> seq_label >> seq_label;

  // kill ws:
  do{
    c=infp.get();
  } while(c==' ' || c=='\t' || c== '\n' || ! infp);

  infp.unget(c);

  tseq.resize(0);

  while(true){
    c=infp.get();
    if(! infp) break;
    if(c==' ' || c=='\t' || c== '\n') continue;
    if(c=='>') {
      infp.unget(c);
      break;
    }
    tseq.push_back(c);
  }

  tseq.push_back
#endif

  FastaSeq f;

  unsigned seqcount(0);

  while( infp >> f ) {
    seqcount++;

    istringstream is(f.header);

    // throw away the '>' char on header
    assert(is.peek() == '>');
    is.get();

    // first ws separated word in header is used as org name
    string taxstr;
    is >> taxstr;

    const unsigned taxid(taxid_lup.assignid(taxstr));
    if(taxid>=seq.size()){ seq.resize(taxid+1); }
    seq[taxid].init(f.seq.size());
    std::copy(f.seq.begin(),f.seq.end(),seq[taxid].val);
  }

  const unsigned n_seqs(seq.size());

  // check that all seq are the same size:
  for(unsigned t(0),n_bases(0);t<n_seqs;++t){
    const unsigned b(seq[t].size());
    if(n_bases==0) n_bases = b;
    if(b==0){
      // sequence absent: non-error skip case
      seq.clear();
      return;
    } else if(n_bases != b){
      die("process_fsa:: invalid alignment format");
    }
  }

  if(n_seqs){
    const unsigned n_bases(seq[0].size());

    // add default info
    info.resize(n_bases);
    for(unsigned n(0);n<n_bases;++n){
      info[n].cat_no = UNASSIGNED_CAT_ID;
      info[n].codon_pos = static_cast<unsigned char>(n%CODON_BASE_SIZE);
      info[n].is_continuous = (n!=0);
    }

    // upper-caseify everything:
    for(unsigned t(0);t<n_seqs;++t){
      char* v(seq[t].val);
      for(unsigned n(0);n<n_bases;++n){ v[n] = toupper(v[n]); }
    }
  }
}




// assumes in_align is sorted on base_no!!
//
// parsing is lazy sstream stuff, should not scale well.
//
// assumes input categories are numbered by codon (zero-indexed)
//
//format:
// first line: [codon_no|*] cat_no
// other line: codon_no cat_no
//
// * sets the default cat_no for all valid codons in the sequence
// else, default cat_no=0
//
static
void
harvest_cat_file(align_dat_t::info_type& info,
                 const char* cat_file){

  // cat file is allowed to be empty or not exist to indicate defualt cat:
  ifstream cat_is(cat_file);
  if( ! cat_is ) return;

  string buf;
  bool is_read_first_noncomment(false);

  unsigned default_cat_no(UNASSIGNED_CAT_ID);

  const unsigned asize(info.size());

  while(getline(cat_is,buf)){
    if( buf[0] == '#') continue;

    istringstream sbuf(buf);

    unsigned codon_no;
    unsigned cat_no;

    if( ! is_read_first_noncomment ){
      is_read_first_noncomment=true;
      if(buf[0] == '*'){
        sbuf.get();
        sbuf >> cat_no;
        if(!sbuf) die((string("irregular cat file format (1): ")+cat_file).c_str());

        if(cat_no == default_cat_no) continue;

        for(unsigned i(0);i<asize;++i){
          info[i].cat_no = static_cast<align_col_t::cat_type>(cat_no);
        }

        // after setting the default, individual codon cats can be
        // specified, this lets downstream know that a new default has
        // been set for this alignment
        default_cat_no = cat_no;
        continue;
      }
    }

    sbuf >> codon_no >> cat_no;

    if(!sbuf) die((string("irregular cat file format (2): ")+cat_file).c_str());

    // skip default cat value:
    if(cat_no == default_cat_no) continue;

    const unsigned base_no(codon_no*CODON_BASE_SIZE);
    for(unsigned j(0);j<CODON_BASE_SIZE;++j) {
      info[base_no+j].cat_no = static_cast<align_col_t::cat_type>(cat_no);
    }
  }
}



/// catlabel format:
/// # comment anyline
/// 1 label1
/// 2 label2
/// ...
///
static
void
harvest_catlabel(map<unsigned,string>& id_2_catlabel,
                 const char* catlabel_file){

  string buf;
  ifstream cat_is(catlabel_file);
  check_nonempty_istream(cat_is,catlabel_file);

  while(getline(cat_is,buf)){
    if( buf[0] == '#' ) continue;
    istringstream sbuf(buf);

    unsigned cat_no;
    string cat_label;
    sbuf >> cat_no >> cat_label;

    if(cat_label.size() == 0) pass_away("empty label in catlabel file");

    if(id_2_catlabel.find(cat_no) != id_2_catlabel.end()){
      if(id_2_catlabel[cat_no] != cat_label &&
         ! (cat_no == UNASSIGNED_CAT_ID && id_2_catlabel[cat_no] == UNASSIGNED_CAT_LABEL)){
        std::cerr << "ERROR:: harvest_catlabel: cat_no: " << cat_no
                  << " is already assigned to label: " << id_2_catlabel[cat_no] << "\n";
        exit(EXIT_FAILURE);
      }
    }
    id_2_catlabel[cat_no] = cat_label;
  }
}




static
void
harvest_file_set(nuc_seq_data& in_seq,
                 align_dat_t::info_type& tmp_info,
                 tmp_seq_type& tmp_seq,
                 const char* full_fsa_file,
                 const bool is_cat_mode,
                 const char* full_cat_file,
                 const seq_filter_options& fopt,
                 const std::string& label){

  harvest_fsa_file(tmp_info,tmp_seq,in_seq.taxid,full_fsa_file);
  if(tmp_seq.empty()) return;

  if(is_cat_mode){
    harvest_cat_file(tmp_info,full_cat_file);
  }

  in_seq.dat.push_back(align_dat_t());
  if(fopt.is_skip_filters){
    no_filter(in_seq.dat.back(),tmp_info,tmp_seq);
  } else {
    codon_filter(in_seq.dat.back(),tmp_info,tmp_seq,fopt);
  }
  in_seq.dat.back().label = label;
}




const char dir_delim('/');
const char* fsa_ext(".fsa");
const char* cat_ext(".cat");




void
process_seq_data(nuc_seq_data& in_seq,
                 const seq_filter_options& fopt,
                 const char* seq_file_node,
                 const char* cat_file_node,
                 const char* cat_labels_file){

  std::ostream& log_os(std::cerr);

  if(! (seq_file_node && strlen(seq_file_node))){
    die("invalid input sequence file/directory");
  }

  // initialize in_seq:
  in_seq.dat.clear();
  in_seq.taxid.clear();
  in_seq.id_2_catlabel.clear();

  // tmp hack: force 0==unassigned
  in_seq.id_2_catlabel[UNASSIGNED_CAT_ID]=UNASSIGNED_CAT_LABEL;

  bool is_cat_mode(false);
  if(cat_file_node && strlen(cat_file_node)){
    is_cat_mode=true;
    if(cat_labels_file && strlen(cat_labels_file)){
      harvest_catlabel(in_seq.id_2_catlabel,cat_labels_file);
    }
  }

  DIR *dir(opendir(seq_file_node));
  DIR *cat_dir(0);
  if(is_cat_mode) cat_dir=opendir(cat_file_node);

  align_dat_t::info_type tmp_info;
  tmp_seq_type tmp_seq;

  if(dir){
    log_os << "reading sequence directory: " << seq_file_node << "\n";
    if(is_cat_mode){
      if(cat_dir==0) die("expected category directory");
      log_os << "             cat directory: " << cat_file_node << "\n";
      closedir(cat_dir);
    }

    string full_fsa_file;
    string full_cat_file;
    unsigned attempted_align_count(0);

    dirent* ent;
    while (0!=(ent=readdir(dir))){
      if( strstr(ent->d_name,fsa_ext) == NULL ) continue;
      const char* fsa_file(ent->d_name);

      full_fsa_file=string(seq_file_node)+dir_delim+fsa_file;

      // strip off filename suffix and store label:
      string fsa_label(fsa_file);
      size_t a(fsa_label.rfind('.'));
      if(a != string::npos) fsa_label.erase(a);

      if(is_cat_mode){
        full_cat_file=string(cat_file_node)+dir_delim+fsa_label+cat_ext;
      }

      harvest_file_set(in_seq,tmp_info,tmp_seq,full_fsa_file.c_str(),is_cat_mode,
                       full_cat_file.c_str(),fopt,fsa_label);

      attempted_align_count++;
      const unsigned readable_align_count(in_seq.dat.size());

      if(attempted_align_count%20==0) log_os << ".";
      if(attempted_align_count%1000==0) log_os << "\n";
      if(fopt.max_align_count>0 && readable_align_count>=fopt.max_align_count) break;
    }
    closedir(dir);
    log_os << "\n";

  } else {
    log_os << "reading sequence file: " << seq_file_node << "\n";

    string fsa_label(seq_file_node);
    size_t a(fsa_label.rfind('.'));
    if(a != string::npos) fsa_label.erase(a);

    if(is_cat_mode){
      if(cat_dir) die("expected category file");
      log_os << "             cat file: " << cat_file_node << "\n";
    }

    harvest_file_set(in_seq,tmp_info,tmp_seq,seq_file_node,is_cat_mode,
                     cat_file_node,fopt,fsa_label);
  }
}
