// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __RATE_GTOR_H
#define __RATE_GTOR_H

#include "cat_manager.h"
#include "condition_func.h"
#include "denovo_types.h"
#include "param_view.h"
#include "rate_gtor_options.h"
#include "subs_ml_model_comex.h"
#include "subs_ml_ptol.h"
#include "subs_ml_types.h"
#include "util/math/matrix_util.h"
#include "util/math/prob_util.h"
#include "util/general/die.h"

#include <iosfwd>
#include <iterator>
#include <vector>

struct root_gtor;


/// \brief parameterized rate matrix class
///
struct rate_gtor_base {

  typedef rate_gtor_base self_t;

  rate_gtor_base(const rate_gtor_sml_share& smls)
    : _smls(smls),
      _psize(0),
      _group_catsize(1),
      _param(0),
      _is_train_param(),
      _is_free_block_start(),
      _is_free_block_stop(),
      _param_variance(){}

  rate_gtor_base(const self_t& s,
                 const rate_gtor_sml_share& smls)
    : _smls(smls),
      _psize(s._psize),
      _group_catsize(s._group_catsize),
      _param(0),
      _is_train_param(s._is_train_param),
      _is_free_block_start(s._is_free_block_start),
      _is_free_block_stop(s._is_free_block_stop),
      _param_variance() {
    if(_psize) {
      _param = new smlfloat[_psize];
    }
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) {
      _param[i] = s._param[i];
    }
  }

  virtual ~rate_gtor_base() {
    if(_param) delete [] _param;
  }

private:
  rate_gtor_base(const self_t& s);

  self_t& operator=(const self_t& rhs); // disable copy
public:

  virtual rate_gtor_base* clone(const rate_gtor_sml_share& smls) const = 0;

  RATE_GTOR_MODEL::index_t model_type() const { return _smls.rate_gtor_model(); }

  unsigned state_size() const { return _smls.state_size(); }

  virtual
  unsigned state_size_conditioned() const { return state_size(); }

  virtual
  unsigned submodel_size() const { return 1; }

  virtual
  unsigned submodel_state_size(unsigned s) const { if(s>=submodel_size()) abort(); return state_size(); }

  virtual
  condition_func* submodel_condition_func_factory(unsigned s) const {
    if(s>=submodel_size()) abort(); return condition_func_factory();
  }

  virtual
  void submodel_pdistro_reduction(const prob_t* in,unsigned s,prob_t* out) const {
    if(s>=submodel_size()) die("invalid pdistro_reduction call");
    std::copy(in,in+state_size(),out);
  }

  virtual
  void submodel_fuse_pdistros(const prob_t* const * in,
                              prob_t* out) const {
    if(submodel_size()>1) die("invalid fuse_pdistros call");
    std::copy(in[0],in[0]+state_size(),out);
  }

  virtual
  void submodel_state_reduction_map(unsigned s,unsigned* out) const {
    if(s>=submodel_size()) abort();
    const unsigned ss(state_size());
    for(unsigned i(0);i<ss;++i) out[i] = i;
    // remap ambiguous state too:
    out[ss] = ss;
  }

  virtual
  void submodel_adjust_p(prob_t*,const unsigned) const {}

  unsigned param_size(const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const;

  unsigned category_size() const { return category_size_site()*category_size_group(); }

  unsigned category_size_site() const {
    return get_cat_manager().get_cat_size();
  }

  unsigned category_size_group() const { return _group_catsize; }

protected:
  unsigned category_size_assigned() const {
    return get_cat_manager().get_assigned_cat_size();
  }

public:
  /// \brief use the background distribution and the context averaged
  /// nuc rates to generate time scale (default)
  virtual smlfloat rate_time_bg_scale() const = 0;

  /// \brief use the background distribution to generate a time scale
  /// for each nucleotide context, average these scales by nuc
  /// background.
  virtual
  smlfloat rate_time_context_avg_bg_scale() const { return rate_time_bg_scale(); }

#if 0
  /// \brief use the rate matrix stationary distro and the context
  /// averaged nuc rates to generate time scale
  virtual smlfloat rate_time_stat_scale() const = 0;
#endif

  SITE_MODEL::index_t site_model() const {
    const SITE_MODEL::index_t sm(RATE_GTOR_MODEL::convert_to_site_model(model_type()));
    assert(sm!=SITE_MODEL::NONE);
    return sm;
  }

  template <typename OutputIterator>
  OutputIterator
  param_state(OutputIterator x, const PARAM_VIEW::index_t pv) const;

  template <typename InputIterator>
  InputIterator
  set_param_state(InputIterator x, const PARAM_VIEW::index_t pv);

  template <typename InputIterator>
  InputIterator
  attach_ci(InputIterator ci){
    const unsigned ps(param_size());
    _param_variance.clear();
    _param_variance.resize(ps,0.);

    for(unsigned i(0);i<ps;++i){
      if(_is_train_param[i]){
        _param_variance[i] = ci->variance;
        ci++;
      }
    }
    return ci;
  }

  virtual
  void store_state(std::ostream&) const {}

  virtual
  void load_state(std::istream&) {}

public:
  void
  dump(const char* label,
       std::ostream& os) const;

  virtual
  void
  report(const char* label,
         std::ostream& os) const = 0;

  virtual
  condition_func* condition_func_factory() const { return new condition_func;}

  /// \brief get the instantaneous rates
  virtual void rates(smlfloat rates[],
                     const rates_func_options& opt = rates_func_options()) const = 0;

  template <typename RandomAccessIterator>
  void
  get_site_category_pdistro(RandomAccessIterator x) const {
    const unsigned cs(category_size_site());
    for(unsigned i(0);i<cs;++i){ *(x+i) = get_site_category_proportional_prob(i); }
    pdistro_norm(x,x+cs);
#ifdef DEBUG
    pdistro_check(x,cs,SUBS_ML_PTOL);
#endif
  }

  template <typename RandomAccessIterator>
  void
  get_group_category_pdistro(RandomAccessIterator x) const {
    const unsigned cs(category_size_group());
    for(unsigned i(0);i<cs;++i){ *(x+i) = get_group_category_prob(i); }
#ifdef DEBUG
    pdistro_check(x,cs,SUBS_ML_PTOL);
#endif
  }

  virtual
  void reset_param(const DENOVO_MODEL::index_t dm,
                   const bool is_protected_call=false);

  // per site lnp penalty
  virtual
  smlfloat param_penalty() const { return 0.;}

  const prob_t* bg_pdistro() const { return _smls.obs_state_distro(); }

  const prob_t* bg_pdistro_cat(const unsigned site_category) const {
    const unsigned acn(category_size_assigned());
    if(acn){
      assert(site_category<acn);
      return _smls.obs_state_distro_assigned_cat(site_category);
    } else {
      return bg_pdistro();
    }
  }

  template <typename OutputIterator>
  OutputIterator
  is_train_param(OutputIterator x) const {
    const unsigned ps(param_size());
    for(unsigned p(0);p<ps;++p) {
      *x = _is_train_param[p];
      ++x;
    }
    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_is_train_param(InputIterator x) {
    const unsigned ps(param_size());
    for(unsigned p(0);p<ps;++p) {
      _is_train_param[p] = *x;
      ++x;
    }
    return x;
  }

  virtual
  void report_root(const root_gtor& rg,
                   const char* label,
                   std::ostream& os) const = 0;

  virtual void bg_pdistro_update() { }

  /////////////////////////////////////////////////////////////////
protected:
  virtual
  smlfloat
  trainable_param_transform(const unsigned i) const {
    return _param[i];
  }

  virtual
  void
  trainable_param_untransform(const unsigned i,
                              const smlfloat p) {
    _param[i] = p;
  }

  unsigned
  param_end() const { return 0; }

  virtual
  unsigned pseudo_param_size() const { return 0; }

  virtual
  void set_pseudo_param(const unsigned param_no,
                        const smlfloat val);

  virtual
  prob_t
  get_site_category_proportional_prob(const int /*site_category*/) const { return 1.; }

  virtual
  prob_t
  get_group_category_prob(const int /*group_category*/) const { return 1.; }

  void check_param() const;

  // fix_input_param sets input parameters to legal values (normalizing
  // prob distros -- it is intended to repair input parameters and will
  // change the meaning of the model.
  //
  virtual void fix_input_param() {}

  void set_param_size(const unsigned p);

  void
  set_category_size_group(unsigned c){ _group_catsize = c; }

  unsigned
  get_typed_site_cat_size(const SITE_CAT_TYPE::index_t sct) const {
    return get_cat_manager().get_typed_cat_size(sct);
  }

  unsigned
  get_typed_cat_from_site_cat(const SITE_CAT_TYPE::index_t sct,
                              const unsigned site_cat) const {
    return get_cat_manager().get_typed_cat_no(sct,site_cat);
  }

  void
  get_typed_cat_dependent_prior_prob(const SITE_CAT_TYPE::index_t sct,
                                     prob_t* p_prob) const;

  smlfloat
  param_variance(const unsigned i) const {
    if(_param_variance.size()>i){
      return _param_variance[i];
    } else {
      return 0.;
    }
  }

  //////////////////////////////////////////////////////////
private:
  template <typename FloatType>
  void
  set_uniform_param_state(FloatType x) {
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) _param[i] = x;
    fix_input_param();
  }

  const cat_manager&
  get_cat_manager() const { return _smls.get_cat_manager(); }

///////////////////////////
  const rate_gtor_sml_share _smls; ///< store external reference to shared subs_ml_model data
  unsigned _psize;
  unsigned _group_catsize;
  /// \todo maybe change param to vector
protected:
  smlfloat* _param;
  std::vector<bool> _is_train_param;
  std::vector<bool> _is_free_block_start;
  std::vector<bool> _is_free_block_stop;
private:
  std::vector<smlfloat> _param_variance;
};

#include "rate_gtor.hh"

#endif
