// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor_nscodon_base.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __RATE_GTOR_NSCODON_BASE_H
#define __RATE_GTOR_NSCODON_BASE_H


#include "rate_gtor_nuc_base.h"
#include "rate_gtor_nscodon_options.h"

#include <iosfwd>
#include <utility>

namespace AA_PARAM {
  enum index_t {
    normal,
    zero,
    one };
}


/// \brief base rate generator for coding sequence state models
///
struct rate_gtor_nscodon_base : public rate_gtor_nuc_base {

  typedef rate_gtor_nuc_base base_t;
  typedef rate_gtor_nscodon_base self_t;

  rate_gtor_nscodon_base(const rate_gtor_nuc_options& nopt,
                         const rate_gtor_nscodon_options& copt,
                         const rate_gtor_sml_share& smls)
    : base_t(nopt,smls), _data() {
    _data.copt = copt;
    init_class_storage();
    init_param_locks(copt.lopt);
    init_cat_labels();
  }

  rate_gtor_nscodon_base(const self_t& s,
                         const rate_gtor_sml_share& smls)
  : base_t(s,smls), _data(s._data) {}

  virtual
  void rates(smlfloat rates[],
             const rates_func_options& opt = rates_func_options()) const;

  virtual
  void rates_variance(irv_t<smlfloat> rates[],
                      const rates_func_options& opt = rates_func_options()) const;

  virtual
  void store_state(std::ostream& os) const;

  virtual
  void load_state(std::istream& is);

  /// \brief prints nuc rates and aa selection parameters
  virtual
  void report(const char* label,std::ostream& os) const;

  virtual
  void reset_param(const DENOVO_MODEL::index_t dm,
                   const bool is_protected_call = false);

  void
  set_site_select_cats_prob(const smlfloat* x){
    const unsigned st = paramset_start(PARAMSET_SITE_SELECT_CATS_PROB);
    const unsigned s = paramset_size(PARAMSET_SITE_SELECT_CATS_PROB);
    for(unsigned i(0);i<(s);++i)  _param[i+st] = x[i];
  }

  void
  set_site_select_matrix_cats_prob(const smlfloat* x){
    const unsigned st = paramset_start(PARAMSET_SITE_SELECT_MATRIX_CATS_PROB);
    const unsigned s = paramset_size(PARAMSET_SITE_SELECT_MATRIX_CATS_PROB);
    for(unsigned i(0);i<(s);++i)  _param[i+st] = x[i];
  }

  irv_t<smlfloat>
  get_category_nsaa_param(NSAA::index_t from,
                          NSAA::index_t to,
                          const unsigned site_category,
                          const unsigned group_category) const;

  smlfloat
  get_codon_bias(const NSCODON::index_t from,
                 const NSCODON::index_t to) const;

  const prob_t* bg_pdistro_nuc_pos(const unsigned pos) const {
    return _data.bg_pdistro_nuc_pos.val[pos];
  }

  const prob_t* bg_pdistro_nuc_pos_cat(const unsigned site_cat,
                                       const unsigned pos) const {
     const unsigned acn(_data.bg_pdistro_nuc_pos_cat.dim1());
     assert(acn==0 || site_cat<acn);
     if(acn) { return _data.bg_pdistro_nuc_pos_cat.val[site_cat][pos]; }
     else    { return bg_pdistro_nuc_pos(pos); }
  }

  const prob_t*
  bg_pdistro_nuc_pos_conditioned_on_5p(const unsigned pos,
                                       const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const {
    return _data.bg_pdistro_nuc_pos_conditioned_on_5p.val[pos]+dep_n*NUC::SIZE;
  }

  const prob_t*
  bg_pdistro_nuc_pos_conditioned_on_5p_cat(const unsigned site_cat,
                                           const unsigned pos,
                                           const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const{
    const unsigned acn(_data.bg_pdistro_nuc_pos_conditioned_on_5p_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn) { return _data.bg_pdistro_nuc_pos_conditioned_on_5p_cat.val[site_cat][pos]+dep_n*NUC::SIZE; }
    else    { return bg_pdistro_nuc_pos_conditioned_on_5p(pos,dep_n); }
  }

  const prob_t* bg_pdistro_nuc_pos_conditioned_on_3p(const unsigned pos,
                                                     const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const{
    return _data.bg_pdistro_nuc_pos_conditioned_on_3p.val[pos]+dep_n*NUC::SIZE;
  }

  const prob_t*
  bg_pdistro_nuc_pos_conditioned_on_3p_cat(const unsigned site_cat,
                                           const unsigned pos,
                                           const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const{
    const unsigned acn(_data.bg_pdistro_nuc_pos_conditioned_on_3p_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn) { return _data.bg_pdistro_nuc_pos_conditioned_on_3p_cat.val[site_cat][pos]+dep_n*NUC::SIZE; }
    else    { return bg_pdistro_nuc_pos_conditioned_on_3p(pos,dep_n); }
  }

  const prob_t* bg_pdistro_nscodon() const { return _data.bg_pdistro_nscodon.val; }

  const prob_t* bg_pdistro_nscodon_cat(const unsigned site_cat) const {
    const unsigned acn(_data.bg_pdistro_nscodon_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn) { return _data.bg_pdistro_nscodon_cat.val[site_cat]; }
    else    { return bg_pdistro_nscodon(); }
  }

private:
  prob_t* bg_pdistro_nscodon() { return _data.bg_pdistro_nscodon.val; }

  prob_t* bg_pdistro_nscodon_cat(const unsigned site_cat) {
    return _data.bg_pdistro_nscodon_cat.val[site_cat];
  }

  prob_t* bg_pdistro_nuc_pos(const unsigned pos) {
    return _data.bg_pdistro_nuc_pos.val[pos];
  }

  prob_t* bg_pdistro_nuc_pos_cat(const unsigned site_cat,
                                 const unsigned pos) {
    return _data.bg_pdistro_nuc_pos_cat.val[site_cat][pos];
  }

public:
  // per site lnp penalty
  virtual
  smlfloat
  param_penalty() const {
    static const smlfloat paa_penalty_val(50.);
    static const smlfloat paa_penalty_factor(0.0001);
    const unsigned st(paramset_start(PARAMSET_AA));
    const unsigned s(paramset_size(PARAMSET_AA));
    smlfloat pen(0.);
    for(unsigned i(st);i<(st+s);++i)  if(_param[i]>paa_penalty_val) {
      const smlfloat x(1.+(_param[i]-paa_penalty_val)*paa_penalty_factor);
      pen += (x*x)-1.;
    }
    return pen;
  }

  std::vector<std::string>& site_select_cats_labels() { return _data.site_select_cats_labels; }
  std::vector<std::string>& site_select_matrix_cats_labels() { return _data.site_select_matrix_cats_labels; }

  virtual
  void report_root(const root_gtor& rg,
                   const char* label,
                   std::ostream& os) const;

  // public versions of these methods to set ci subset (temporary!!)
  unsigned
  param_size_aa() const {
    return paramset_size(PARAMSET_AA);
  }

  unsigned
  param_start_aa() const {
    return paramset_start(PARAMSET_AA);
  }

  virtual void bg_pdistro_update();

  /////////////////////////////////////////////////////////////////////////////
protected:

  unsigned
  param_end() const { return paramset_start(PARAMSET_END); }

  virtual
  unsigned pseudo_param_size() const;

  virtual
  void set_pseudo_param(const unsigned param_no,
                        const smlfloat val);

  virtual
  prob_t
  get_site_category_proportional_prob(const int site_category) const{
    if(site_category<0) return 1.;

    const prob_t nuc_base_p(base_t::get_site_category_proportional_prob(site_category));

    const unsigned sc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::SEL_STRENGTH,site_category));
    const prob_t select_p(get_paramset_pdistro_member(PARAMSET_SITE_SELECT_CATS_PROB,sc));

    const unsigned smc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::SEL_MATRIX,site_category));
    const prob_t select_m_p(get_paramset_pdistro_member(PARAMSET_SITE_SELECT_MATRIX_CATS_PROB,smc));

    return nuc_base_p*select_p*select_m_p;
  }

  virtual
  prob_t
  get_group_category_prob(const int group_category) const{
    if(group_category<0) return 1.;
    const unsigned rc = parse_group_rate_category(group_category);
    const unsigned sc = parse_group_select_category(group_category);

    const prob_t rate_p = base_t::get_group_category_prob(rc);
    const prob_t select_p = get_paramset_pdistro_member(PARAMSET_GROUP_SELECT_CATS_PROB,sc);

    return rate_p*select_p;
  }

  unsigned
  parse_group_select_category(const unsigned group_category) const {
    return (group_category/category_size_group_rate())%category_size_group_select();
  }

  std::pair<unsigned,AA_PARAM::index_t>
  get_nsaa_param_index(NSAA::index_t from,
                       NSAA::index_t to,
                       const unsigned site_select_matrix_cat) const;

  irv_t<smlfloat>
  get_nsaa_param(const NSAA::index_t from,
                 const NSAA::index_t to,
                 const unsigned site_select_matrix_cat) const;

  void state_pdistro_to_nscodon_pdistro(const prob_t* state_pdistro,
                                        prob_t* nscodon_pdistro) const;

  void state_pdistro_to_nsaa_pdistro(const prob_t* state_pdistro,
                                     prob_t* nsaa_pdistro) const;

  /////////////////////////////////////////////////////////////////////////////
private:
  virtual
  void fix_input_param();

  virtual
  void state_pdistro_to_nuc_pdistro(const prob_t* state_pdistro,
                                    prob_t* nuc_pdistro) const;

  void stat_pdistro_nscodon(prob_t* sd,
                            const rates_func_options& opt = rates_func_options()) const;

  void stat_pdistro_nsaa(prob_t* sd,
                         const rates_func_options& opt = rates_func_options()) const;

  void bg_pdistro_nscodon_update(){
    state_pdistro_to_nscodon_pdistro(bg_pdistro(),bg_pdistro_nscodon());
    const unsigned acn(category_size_assigned());
    if(acn){
      _data.bg_pdistro_nscodon_cat.init(acn,NSCODON::SIZE);
      for(unsigned c(0);c<acn;++c){
        state_pdistro_to_nscodon_pdistro(bg_pdistro_cat(c),bg_pdistro_nscodon_cat(c));
      }
    }
  }

  void bg_pdistro_nuc_pos_update();

  void bg_pdistro_nuc_pos_conditioned_update();

  void init_class_storage();

  void init_param_locks(const rate_gtor_nscodon_lock_options& lopt);

  void init_cat_labels();

#if 0
  static smlfloat get_paa_switch_value() { return 5.; }
#endif

protected:

  // parameter arrangement:
  enum paramset_t {
    PARAMSET_SITE_SELECT_CATS = base_t::PARAMSET_END,
    PARAMSET_SITE_SELECT_CATS_PROB,
    PARAMSET_SITE_SELECT_MATRIX_CATS_PROB,
    PARAMSET_GROUP_SELECT_CATS,
    PARAMSET_GROUP_SELECT_CATS_PROB,
    PARAMSET_AA,
    PARAMSET_CODON_BIAS,
    PARAMSET_END
  };

  virtual
  unsigned paramset_size(const int p) const;

private:
  // dumps a string identifying each paramset to os
  void
  paramset_label(const paramset_t p,
                 unsigned n,
                 std::ostream& os) const;

  void
  get_aa_param_index_name(const unsigned matrix_cat_no,
                          const unsigned matrix_param_no,
                          std::ostream& os) const;

  /// cat sizes:
  unsigned
  category_size_site_select() const { return get_typed_site_cat_size(SITE_CAT_TYPE::SEL_STRENGTH); }

  unsigned
  category_size_site_select_matrix() const { return get_typed_site_cat_size(SITE_CAT_TYPE::SEL_MATRIX); }

  unsigned
  category_size_group_select() const { return _data.copt.n_group_select_cats; }


  /// other:
  SELECT_MODEL::index_t select_model(const unsigned i) const { return _data.copt.select_model(i); }

  void
  get_nsaa_selection_matrix(irv_t<smlfloat> nsaa_selection[NSAA::SIZE*NSAA::SIZE],
                            const unsigned site_select_matrix_category = 0) const;

  smlfloat
  get_codon_bias_factor(const NSCODON::index_t to) const;

  smlfloat
  get_site_select_cats_expect() const {
#if 0
    const unsigned st = paramset_start(PARAMSET_SITE_SELECT_CATS);
    const unsigned s = paramset_size(PARAMSET_SITE_SELECT_CATS);

    smlfloat sum(0.);
    for(unsigned i(0);i<(s);++i){ sum += _param[st+i]*get_paramset_pdistro_member(PARAMSET_SITE_SELECT_CATS_PROB,i); }
    return sum;
#else
    return 1.;
#endif
  }

  smlfloat
  get_group_select_cats_expect() const {
    const unsigned st = paramset_start(PARAMSET_GROUP_SELECT_CATS);
    const unsigned s = paramset_size(PARAMSET_GROUP_SELECT_CATS);

    smlfloat sum(0.);
    for(unsigned i(0);i<(s);++i){ sum += _param[st+i]*get_paramset_pdistro_member(PARAMSET_GROUP_SELECT_CATS_PROB,i); }
    return sum;
  }

  void reset_random_cat_prob(const paramset_t p);

  static
  unsigned
  paramsize_select_model_unscaled(const SELECT_MODEL::index_t sm);

  static
  unsigned
  paramsize_select_model(const SELECT_MODEL::index_t sm);

  unsigned
  paramset_aa_select_matrix_cat_size(const unsigned cat_no) const {
    return paramsize_select_model(select_model(cat_no));
  }

  unsigned
  paramset_aa_select_matrix_cat_start(const unsigned cat_no) const {
    unsigned val(paramset_start(PARAMSET_AA));

    for(unsigned i(0);i<cat_no;++i){
      val += paramset_aa_select_matrix_cat_size(i);
    }

    return val;
  }

  // to simplify copy ctor:
  struct rate_gtor_nscodon_base_autocopy_data {

    rate_gtor_nscodon_base_autocopy_data()
      : bg_pdistro_nscodon(NSCODON::SIZE),
        bg_pdistro_nuc_pos(CODON::BASE_SIZE,NUC::SIZE),
        bg_pdistro_nuc_pos_conditioned_on_5p(CODON::BASE_SIZE,NUC::SIZE*NUC::SIZE),
        bg_pdistro_nuc_pos_conditioned_on_3p(CODON::BASE_SIZE,NUC::SIZE*NUC::SIZE) {}

    rate_gtor_nscodon_options_base copt;
    std::vector<std::string> site_select_cats_labels;
    std::vector<std::string> site_select_matrix_cats_labels;
    std::vector<std::string> group_select_cats_labels;

    simple_array<prob_t> bg_pdistro_nscodon;
    simple_matrix<prob_t> bg_pdistro_nuc_pos;
    simple_matrix<prob_t> bg_pdistro_nuc_pos_conditioned_on_5p;
    simple_matrix<prob_t> bg_pdistro_nuc_pos_conditioned_on_3p;

    simple_init_matrix<prob_t> bg_pdistro_nscodon_cat;
    simple_init_matrix3d<prob_t> bg_pdistro_nuc_pos_cat;
    simple_init_matrix3d<prob_t> bg_pdistro_nuc_pos_conditioned_on_5p_cat;
    simple_init_matrix3d<prob_t> bg_pdistro_nuc_pos_conditioned_on_3p_cat;
  };

  // ext. data
  rate_gtor_nscodon_base_autocopy_data _data;

  // static data:
  //
  struct aa_exchange_param_info {
    aa_exchange_param_info();

    bool is_1nuc_aa_exchange[NSAA::SIZE*NSAA::SIZE];
    unsigned asymm_size;
    unsigned symm_size;
    unsigned asymm_aatoparam_map[NSAA::SIZE*NSAA::SIZE];
    unsigned symm_aatoparam_map[NSAA::SIZE*NSAA::SIZE];
    std::vector<unsigned> asymm_paramtoaa_map;
    std::vector<unsigned> symm_paramtoaa_map;
  };

  struct aa_codon_map_info {
    aa_codon_map_info();

    NSCODON::index_t nscodon_c2aa_ordermap[NSCODON::SIZE];
    NSCODON::index_t nscodon_aa2c_ordermap[NSCODON::SIZE];
    unsigned aa_codon_size[NSAA::SIZE];
    unsigned aa_codon_start_offset[NSAA::SIZE];
  };

  static const aa_exchange_param_info _aaex;
  static const aa_codon_map_info _aacodon;
};


#endif
