// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor_nscodon_base_report.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "bioseq_report.h"
#include "nsaa_maps.h"
#include "rate_gtor_nscodon_base.h"
#include "rate_gtor_nscodon_base_util.h"
#include "root_gtor.h"
#include "simple_util.h"
#include "stationary_pdistro.h"
#include "util/bio/bioseq_util_io.h"
#include "util/bio/bioseq_util_flux_rate_conversions.h"
#include "util/math/array_util.h"
#include "util/math/matrix_util.h"
#include "util/math/matrix_util_io.h"

#include <cmath>

#include <iomanip>
#include <iostream>
#include <utility>
#include <vector>



/// keep things consistent
static
void
format_root_header(const unsigned root_cat_no,
                   const char* root_label,
                   std::ostream& os){

  os << root_label << " prior prob: cat_no " << static_cast<int>(root_cat_no);
}


static
void
report_nuc_distro_from_nscodon(const prob_t nsp[],
                               const char* label,
                               std::ostream& os){

  os << label << "\n";
  os << "nuc distro:\n";

  prob_t nuc_pdf[NUC::SIZE];
  nscodon_pdf_2_nuc_pdf(nsp,nuc_pdf);
  pdistro_report(nuc_pdf,NUC::SIZE,NUC::syms,os);
}


static
void
report_nuc_distro_from_nscodon_pos(const prob_t nscodon_pdf[],
                                   const char* label,
                                   const unsigned pos,
                                   std::ostream& os){

  os << label << "\n";
  os << "nuc distro from codon pos " << pos << ":\n";

  prob_t nuc_pdf[NUC::SIZE];
  nscodon_pdf_2_nuc_pdf_pos(nscodon_pdf,pos,nuc_pdf);
  pdistro_report(nuc_pdf,NUC::SIZE,NUC::syms,os);
}


static
void
report_nsaa_distro_from_nscodon(const prob_t nsp[],
                                const char* label,
                                std::ostream& os){
  os << label << "\n";
  os << "aa distro:\n";

  prob_t nsaa_pdf[NSAA::SIZE];
  nscodon_pdf_2_nsaa_pdf(nsp,nsaa_pdf);
  pdistro_report(nsaa_pdf,NSAA::SIZE,NSAA::syms,os);
}



const unsigned pdistro_row_write_width(7);



static
void
pdistro_row_write(const prob_t* pdistro,
                  const unsigned N,
                  std::ostream& os,
                  const unsigned width = pdistro_row_write_width,
                  const unsigned prec = 3){

  os.unsetf(std::ios::fixed | std::ios::scientific);
  os.setf(std::ios::showpoint);
  for(unsigned i(0);i<N;++i){
    os << " "
       << std::right
       << std::setw(width-1)
       << std::setprecision(prec)
       << pdistro[i];
  }
  os << "\n";
}



static
smlfloat
normalized_dot(const smlfloat* v1,
               const smlfloat* v2,
               const unsigned N){
  return array_dot(v1,v2,N)/(array_norm2(v1,N)*array_norm2(v2,N));
}



static
smlfloat
rel_ent(const prob_t* pdf1,
        const prob_t* pdf2,
        const unsigned N){

  // relative entroy of two distros:
  smlfloat re(0.);
  for(unsigned i(0);i<N;++i){
    if(fabs(pdf1[i]) <= 0.) continue; // 0*-inf=0
    re += pdf1[i]*std::log(pdf1[i]/pdf2[i]);
  }
  return re;
}



void
rate_gtor_nscodon_base::
report_root(const root_gtor& rg,
            const char* label,
            std::ostream& os) const {

  prob_t bg_pdistro_nsaa[NSAA::SIZE];
  nscodon_pdf_2_nsaa_pdf(bg_pdistro_nscodon(),bg_pdistro_nsaa);

  prob_t stack_nscodon_pdf[NSCODON::SIZE];

  for(unsigned i(0);i<rg.root_cat_size();++i){
    const prob_t* root_pdf(rg.root_cat_distro(i));
    const prob_t* nscodon_pdf(root_pdf);

    const BIO_PDISTRO::index_t bp(BIO_PDISTRO::convert_from_site_model(site_model()));
    if(bp != BIO_PDISTRO::NSCODON){
      bio_pdistro_convert(bp,BIO_PDISTRO::NSCODON,root_pdf,stack_nscodon_pdf);
      nscodon_pdf=stack_nscodon_pdf;
    }

    std::string distro_label;
    {
      std::ostringstream oss;
      format_root_header(i,label,oss);
      distro_label = oss.str();
    }

    // nuc
    report_nuc_distro_from_nscodon(nscodon_pdf,distro_label.c_str(),os);

    // nuc by codon pos
    for(unsigned k(0);k<CODON::BASE_SIZE;++k){
      report_nuc_distro_from_nscodon_pos(nscodon_pdf,distro_label.c_str(),k,os);
    }

    // aa
    report_nsaa_distro_from_nscodon(nscodon_pdf,distro_label.c_str(),os);

    // codons
    os << distro_label.c_str() << "\n";
    print_site_model_distro(nscodon_pdf,SITE_MODEL::NSCODON,os);

    const SITE_MODEL::index_t sm(site_model());
    if(sm==SITE_MODEL::NSC4POST){
      const unsigned ss(state_size());

      // get sequence stationary root
      //
      static const unsigned repeat_pos(3);
      static const unsigned dep_pos(0);

      // get stationary dep nuc distro
      prob_t dep_nuc_stat[NUC::SIZE];
      {
        // extract dependent nuc distro
        prob_t dep_nuc_p[NUC::SIZE*NUC::SIZE];

        std::fill(dep_nuc_p,dep_nuc_p+NUC::SIZE*NUC::SIZE,0.);

        NUC::index_t sn[SITE_MODEL::MAX_BASE_SIZE];
        for(unsigned j(0);j<ss;++j){
          SITE_MODEL::decode_nuc(sm,j,sn);
          dep_nuc_p[sn[repeat_pos]*NUC::SIZE+sn[dep_pos]] += root_pdf[j];
        }

        for(unsigned j(0);j<NUC::SIZE;++j){
          pdistro_norm(dep_nuc_p+j*NUC::SIZE,dep_nuc_p+(j+1)*NUC::SIZE);
        }
        get_stationary_pdistro_from_pmatrix(dep_nuc_stat,dep_nuc_p,NUC::SIZE);
      }

      // get sequence-stationary codon distro
      prob_t sstat_nscodon_pdf[NSCODON::SIZE];
      {
        prob_t nscodon_by_nx_base[NSC4::SIZE];
        for(unsigned j(0);j<NSC4::SIZE;++j) nscodon_by_nx_base[j] = root_pdf[j];

        prob_t* nscodon_by_nx[NUC::SIZE];
        for(unsigned j(0);j<NUC::SIZE;++j){
          nscodon_by_nx[j] = nscodon_by_nx_base+NSC4::get_nscodon_offset(static_cast<NUC::index_t>(j));
        }

        for(unsigned j(0);j<NUC::SIZE;++j){
          pdistro_norm(nscodon_by_nx[j],nscodon_by_nx[j]+NSCODON::SIZE);
        }

        std::fill(sstat_nscodon_pdf,sstat_nscodon_pdf+NSCODON::SIZE,0.);

        for(unsigned k(0);k<NUC::SIZE;++k){
          for(unsigned j(0);j<NSCODON::SIZE;++j){
            sstat_nscodon_pdf[j] += nscodon_by_nx[k][j]*dep_nuc_stat[k];
          }
        }
      }

      os << distro_label.c_str() << " (sequence-stationary)\n";
      print_site_model_distro(sstat_nscodon_pdf,SITE_MODEL::NSCODON,os);
    }

    // report differences between root distribution and obs:
    //
    {
      const prob_t* root_nscodon_pdf(nscodon_pdf);

      {
        prob_t root_nuc_pdf[NUC::SIZE];
        nscodon_pdf_2_nuc_pdf(root_nscodon_pdf,root_nuc_pdf);

        os << distro_label << "\n";
        os << "root comparison to obs distro (nuc):\n";

        os << "NURO:      ";
        for(unsigned k(0);k<NUC::SIZE;++k) {
          for(unsigned j(0);j<((pdistro_row_write_width+2)-1);++j) os << " ";
          os << NUC::syms[k];
        }
        os << "\n";

        os << "NURO ROOT: ";
        pdistro_row_write(root_nuc_pdf,NUC::SIZE,os,pdistro_row_write_width+2,5);

        os << "NURO OBS:  ";
        pdistro_row_write(bg_pdistro_nuc(),NUC::SIZE,os,pdistro_row_write_width+2,5);

        os << std::setprecision(7);
        os << "NURO dot(ROOT,OBS):  ";
        os << normalized_dot(root_nuc_pdf,bg_pdistro_nuc(),NUC::SIZE) << "\n";

        os << "NURO re(ROOT,OBS):   ";
        os << rel_ent(root_nuc_pdf,bg_pdistro_nuc(),NUC::SIZE) << "\n";
        os << "\n";
      }

      {
        prob_t root_nsaa_pdf[NSAA::SIZE];
        nscodon_pdf_2_nsaa_pdf(root_nscodon_pdf,root_nsaa_pdf);

        os << distro_label << "\n";
        os << "root comparison to obs distro (nsaa):\n";

        os << "AARO:      ";
        for(unsigned k(0);k<NSAA::SIZE;++k) {
          for(unsigned j(0);j<(pdistro_row_write_width-1);++j) os << " ";
          os << NSAA::syms[k];
        }
        os << "\n";

        os << "AARO ROOT: ";
        pdistro_row_write(root_nsaa_pdf,NSAA::SIZE,os);

        os << "AARO OBS:  ";
        pdistro_row_write(bg_pdistro_nsaa,NSAA::SIZE,os);

        os << std::setprecision(7);
        os << "AARO dot(ROOT,OBS):  ";
        os << normalized_dot(root_nsaa_pdf,bg_pdistro_nsaa,NSAA::SIZE) << "\n";

        os << "AARO re(ROOT,OBS):   ";
        os << rel_ent(root_nsaa_pdf,bg_pdistro_nsaa,NSAA::SIZE) << "\n";
        os << "\n";
      }

      os << distro_label << "\n";
      os << "root comparison to obs distro (nscodon):\n";

      os << "NCRO dot(ROOT,OBS):  ";
      os << normalized_dot(root_nscodon_pdf,bg_pdistro_nscodon(),NSCODON::SIZE) << "\n";

      os << "NCRO re(ROOT,OBS):   ";
      os << rel_ent(root_nscodon_pdf,bg_pdistro_nscodon(),NSAA::SIZE) << "\n";
      os << "\n";
    }
  }
}




struct smc2{
  smc2(unsigned _i,
       unsigned _j,
       irv_t<smlfloat> _val1,
       irv_t<smlfloat> _val2,
       bool is_doswap=true)
  : i(_i),j(_j),val1(_val1),val2(_val2) {
    if(is_doswap){
      if(val1<val2){
        std::swap(val1,val2);
        std::swap(i,j);
      }
    }
    ratio = val1/val2;
  }

  bool
  operator<(const smc2& right) const {
    return ratio < right.ratio;
  }

  unsigned i;
  unsigned j;
  irv_t<smlfloat> val1;
  irv_t<smlfloat> val2;
  irv_t<smlfloat> ratio;
};


struct smc3{
  smc3(unsigned _i,
       unsigned _j,
       irv_t<smlfloat> _val1,
       irv_t<smlfloat> _val2)
  : i(_i),j(_j),val1(_val1),val2(_val2) { valdiff = val1-val2; }

  bool
  operator<(const smc3& right) const {
    return valdiff < right.valdiff;
  }

  unsigned i;
  unsigned j;
  irv_t<smlfloat> val1;
  irv_t<smlfloat> val2;
  irv_t<smlfloat> valdiff;
};




static
void
report_sorted_nsaa_exchange_parameters(const irv_t<smlfloat>* nsaa_rates,
                                       std::ostream& os,
                                       const char* label){

  report_sorted_nsaa_pair_values(nsaa_rates,os,label);

  {
    std::vector<smc2> vsm;

    /// dump rate pairs -- sorted by abs(log(a/b))
    for(unsigned i(0);i<NSAA::SIZE;++i){
      for(unsigned j(i+1);j<NSAA::SIZE;++j){
        if(i == j) continue;
        irv_t<smlfloat> val1 = nsaa_rates[j+i*NSAA::SIZE];
        if(std::fabs(val1) < std::numeric_limits<smlfloat>::epsilon() ) continue;
        irv_t<smlfloat> val2 = nsaa_rates[i+j*NSAA::SIZE];
        vsm.push_back(smc2(i,j,val1,val2));
      }
    }
    std::sort(vsm.begin(),vsm.end());
    std::reverse(vsm.begin(),vsm.end());

    os << "RATIO SORTED " << label << " PARAMETERS:\n";
    os << "(from->to=param & from->to=param & ratio )\n";
    for(unsigned i(0);i<vsm.size();++i){
      os << NSAA::syms[vsm[i].i] << "->"
         << NSAA::syms[vsm[i].j] << "="
         << std::setw(5)
         << std::left
         << vsm[i].val1 << " & "
         << NSAA::syms[vsm[i].j] << "->"
         << NSAA::syms[vsm[i].i] << "="
         << std::setw(5)
         << std::left
         << vsm[i].val2 << " & "
         << std::setw(4)
         << std::left
         << vsm[i].ratio;

      { // add codon transition list:
        std::vector<nscodon_pair> vc;
        get_codon_paths(vc,
                        static_cast<NSAA::index_t>(vsm[i].i),
                        static_cast<NSAA::index_t>(vsm[i].j));

        os << " &   "
           << NSAA::syms[vsm[i].i] << "->"
           << NSAA::syms[vsm[i].j] << " codons: ";
        for(unsigned j(0);j<vc.size();++j){
          if(j) os << ",";
          os << NSCODON::print(vc[j].c1) << "->" << NSCODON::print(vc[j].c2);
        }
      }
      os << "\n";
    }
    os << "\n";
  }
}




static
void
report_paired_aa_pdistro(const irv_t<smlfloat> distro1[NSAA::SIZE],
                         const irv_t<smlfloat> distro2[NSAA::SIZE],
                         std::ostream& os,
                         const char* label,
                         const char* label1,
                         const char* label2){

  //// dump sorted mut/acc rates:
  std::vector<std::pair<irv_t<smlfloat>,unsigned> > vs1,vs2;

  for(unsigned i(0);i<NSAA::SIZE;++i){
    vs1.push_back(std::make_pair(distro1[i],i));
    vs2.push_back(std::make_pair(distro2[i],i));
  }

  std::sort(vs1.begin(),vs1.end());
  std::reverse(vs1.begin(),vs1.end());
  std::sort(vs2.begin(),vs2.end());
  std::reverse(vs2.begin(),vs2.end());

  os << "SORTED " << label << " " << label1 << ":\n";
  for(unsigned i(0);i<NSAA::SIZE;++i)
    os << NSAA::syms[vs1[i].second] << "->X: "
       << std::setprecision(7) << vs1[i].first << "\n";
  os << "\n";

  os << "SORTED " << label << " " << label2 << ":\n";
  for(unsigned i(0);i<NSAA::SIZE;++i)
    os << "X->" << NSAA::syms[vs2[i].second] << ": "
       << std::setprecision(7) << vs2[i].first << "\n";
  os << "\n";


  std::vector<smc2> vs_ratio;
  for(unsigned i(0);i<NSAA::SIZE;++i){
    vs_ratio.push_back(smc2(i,i,distro1[i],distro2[i],false));
  }
  std::sort(vs_ratio.begin(),vs_ratio.end());
  std::reverse(vs_ratio.begin(),vs_ratio.end());

  os << label << " (" << label1 << "/" << label2 << ") RATIO SORTED:\n";
  os << "(from->to=param & from->to=param & ratio )\n";
  for(unsigned i(0);i<vs_ratio.size();++i){
    os << NSAA::syms[vs_ratio[i].i] << "->X="
       << std::setw(5)
       << std::left
       << vs_ratio[i].val1 << " & X->"
       << NSAA::syms[vs_ratio[i].i] << "="
       << std::setw(5)
       << std::left
       << vs_ratio[i].val2 << " & "
       << std::setw(4)
       << std::left
       << vs_ratio[i].ratio;
    os << "\n";
  }
  os << "\n";


  std::vector<smc3> vs_diff;
  for(unsigned i(0);i<NSAA::SIZE;++i){
    vs_diff.push_back(smc3(i,i,distro1[i],distro2[i]));
  }
  std::sort(vs_diff.begin(),vs_diff.end());
  std::reverse(vs_diff.begin(),vs_diff.end());

  os << label << " (" << label1 << " - " << label2 << ") DIFF SORTED:\n";
  os << "(from->to=param & from->to=param & diff(1-2) )\n";
  for(unsigned i(0);i<vs_diff.size();++i){
    os << NSAA::syms[vs_diff[i].i] << "->X="
       << std::setw(5)
       << std::left
       << vs_diff[i].val1 << " & X->"
       << NSAA::syms[vs_diff[i].i] << "="
       << std::setw(5)
       << std::left
       << vs_diff[i].val2 << " & "
       << std::setw(4)
       << std::left
       << vs_diff[i].valdiff;
    os << "\n";
  }
  os << "\n";
}




static
void
report_nsaa_exchange_averages(const irv_t<smlfloat> nsaa_mat[NSAA::SIZE*NSAA::SIZE],
                              std::ostream& os,
                              const char* label){

  // find simple averages for "to" and "from" values in matrix
  irv_t<smlfloat> from_avg[NSAA::SIZE];
  unsigned from_count[NSAA::SIZE];
  irv_t<smlfloat> to_avg[NSAA::SIZE];
  unsigned to_count[NSAA::SIZE];
  for(unsigned i(0);i<NSAA::SIZE;++i){
    from_avg[i] = 0.;
    from_count[i] = 0;
    to_avg[i] = 0.;
    to_count[i] = 0;
  }
  for(unsigned i(0);i<NSAA::SIZE;++i){
    for(unsigned j(0);j<NSAA::SIZE;++j){
      if(i == j) continue;
      const irv_t<smlfloat> val = nsaa_mat[j+i*NSAA::SIZE];
      if(std::fabs(val) < std::numeric_limits<smlfloat>::epsilon() ) continue;
      from_avg[i] += val;
      from_count[i]++;
      to_avg[j] += val;
      to_count[j]++;
    }
  }
  for(unsigned i(0);i<NSAA::SIZE;++i){
    from_avg[i] /= static_cast<smlfloat>(from_count[i]);
    to_avg[i] /= static_cast<smlfloat>(to_count[i]);
  }

  report_paired_aa_pdistro(from_avg,to_avg,os,label,"FROM AVG","TO AVG");
}




static
void
report_nsaa_exchange_sums(const irv_t<smlfloat> nsaa_mat[NSAA::SIZE*NSAA::SIZE],
                          std::ostream& os,
                          const char* label){

  // find simple averages for to and from values in matrix
  irv_t<smlfloat> from_stat[NSAA::SIZE];
  irv_t<smlfloat> to_stat[NSAA::SIZE];
  for(unsigned i(0);i<NSAA::SIZE;++i){
    from_stat[i] = 0.;
    to_stat[i] = 0.;
  }
  for(unsigned i(0);i<NSAA::SIZE;++i){
    for(unsigned j(0);j<NSAA::SIZE;++j){
      if(i == j) continue;
      const irv_t<smlfloat> val = nsaa_mat[j+i*NSAA::SIZE];
      if(std::fabs(val) < std::numeric_limits<smlfloat>::epsilon() ) continue;
      from_stat[i] += val;
      to_stat[j] += val;
    }
  }

  report_paired_aa_pdistro(from_stat,to_stat,os,label,"FROM ALL","TO ALL");
}




static
void
report_nsaa_matrix_full(const irv_t<smlfloat>* nsaa_mat,
                        std::ostream& os,
                        const char* label){

  os << label << " MATRIX:\n";
  matrix_report(nsaa_mat,NSAA::SIZE,NSAA::syms,os);

  report_sorted_nsaa_exchange_parameters(nsaa_mat,os,label);
}


#if 0
static
void
report_nsaa_grouping(const irv_t<smlfloat>* nsaa_selection,
                     const irv_t<smlfloat>* nsaa_flux,
                     const char* label,
                     const unsigned* reduction_map,
                     const char* syms,
                     const unsigned N,
                     std::ostream& os){

  irv_t<smlfloat>* selection_avg(new irv_t<smlfloat>[N*N]);
  matrix_state_reduction(selection_avg,N,nsaa_selection,NSAA::SIZE,reduction_map,true);

  os << "AA SELECTION :: AVERAGE BY " << label << ":\n";
  matrix_report(selection_avg,N,syms,os);

  irv_t<smlfloat>* flux_sum(selection_avg);
  matrix_state_reduction(flux_sum,N,nsaa_flux,NSAA::SIZE,reduction_map);

  os << "AA FLUX :: SUM BY " << label << ":\n";
  matrix_report(flux_sum,N,syms,os);

  delete [] flux_sum;
}
#endif





// extremely sloppy amino acid mutability set dfs
void
aaset_dfs(const bool* is_connected,
          const unsigned* aa_list,
          std::vector<bool>& hit_list,
          const unsigned N,
          const unsigned x){

  hit_list[x] = true;

  for(unsigned i(0);i<N;++i){
    if(i==x) continue;
    if( is_connected[aa_list[i]+NSAA::SIZE*aa_list[x]] ){
      if(! hit_list[i]){
        aaset_dfs(is_connected,aa_list,hit_list,N,i);
      }
    }
  }
}




bool
is_aaset_connected(const bool* is_connected,
                   const unsigned* aa_list,
                   const unsigned N){

  std::vector<bool> hit_list(N,false);

  // run a dfs and see if anything isn't colored
  aaset_dfs(is_connected,aa_list,hit_list,N,0);

  for(unsigned i(0);i<N;++i) if(! hit_list[i]) return false;

  return true;
}


irv_t<smlfloat>
get_Nset_selavg(const irv_t<smlfloat>* neutral_nsaa_flux,
                const irv_t<smlfloat>* nsaa_flux,
                const unsigned* aa_list,
                const unsigned N){

  irv_t<smlfloat> sum(0.);
  unsigned count(0);
  for(unsigned i(0);i<N;++i){
    for(unsigned j(0);j<N;++j){
      if(i==j) continue;
      if(neutral_nsaa_flux[aa_list[i]+aa_list[j]*NSAA::SIZE] <= 0.) continue;
      sum += nsaa_flux[aa_list[i]+aa_list[j]*NSAA::SIZE]/
        neutral_nsaa_flux[aa_list[i]+aa_list[j]*NSAA::SIZE];
      count++;
    }
  }
  if(count) return sum/static_cast<smlfloat>(count);
  else      return irv_t<smlfloat>(0.);
}


struct aaset_selavg_score {

  irv_t<smlfloat>
  score(unsigned* aaid_set,
        const unsigned set_size) const {
    return get_Nset_selavg(neutral_nsaa_flux,nsaa_flux,aaid_set,set_size);
  }

  const irv_t<smlfloat>* neutral_nsaa_flux;
  const irv_t<smlfloat>* nsaa_flux;
};


irv_t<smlfloat>
get_Nset_dn_ds(const irv_t<smlfloat>* neutral_nsaa_flux,
               const irv_t<smlfloat>* nsaa_flux,
               const unsigned* aa_list,
               const unsigned N){

  smlfloat flux (0.);
  smlfloat nflux(0.);
  for(unsigned i(0);i<N;++i){
    for(unsigned j(0);j<N;++j){
      if(i==j) continue;
      flux += nsaa_flux[aa_list[i]+aa_list[j]*NSAA::SIZE];
      nflux += neutral_nsaa_flux[aa_list[i]+aa_list[j]*NSAA::SIZE];
    }
  }
  if(nflux>0) return flux/nflux;
  else        return irv_t<smlfloat>(0.);
}


struct aaset_dn_ds_score {

  irv_t<smlfloat>
  score(unsigned* aaid_set,
        const unsigned set_size) const {
    return get_Nset_dn_ds(neutral_nsaa_flux,nsaa_flux,aaid_set,set_size);
  }

  const irv_t<smlfloat>* neutral_nsaa_flux;
  const irv_t<smlfloat>* nsaa_flux;
};



template <typename scorefunc>
void
score_aaset(const scorefunc& sf,
            const bool* is_reachable,
            unsigned* aaid_set,
            const unsigned set_size,
            unsigned s,
            std::ostream& os){

  const unsigned end(s==0 ? static_cast<unsigned>(NSAA::SIZE) : aaid_set[s-1]);

  for(unsigned i(0);i<end;++i){
    aaid_set[s] = i;
    if((s+1)==set_size){
      if(! is_aaset_connected(is_reachable,aaid_set,set_size)) continue;
      os << "GKEY " << set_size << " ";
      os << std::setw(9);
      os << sf.score(aaid_set,set_size);
      for(unsigned j(0);j<set_size;++j){
        os << " " << NSAA::syms[aaid_set[j]] << " ";
      }
      os << "\n";
    }else {
      score_aaset(sf,is_reachable,aaid_set,set_size,s+1,os);
    }
  }
}



template <typename scorefunc>
void
score_aaset_start(const irv_t<smlfloat>* neutral_nsaa_flux,
                  const irv_t<smlfloat>* nsaa_flux,
                  const bool* is_reachable,
                  std::ostream& os){

  const unsigned max_set_size(11);
  scorefunc sf;
  sf.neutral_nsaa_flux=neutral_nsaa_flux;
  sf.nsaa_flux=nsaa_flux;

  for(unsigned set_size(2);set_size<max_set_size;++set_size){
    simple_array<unsigned> aaid_set(set_size);
    score_aaset(sf,is_reachable,aaid_set.val,set_size,0,os);
  }
}




void
report_nsaa_sets(const irv_t<smlfloat>* neutral_nsaa_flux,
                 const irv_t<smlfloat>* nsaa_flux,
                 std::ostream& os){

  //  irv_t<smlfloat>* pair_dn_ds(new irv_t<smlfloat>[NSAA::SIZE*NSAA::SIZE]);

  bool is_reachable[NSAA::SIZE*NSAA::SIZE];

  make_reachable_aa_matrix(is_reachable);
  os << std::setprecision(6);

  // use dn/ds sets
  //  score_aaset_start<aaset_dn_ds_score>(neutral_nsaa_flux,nsaa_flux,is_reachable,os);

  // use selection strength average sets...
  score_aaset_start<aaset_selavg_score>(neutral_nsaa_flux,nsaa_flux,is_reachable,os);
}



#if 0
  // over all pairs:
  for(unsigned i(0);i<NSAA::SIZE;++i){
    for(unsigned j(i+1);j<NSAA::SIZE;++j){

#if 0
      if(neutral_nsaa_flux[j+i*NSAA::SIZE] <= 0.) {
        pair_dn_ds[j+i*NSAA::SIZE]=0.;
      } else {
        smlfloat flux(nsaa_flux[j+i*NSAA::SIZE]+nsaa_flux[i+j*NSAA::SIZE]);
        smlfloat nflux(neutral_nsaa_flux[j+i*NSAA::SIZE]+neutral_nsaa_flux[i+j*NSAA::SIZE]);
        pair_dn_ds[j+i*NSAA::SIZE] = flux/nflux;
      }
#endif
#if 0
      for(unsigned k(j+1);k<NSAA::SIZE;++k){

        unsigned nsfoo[] = {i,j,k};
        if(! is_aaset_connected(is_reachable,nsfoo,3)) continue;
        os << "GKEY " << NSAA::syms[i] << " " << NSAA::syms[j] << " " << NSAA::syms[k] << " ";
        os << get_Nset_dn_ds(neutral_nsaa_flux,nsaa_flux,3,nsfoo);
        os << "\n";
      }
#endif
    }
  }
#if 0
  os << "PAIR DN/DS\n";
  matrix_report(pair_dn_ds,NSAA::SIZE,NSAA::syms,os);

  report_nsaa_matrix_full(pair_dn_ds,os,"PAIR DN/DS");
#endif
#endif




static
void
report_nsaa_grouping_flux_average(const irv_t<smlfloat>* neutral_nsaa_flux,
                                  const irv_t<smlfloat>* nsaa_flux,
                                  const char* label,
                                  const unsigned* reduction_map,
                                  const char* syms,
                                  const unsigned N,
                                  std::ostream& os){

  irv_t<smlfloat>* neutral_reduced_flux(new irv_t<smlfloat>[N*N]);
  matrix_state_reduction(neutral_reduced_flux,N,neutral_nsaa_flux,NSAA::SIZE,reduction_map);
#if 0
  os << "NEUTRAL AA FLUX :: SUM BY " << label << ":\n";
  matrix_report(neutral_reduced_flux,N,syms,os);
#endif
  irv_t<smlfloat>* reduced_flux(new irv_t<smlfloat>[N*N]);
  matrix_state_reduction(reduced_flux,N,nsaa_flux,NSAA::SIZE,reduction_map);

  irv_t<smlfloat>* reduced_dn_ds(new irv_t<smlfloat>[N*N]);

  for(unsigned i(0);i<(N*N);++i){
    if(neutral_reduced_flux[i]<=0.) reduced_dn_ds[i] = 0.;
    else                            reduced_dn_ds[i] = reduced_flux[i]/neutral_reduced_flux[i];
  }

  static const unsigned rwid(9);
  static const unsigned rpre(3);

  os << "AA SELECTION :: NSAA FLUX AVERAGE BY " << label << ":\n";
  matrix_report(reduced_dn_ds,N,syms,os,rwid,rpre);

  os << "AA FLUX :: SUM BY " << label << ":\n";
  matrix_report(reduced_flux,N,syms,os,rwid,rpre);

  os << "AA NEUTRAL FLUX :: SUM BY " << label << ":\n";
  matrix_report(neutral_reduced_flux,N,syms,os,rwid,rpre);

  delete [] neutral_reduced_flux;
  delete [] reduced_flux;
  delete [] reduced_dn_ds;
}


static
void
print_section_divider(std::ostream& os){
  os << "######################################################################\n";
}


static
void
print_site_matrix_category_header(const std::string& matrix_category_label,
                                  std::ostream& os){
  print_section_divider(os);
  os << "SELECTION MATRIX CATEGORY: [ " << matrix_category_label << " ]\n";
}




struct ssc_data {
  ssc_data()
    : prior(0.), exp_dn_ds(0.), label(0) {}

  ssc_data(prob_t p,
           smlfloat e,
           const std::string* l)
    : prior(p), exp_dn_ds(e), label(l) {}

  bool operator<(const ssc_data&) const { return false;}

  prob_t prior;
  smlfloat exp_dn_ds;
  const std::string* label;
};




void
rate_gtor_nscodon_base::
report(const char* label,
       std::ostream& os) const {

  // report from mutation model:
  base_t::report(label,os);

  {
    using namespace CODON_BIAS_MODEL;

    const index_t& cbm(_data.copt.codon_bias_model);
    if(cbm == SYNON_RATIO ||
       cbm == SYNON_NORM_RATIO){
      // dump codon bias parameters
      os << "CODON BIAS SELECTION:\n";
      os << std::setprecision(6) << std::fixed;
      for(unsigned i(0);i<NSCODON::SIZE;++i){
        const NSCODON::index_t nsi(static_cast<NSCODON::index_t>(i));
        os << NSCODON::print(nsi) << " : " << get_codon_bias_factor(nsi) << "\n";
      }
      os << "\n";
    }
  }

  const unsigned smc_size(category_size_site_select_matrix());
  prob_t* smc_prior(new prob_t[smc_size]);
  get_typed_cat_dependent_prior_prob(SITE_CAT_TYPE::SEL_MATRIX,smc_prior);

  const unsigned ssc_size(category_size_site_select());
  prob_t* ssc_prior(new prob_t[ssc_size]);
  get_typed_cat_dependent_prior_prob(SITE_CAT_TYPE::SEL_STRENGTH,ssc_prior);

  irv_t<smlfloat> e_dn_ds(0.);
  std::vector<irv_t<smlfloat> > smc_dn_ds(smc_size);
  std::vector<irv_t<smlfloat> > smc_exp_dn_ds(smc_size);
  std::vector<irv_t<smlfloat> > ssc_exp_dn_ds(ssc_size);

  irv_t<smlfloat> nsaa_selection[NSAA::SIZE*NSAA::SIZE];

  irv_t<smlfloat> neutral_nscodon_flux[NSCODON::SIZE*NSCODON::SIZE];
  get_nscodon_flux_no_aa_selection(*this,neutral_nscodon_flux);


  //adjust to reportable mutation rates:
  //
  const smlfloat rate_scale(1./rate_time_bg_scale());
  for(unsigned i(0);i<NSCODON::SIZE*NSCODON::SIZE;++i){
    neutral_nscodon_flux[i] *= rate_scale;
  }

  {
    // calculate expected dN/dS:
    for(unsigned smc(0);smc<smc_size;++smc){
      // get matrix dn_ds first
      get_nsaa_selection_matrix(nsaa_selection,smc);
      matrix_scale(nsaa_selection,NSAA::SIZE,get_group_select_cats_expect());
      // use neutral nscodon to flux build dN/dS for each selection matrix type
      smc_dn_ds[smc] = neutral_nscodon_flux_to_dn_ds(neutral_nscodon_flux,nsaa_selection);
    }

    const smlfloat ssc_norm(get_site_select_cats_expect());

    std::vector<irv_t<smlfloat> > smc_exp_dn_ds_norm(smc_size,0.);
    std::vector<irv_t<smlfloat> > ssc_exp_dn_ds_norm(ssc_size,0.);

    const unsigned n_site_cats(category_size_site());
    prob_t* site_cat_prior_pdistro(new prob_t[n_site_cats]);
    get_site_category_pdistro(site_cat_prior_pdistro);

    for(unsigned sc(0);sc<n_site_cats;++sc){
      const unsigned smc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::SEL_MATRIX,sc));
      const unsigned ssc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::SEL_STRENGTH,sc));

      const irv_t<smlfloat> ssc_val(get_paramset_member(PARAMSET_SITE_SELECT_CATS,ssc)/ssc_norm);
      const irv_t<smlfloat> sc_dn_ds(ssc_val*smc_dn_ds[smc]*site_cat_prior_pdistro[sc]);

      smc_exp_dn_ds[smc] += sc_dn_ds;
      smc_exp_dn_ds_norm[smc] += site_cat_prior_pdistro[sc];
      ssc_exp_dn_ds[ssc] += sc_dn_ds;
      ssc_exp_dn_ds_norm[ssc] += site_cat_prior_pdistro[sc];
      e_dn_ds += sc_dn_ds;
    }

    delete [] site_cat_prior_pdistro;

    for(unsigned smc(0);smc<smc_size;++smc){ smc_exp_dn_ds[smc] /= smc_exp_dn_ds_norm[smc]; }
    for(unsigned ssc(0);ssc<ssc_size;++ssc){ ssc_exp_dn_ds[ssc] /= ssc_exp_dn_ds_norm[ssc]; }
  }

  os.unsetf(std::ios::fixed);
  os << std::setprecision(5);
  os << std::left;
  os << "Expected dN/dS: " << e_dn_ds << "\n\n";

  {
    // summarize select matrix cats with dN/dS
    //
    os << "SITE SELECTION MATRIX CATEGORIES (" << smc_size << ") dN/dS:\n";

    for(unsigned smc(0);smc<smc_size;++smc){
      os << "[ " << _data.site_select_matrix_cats_labels[smc] << " ] : "
         << "Matrix_dN/dS= "
         << std::setw(7) << smc_dn_ds[smc]
         << " expected_dN/dS= "
         << std::setw(7) << smc_exp_dn_ds[smc]
         << " dependent_prior_prob= "
         << std::setw(7) << smc_prior[smc] << "\n";
    }
    os << "\n";
  }

  {
    // dump site selection categories (normalized)
    //
    const smlfloat ssc_norm(get_site_select_cats_expect());

    std::vector<std::pair<smlfloat,ssc_data> > tmp_cats(ssc_size);
    for(unsigned ssc(0);ssc<ssc_size;++ssc){
      //const prob_t independent_ssc_prior(get_paramset_pdistro_member(PARAMSET_SITE_SELECT_CATS_PROB,i));
      tmp_cats[ssc] = std::make_pair(get_paramset_member(PARAMSET_SITE_SELECT_CATS,ssc)/ssc_norm,
                                     ssc_data(ssc_prior[ssc],ssc_exp_dn_ds[ssc],&(_data.site_select_cats_labels[ssc])));
    }
    std::sort(tmp_cats.begin(),tmp_cats.end());

    os << "SITE SELECTION CATEGORIES (" << ssc_size << "):\n";
    for(unsigned ssc(0);ssc<ssc_size;++ssc){
      os << "[ " << *(tmp_cats[ssc].second.label) << " ] scale_factor= "
         << std::setw(7) << tmp_cats[ssc].first
         << " expected_dN/dS= "
         << std::setw(7) << tmp_cats[ssc].second.exp_dn_ds
         << " dependent_prior_prob= "
         << std::setw(7) << tmp_cats[ssc].second.prior << "\n";
    }
    os << "\n";
  }

  delete [] smc_prior;
  delete [] ssc_prior;


  irv_t<smlfloat> neutral_nsaa_flux[NSAA::SIZE*NSAA::SIZE];
  irv_t<smlfloat> nsaa_flux[NSAA::SIZE*NSAA::SIZE];
  irv_t<smlfloat> nsaa_rate[NSAA::SIZE*NSAA::SIZE];

  prob_t bg_pdistro_nsaa[NSAA::SIZE];
  nscodon_pdf_2_nsaa_pdf(bg_pdistro_nscodon(),bg_pdistro_nsaa);

  for(unsigned mc(0);mc<category_size_site_select_matrix();++mc){
    const std::string& mcat_label(_data.site_select_matrix_cats_labels[mc]);

    if(select_model(mc) == SELECT_MODEL::NONE ||
       select_model(mc) == SELECT_MODEL::SINGLE) continue;

    get_nsaa_selection_matrix(nsaa_selection,mc);

    // normalize nsaa for selection categories
    matrix_scale(nsaa_selection,NSAA::SIZE,get_site_select_cats_expect()*get_group_select_cats_expect());

    print_site_matrix_category_header(mcat_label,os);
    report_nsaa_matrix_full(nsaa_selection,os,"AA SELECTION");
    report_nsaa_exchange_averages(nsaa_selection,os,"AA SELECTION");

    {  // average aa selection from the background codon distro and neutral rates
      irv_t<smlfloat> to_avg[NSAA::SIZE];
      irv_t<smlfloat> from_avg[NSAA::SIZE];
      neutral_nscodon_flux_to_nsaa_dn_ds(neutral_nscodon_flux,nsaa_selection,to_avg,from_avg);

      report_paired_aa_pdistro(from_avg,to_avg,os,"AA SELECTION/NSCODON AVG","FROM AVG","TO AVG");
    }


    neutral_nscodon_flux_to_nsaa_flux(neutral_nscodon_flux,nsaa_selection,nsaa_flux);

    print_site_matrix_category_header(mcat_label,os);
    report_nsaa_matrix_full(nsaa_flux,os,"AA FLUX");
    report_nsaa_exchange_sums(nsaa_flux,os,"AA FLUX");

    print_site_matrix_category_header(mcat_label,os);

    std::copy(nsaa_flux,nsaa_flux+NSAA::SIZE*NSAA::SIZE,nsaa_rate);

    /// \todo introduce a category specific background distro:
    flux_to_rates_inplace(NSAA::SIZE,bg_pdistro_nsaa,nsaa_rate);

    report_nsaa_matrix_full(nsaa_rate,os,"AA RATE");

#if 0
    // report the stationary aa distro of each selection matrix, assuming
    // no other category type exists in this model / Also, note this only
    // uses one overall aa bkgd...
    //
    if(category_size() == category_size_site_select_matrix()){
      static const unsigned cell_width(7);
      static const unsigned prec(3);

      os << "\n";

      prob_t stat_nsaa[NSAA::SIZE];
      rates_func_options rf;
      {
        const unsigned nsc(category_size_site());
        for(unsigned sc(0);sc<nsc;++sc){
          unsigned test_scm(get_typed_cat_from_site_cat(SITE_CAT_TYPE::SEL_MATRIX,sc));
          if( test_scm == mc ){
            rf.site_category=sc;
            break;
          }
        }
      }

      stat_pdistro_nsaa(stat_nsaa,rf);

      os << "AAST AA:    ";
      for(unsigned i(0);i<NSAA::SIZE;++i) {
        for(unsigned j(0);j<(cell_width-1);++j) os << " ";
        os << NSAA::syms[i];
      }
      os << "\n";

      os << "AAST bgd:   ";
      os << std::setprecision(2);
      for(unsigned i(0);i<NSAA::SIZE;++i){
        os << " "
           << std::right
           << std::setw(cell_width-1)
           << std::setprecision(prec)
           << bg_pdistro_nsaa[i];
      }
      os << "\n";

      os << "AAST statd: ";
      for(unsigned i(0);i<NSAA::SIZE;++i){
        os << " "
           << std::right
           << std::setw(cell_width-1)
           << std::setprecision(prec)
           << stat_nsaa[i];
      }
      os << "\n";
      os << "\n";
    }
#endif


    nscodon_flux_to_nsaa_flux(neutral_nscodon_flux,neutral_nsaa_flux);

    { // report exchanges from aa-groupings
      using namespace NSAA_MAPS;
      print_site_matrix_category_header(mcat_label,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"LEHNINGER GROUPS",
                                        LEHNINGER::get_reduction_map(),LEHNINGER::syms,LEHNINGER::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"LANGE GROUPS",
                                        LANGE::get_reduction_map(),LANGE::syms,LANGE::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"HP",
                                        HP::get_reduction_map(),HP::syms,HP::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"HPO",
                                        HPO::get_reduction_map(),HPO::syms,HPO::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"BETA BRANCHING",
                                        BETA_BRANCH::get_reduction_map(),BETA_BRANCH::syms,BETA_BRANCH::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"HELIX FORMING",
                                        HELIX_FORM::get_reduction_map(),HELIX_FORM::syms,HELIX_FORM::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"AROMATIC",
                                        AROMATIC::get_reduction_map(),AROMATIC::syms,AROMATIC::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"CYS",
                                        PRO::get_reduction_map(),PRO::syms,PRO::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"GLY",
                                        PRO::get_reduction_map(),PRO::syms,PRO::SIZE,os);

      report_nsaa_grouping_flux_average(neutral_nsaa_flux,nsaa_flux,"PRO",
                                        PRO::get_reduction_map(),PRO::syms,PRO::SIZE,os);
    }

#ifdef REPORT_CLUSTERS
    report_nsaa_sets(neutral_nsaa_flux,nsaa_flux,os);
#endif
  }


  // report and crudely analyze the stationary distro for the model as a whole
  //
  { // sum category stat distros...
    //

    prob_t stat_nscodon[NSCODON::SIZE];
    {
      std::fill(stat_nscodon,stat_nscodon+NSCODON::SIZE,0.);

      const unsigned nsc(category_size_site());
      const unsigned ngc(category_size_group());

      std::vector<prob_t> site_cat_prior_pdistro(nsc);
      std::vector<prob_t> group_cat_prior_pdistro(ngc);

      get_site_category_pdistro(site_cat_prior_pdistro.begin());
      get_group_category_pdistro(group_cat_prior_pdistro.begin());

      prob_t stat_nscodon_cat[NSCODON::SIZE];
      rates_func_options rf;

      for(unsigned gc(0);gc<ngc;++gc){
        for(unsigned sc(0);sc<nsc;++sc){
          const prob_t cat_prob(site_cat_prior_pdistro[sc]*group_cat_prior_pdistro[gc]);

          rf.site_category=sc;
          rf.group_category=gc;
          stat_pdistro_nscodon(stat_nscodon_cat,rf);

          for(unsigned i(0);i<NSCODON::SIZE;++i){
            stat_nscodon[i] += stat_nscodon_cat[i]*cat_prob;
          }
        }
      }
    }

     print_section_divider(os);

    {
      prob_t stat_nuc[NUC::SIZE];
      nscodon_pdf_2_nuc_pdf(stat_nscodon,stat_nuc);

      os << "Category averaged nucleotide stationary distribution:\n\n";

      os << "NUST NUC:    ";
      for(unsigned i(0);i<NUC::SIZE;++i) {
        for(unsigned j(0);j<((pdistro_row_write_width+2)-1);++j) os << " ";
        os << NUC::syms[i];
      }
      os << "\n";

      os << "NUST STAT:   ";
      pdistro_row_write(stat_nuc,NUC::SIZE,os,pdistro_row_write_width+2,5);

      os << "NUST OBS:    ";
      pdistro_row_write(bg_pdistro_nuc(),NUC::SIZE,os,pdistro_row_write_width+2,5);

      os << std::setprecision(7);
      const smlfloat stdot(normalized_dot(stat_nuc,bg_pdistro_nuc(),NUC::SIZE));
      os << "NUST dot(STAT,OBS):  " << stdot << "\n";

      const smlfloat stre(rel_ent(stat_nuc,bg_pdistro_nuc(),NUC::SIZE));
      os << "NUST re(STAT,OBS):   " << stre << "\n";
      os << "\n";
    }

    {
      prob_t stat_nsaa[NSAA::SIZE];
      nscodon_pdf_2_nsaa_pdf(stat_nscodon,stat_nsaa);

      os << "Category averaged amino acid stationary distribution:\n\n";

      os << "AAST AA:    ";
      for(unsigned i(0);i<NSAA::SIZE;++i) {
        for(unsigned j(0);j<(pdistro_row_write_width-1);++j) os << " ";
        os << NSAA::syms[i];
      }
      os << "\n";

      os << "AAST STAT:  ";
      pdistro_row_write(stat_nsaa,NSAA::SIZE,os);

      os << "AAST OBS:   ";
      pdistro_row_write(bg_pdistro_nsaa,NSAA::SIZE,os);

      os << std::setprecision(7);
      const smlfloat stdot(normalized_dot(stat_nsaa,bg_pdistro_nsaa,NSAA::SIZE));
      os << "AAST dot(STAT,OBS):  " << stdot << "\n";

      const smlfloat stre(rel_ent(stat_nsaa,bg_pdistro_nsaa,NSAA::SIZE));
      os << "AAST re(STAT,OBS):   " << stre << "\n";
      os << "\n";
    }

    {
      os << "Category averaged codon stationary distribution:\n\n";

      os << std::setprecision(7);
      const smlfloat stdot(normalized_dot(stat_nscodon,bg_pdistro_nscodon(),NSCODON::SIZE));

      os << "NCST dot(STAT,OBS):  " << stdot << "\n";

      const smlfloat stre(rel_ent(stat_nscodon,bg_pdistro_nscodon(),NSCODON::SIZE));
      os << "NCST re(STAT,OBS):   " << stre << "\n";
      os << "\n";
    }
  }
}
