// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor_nuc_base.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "context_group.h"
#include "rate_gtor_nuc_base.h"
#include "rate_gtor_util.h"
#include "rate_gtor_nuc_base_util.h"
#include "root_gtor.h"
#include "simple_util.h"
#include "subs_ml_ptol.h"
#include "substk_exception.h"
#include "util/bio/bioseq_util_io.h"
#include "util/bio/bioseq_util_flux_rate_conversions.h"
#include "util/general/die.h"
#include "util/general/metatags.h"
#include "util/math/indy_random_var_io.h"
#include "util/math/matrix_util_io.h"
#include "util/math/polynom_util.h"
#include "util/math/prob_util_io.h"
#include "util/math/random_util.h"

#include <cassert>

#include <iostream>
#include <utility>
#include <vector>


static
unsigned
get_rate_model_param_size(const RATE_MODEL_NUC::index_t i){

  using namespace RATE_MODEL_NUC;

  switch(i){
  case JC69:   return 1;
  case K80:    return 2;
  case F81:    return 1;
  case HKY85:  return 2;
  case REV:    return (off_diag_size(NUC::SIZE)/2);
  case NONREV: return off_diag_size(NUC::SIZE);
  default: die("Unsupported nuc rate model");
  }
}



static
unsigned
context_adjust_nuc_param_size(const CONTEXT_MODEL_NUC::index_t i,
                              const unsigned s){

  using namespace CONTEXT_MODEL_NUC;

  switch(i){
  case INDY:             return s;
  case CPG_1TI:          return s+1;
  case CPG_ONLY:
  case TPA_ONLY:
  case CPG_2TI:          return s+2;
  case CPG_NONREV:       return s+2*(NUC::SIZE-1);
  case PRE_DOUBLET:
  case POST_DOUBLET:     return s*NUC::SIZE;
  case FACTORED_TRIPLET: return s*2*NUC::SIZE;
  case TRIPLET:          return s*NUC::SIZE*NUC::SIZE;
  default: die("Unsupported nuc context model");
  }
}



unsigned
rate_gtor_nuc_base::
get_rate_context_model_param_size(const RATE_MODEL_NUC::index_t rm,
                                  const CONTEXT_MODEL_NUC::index_t cm) {
  unsigned model_param_size(get_rate_model_param_size(rm));
  return context_adjust_nuc_param_size(cm,model_param_size);

}



void
rate_gtor_nuc_base::
init_class_storage(){

  _data.site_rate_cats_labels.resize(category_size_site_rate());

  _data.site_mutation_model_cats_labels.resize(category_size_site_mutation_model());
  _data.nopt.set_n_site_mutation_model_cats(category_size_site_mutation_model());

  _data.group_rate_cats_labels.resize(category_size_group_rate());

  set_category_size_group(category_size_group()*
                          category_size_group_rate());

  {
    bool is_all_indy(true);
    for(unsigned i(0);i<category_size_site_mutation_model();++i){
      if(context_model_nuc(i)!=CONTEXT_MODEL_NUC::INDY){
        is_all_indy=false;
      }
    }

    if(is_all_indy) _data.nopt.is_use_edge_correction = false;
  }

  set_param_size(param_end());

  {
    const unsigned st(base_t::param_end());
    const unsigned sts(param_end());
    for(unsigned i(st);i<sts;++i){
      _is_free_block_start[i] = false;
      _is_free_block_stop[i] = false;
    }
  }

  set_paramset_dependent_block(PARAMSET_SITE_RATE_CATS);
  set_paramset_dependent_block(PARAMSET_SITE_RATE_CATS_PROB);
  set_paramset_dependent_block(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB);
  set_paramset_dependent_block(PARAMSET_GROUP_RATE_CATS);
  set_paramset_dependent_block(PARAMSET_GROUP_RATE_CATS_PROB);
  set_paramset_dependent_block(PARAMSET_NUC);
}




void
rate_gtor_nuc_base::
init_cat_labels() {
  static const unsigned tmp_label_size(32);
  char label[tmp_label_size];
  for(unsigned i(0);i<_data.site_rate_cats_labels.size();++i) {
    snprintf(label,tmp_label_size,"%i",i);
    _data.site_rate_cats_labels[i] = label;
  }
  for(unsigned i(0);i<_data.site_mutation_model_cats_labels.size();++i) {
    snprintf(label,tmp_label_size,"%i",i);
    _data.site_mutation_model_cats_labels[i] = label;
  }
  for(unsigned i(0);i<_data.group_rate_cats_labels.size();++i) {
    snprintf(label,tmp_label_size,"%i",i);
    _data.group_rate_cats_labels[i] = label;
  }
}




void
rate_gtor_nuc_base::
init_param_locks(const rate_gtor_nuc_lock_options& lopt){

  // lock parameters which won't be trained
  if(lopt.is_lock_site_rate_cats || category_size_site_rate()==1){
    set_paramset_train(PARAMSET_SITE_RATE_CATS,false);
  }
  if(lopt.is_lock_site_rate_cats_prob || category_size_site_rate()==1){
    set_paramset_train(PARAMSET_SITE_RATE_CATS_PROB,false);
  }
  if(lopt.is_lock_site_mutation_model_cats_prob || category_size_site_mutation_model()==1){
    set_paramset_train(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB,false);
  }

  if(lopt.is_lock_group_rate_cats || category_size_group_rate()==1){
    set_paramset_train(PARAMSET_GROUP_RATE_CATS,false);
  }
  if(lopt.is_lock_group_rate_cats_prob || category_size_group_rate()==1){
    set_paramset_train(PARAMSET_GROUP_RATE_CATS_PROB,false);
  }
#if 0
  // disabled this, b/c can't find a way to set val correctly first
  { // if jukes rate model, then lock nuc param
    const unsigned st(paramset_start(PARAMSET_NUC));
    const unsigned s(paramset_size(PARAMSET_NUC));
    if(s==1){
      _is_train_param[st]=false;
    }
  }
#endif
}




void
rate_gtor_nuc_base::
rates(smlfloat* rts,
      const rates_func_options& opt) const {
  rates_nuc_context(*this,rts,opt);
}



static
smlfloat
get_nuc_rate_matrix_scale(const smlfloat* nrate,
                          const smlfloat* nbg){
  smlfloat sum(0.);
  for(unsigned i(0);i<NUC::SIZE;++i)
    for(unsigned j(0);j<NUC::SIZE;++j)
      if(i!=j) sum += nrate[j+i*NUC::SIZE]*nbg[i];

  return sum;
}



smlfloat
rate_gtor_nuc_base::
rate_time_context_bg_scale(const NUC::index_t pre_n,
                           const NUC::index_t post_n) const {

  smlfloat nr[NUC::SIZE*NUC::SIZE];
  smlfloat scale(0.);

  for(unsigned gc(0);gc<category_size_group();++gc){
    for(unsigned sc(0);sc<category_size_site();++sc){

      rates_nuc(nr,sc,gc,pre_n,post_n);
      const smlfloat cat_scale(get_nuc_rate_matrix_scale(nr,bg_pdistro_nuc()));
      const prob_t cat_p(get_site_category_proportional_prob(sc)*get_group_category_prob(gc));
      scale += cat_scale*cat_p;
    }
  }

  return scale;
}



/// \todo change to use first order nucleotide dependency info for
/// codon and 4mer case.
///
smlfloat
rate_gtor_nuc_base::
rate_time_context_avg_bg_scale() const {

  smlfloat scale(0.);

  for(unsigned n0(0);n0<NUC::SIZE;++n0){
    for(unsigned n2(0);n2<NUC::SIZE;++n2){
      const smlfloat context_scale(rate_time_context_bg_scale(static_cast<NUC::index_t>(n0),
                                                              static_cast<NUC::index_t>(n2)));
      const prob_t context_p(bg_pdistro_nuc()[n0]*bg_pdistro_nuc()[n2]);
      scale += context_scale*context_p;
    }
  }

  return scale;
}


#if 0
smlfloat
rate_gtor_nuc_base::
rates_nuc_context_bg_scale(const NUC::index_t pre_n,
                           const NUC::index_t post_n) const {

  smlfloat nr[NUC::SIZE*NUC::SIZE];

  for(unsigned n1(0);n1<NUC::SIZE;++n1){
    for(unsigned n2(0);n2<NUC::SIZE;++n2){
      if(n1 == n2) continue;

      const smlfloat current_val(get_category_nuc_context_mut_rate(pre_n ,static_cast<NUC::index_t>(n1),
                                                                   post_n,static_cast<NUC::index_t>(n2)));
      nr[n2+n1*NUC::SIZE] = current_val;
    }
  }

  return get_nuc_rate_matrix_scale(nr,bg_pdistro_nuc());
}
#endif


smlfloat
rate_gtor_nuc_base::
rates_nuc_norm(const unsigned mut_model_cat,
               const trainable_mode::index_t tm) const {
  // target expectation for rate matrix offdiag sum is NUC::SIZE,
  // implicitly works from a uniform background distro at the mutation
  // site itself
  //
  smlfloat sum(0.);
  for(unsigned n1(0);n1<NUC::SIZE;++n1){
    for(unsigned n2(0);n2<NUC::SIZE;++n2){
      if(n1 == n2) continue;
      sum += get_nuc_context_mut_rate(NUC::N,static_cast<NUC::index_t>(n1),
                                      NUC::N,static_cast<NUC::index_t>(n2),mut_model_cat,
                                      bg_pdistro_nuc(),bg_pdistro_nuc(),tm);
    }
  }

  return sum/static_cast<smlfloat>(NUC::SIZE);
}



void
rate_gtor_nuc_base::
stat_pdistro_nuc(prob_t* nuc_sd) const {
  prob_t* sd(new prob_t[state_size()]);
  get_stationary_pdistro(sd,*this);
  state_pdistro_to_nuc_pdistro(sd,nuc_sd);
  delete [] sd;
}


void
rate_gtor_nuc_base::
rates_nuc(smlfloat rts[NUC::SIZE*NUC::SIZE],
          const unsigned site_category,
          const unsigned group_category,
          const NUC::index_t pre_n,
          const NUC::index_t post_n) const {

  for(unsigned n1(0);n1<NUC::SIZE;++n1){
    for(unsigned n2(0);n2<NUC::SIZE;++n2){
      if(n1 == n2) continue;
      const smlfloat current_val(get_category_nuc_context_mut_rate(pre_n,static_cast<NUC::index_t>(n1),
                                                                   post_n,static_cast<NUC::index_t>(n2),
                                                                   site_category,group_category,
                                                                   bg_pdistro_nuc(),bg_pdistro_nuc()));
      rts[n2+n1*NUC::SIZE] = current_val;
    }
  }
  matrix_balance_diagonal(rts,NUC::SIZE);
}



void
rate_gtor_nuc_base::
fix_input_param_cat_paramset(const paramset_t cats,
                             const paramset_t cats_prob){

  // fix cats prior distro:
  set_paramset_pdistro_norm_free_param(cats_prob);

  // run uncompensated normalization on free parameters in cat paramset
  const unsigned st(paramset_start(cats_prob));
  const unsigned st2(paramset_start(cats));
  const unsigned s2(paramset_size(cats));

  const smlfloat scale =
    array_scale_expectation_selected_only(_param+st2,
                                          _param+st2+s2,
                                          _param+st,
                                          _is_train_param.begin()+st2,
                                          1.);
  if(scale<0) {
    throw substk_exception("rate_gtor_nuc_base: locked parameter is out of range");
  }
}



void
rate_gtor_nuc_base::
fix_input_param(){

  base_t::fix_input_param();

  fix_input_param_cat_paramset(PARAMSET_SITE_RATE_CATS,PARAMSET_SITE_RATE_CATS_PROB);
  fix_input_param_cat_paramset(PARAMSET_GROUP_RATE_CATS,PARAMSET_GROUP_RATE_CATS_PROB);

  set_paramset_pdistro_norm_free_param(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB);

  smlfloat mut_model_scale(0.);
  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    // run uncompensated normalization on each category of mutation rate parameters:

    // take into consideration which parameters are locked to
    // calculate normalization scale:
    //
    static const smlfloat expect(1.);
    const smlfloat r_all(rates_nuc_norm(i,trainable_mode::NONE));
    const smlfloat r_full(rates_nuc_norm(i,trainable_mode::FULL));
    const smlfloat r_half(rates_nuc_norm(i,trainable_mode::HALF_JOINT));

    const smlfloat r_lock(r_all - ( r_full + r_half ));

    smlfloat s1,s2;
    solve_quadratic(r_full,r_half,r_lock-expect,s1,s2);

    const smlfloat cat_scale(s1*s1);
    const prob_t cat_p(get_paramset_pdistro_member(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB,i));

    mut_model_scale += cat_scale*cat_p;
  }

  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    scale_nuc_mut_rate_trainable_param(i,mut_model_scale);
  }

#if 0
  std::cerr << "rate norm before,full,half,scale,s1,s2: " << r_all << " " << r_full << " " << r_half << " " << s1 << " " << s2 << "\n";
  const smlfloat r2 = rates_nuc_norm(i);
    std::cerr << "rate norm after: " << r2 << "\n";
#endif
}



void
rate_gtor_nuc_base::
paramset_label(const paramset_t p,
               const unsigned n,
               std::ostream& os) const {

  switch(p){
  case PARAMSET_SITE_RATE_CATS :
    os << "site_rate_cats_" << n;
    break;
  case PARAMSET_SITE_RATE_CATS_PROB :
    os << "site_rate_cats_prior_" << n;
    break;
  case PARAMSET_SITE_MUTATION_MODEL_CATS_PROB :
    os << "site_mutation_model_prior_" << n;
    break;
  case PARAMSET_GROUP_RATE_CATS :
    os << "group_rate_cats_" << n;
    break;
  case PARAMSET_GROUP_RATE_CATS_PROB :
    os << "group_rate_cats_prior_" << n;
    break;
  case PARAMSET_NUC :
    {
      bool is_valid(false);
      unsigned mut_model_no(0);
      unsigned model_param_no(0);
      {
        const unsigned nsmm(category_size_site_mutation_model());
        int ncopy(n);
        for(unsigned i(0);i<nsmm;++i){
          const int s(paramset_mutation_model_cat_size(i));
          if(ncopy<s){
            if(ncopy>=0){
              mut_model_no=i;
              model_param_no=ncopy;
              is_valid=true;
            }
            break;
          }
          ncopy -= s;
        }
      }
      if(! is_valid) die("invalid call to paramset_label");
      get_nuc_param_index_name(mut_model_no,model_param_no,os);
    }
    break;
  case PARAMSET_EDGE_STRENGTH :
    os << "edge_strength_" << n;
    break;
  default :
    os << "unknown_"  << n;
  }
}



const char* const section_id = "rate_gtor_nuc_base";
const char* const end_label = "END";

void
rate_gtor_nuc_base::
store_state(std::ostream& os) const {

  base_t::store_state(os);

  os << section_id << " n_group_rate_cats " << category_size_group_rate() << "\n";
  os << section_id << " is_use_edge_correction " << is_use_edge_correction() << "\n";

  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    os << section_id << " rate_model_" << i << " " << rate_model_nuc(i) << "\n";
  }
  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    os << section_id << " context_model_" << i << " " << context_model_nuc(i) << "\n";
  }

  // write site rate labels:
  for(unsigned i(0);i<category_size_site_rate();++i){
    os << section_id << " site_rate_cat_label_" << i << " " << _data.site_rate_cats_labels[i] << "\n";
  }
  // write site model labels:
  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    os << section_id << " site_mutation_model_label_" << i << " " << _data.site_mutation_model_cats_labels[i] << "\n";
  }
  // write group rate labels:
  for(unsigned i(0);i<category_size_group_rate();++i){
    os << section_id << " group_rate_cat_label_" << i << " " << _data.group_rate_cats_labels[i] << "\n";
  }

  // write all paramsets:
  //
  for(unsigned p(0);p<PARAMSET_END;++p){
    const unsigned st=paramset_start(static_cast<paramset_t>(p));
    const unsigned s=paramset_size(static_cast<paramset_t>(p));
    for(unsigned i(0);i<s;++i){
      os << section_id << " ";
      paramset_label(static_cast<paramset_t>(p),i,os);
      os << " " << _is_train_param[st+i] << " " << _param[st+i] << "\n";
    }
  }

  os << section_id << " " << end_label << "\n";
}



static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid model file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  std::cerr << is.rdbuf();
  abort();
}




static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(tmpstr != section_id) load_die(is,section_id,tmpstr);
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}




void
rate_gtor_nuc_base::
load_state(std::istream& is) {

  base_t::load_state(is);


  load_advance(is,"n_group_rate_cats");
  is >> _data.nopt.n_group_rate_cats;
  load_advance(is,"is_use_edge_correction");
  is >> _data.nopt.is_use_edge_correction;

  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    std::ostringstream oss;
    oss << "rate_model_" << i;
    load_advance(is,oss.str().c_str());
    int tmpi;
    is >> tmpi;
    _data.nopt.set_rate_model(i,static_cast<RATE_MODEL_NUC::index_t>(tmpi));
  }

  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    std::ostringstream oss;
    oss << "context_model_" << i;
    load_advance(is,oss.str().c_str());
    int tmpi;
    is >> tmpi;
    _data.nopt.set_context_model(i,static_cast<CONTEXT_MODEL_NUC::index_t>(tmpi));
  }

  init_class_storage();

  // read site rate labels
  for(unsigned i(0);i<category_size_site_rate();++i){
    load_advance(is);
    is >> _data.site_rate_cats_labels[i];
  }

  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    load_advance(is);
    is >> _data.site_mutation_model_cats_labels[i];
  }

  // read group rate labels
  for(unsigned i(0);i<category_size_group_rate();++i){
    load_advance(is);
    is >> _data.group_rate_cats_labels[i];
  }

  // read all paramsets:
  //
  for(unsigned p(0);p<PARAMSET_END;++p){
    const unsigned st=paramset_start(static_cast<paramset_t>(p));
    const unsigned s=paramset_size(static_cast<paramset_t>(p));
    for(unsigned i(0);i<s;++i){
      load_advance(is);
      bool tmpb;
      is >> tmpb >> _param[st+i];
      _is_train_param[st+i] = tmpb;
    }
  }

  load_advance(is,end_label);

  self_t::fix_input_param();
}




// helper func to set_ts_tv_ratio
void
rate_gtor_nuc_base::
set_ts_tv_ratio_nhood(const unsigned mut_model_cat,
                      const smlfloat ratio,
                      const NUC::index_t n0,
                      const NUC::index_t n2){
  for(unsigned n1(0);n1<NUC::SIZE;++n1){
    for(unsigned to(0);to<NUC::SIZE;++to){
      if(n1==to) continue;

      const NUC::index_t n1x = static_cast<NUC::index_t>(n1);
      const NUC::index_t tox = static_cast<NUC::index_t>(to);
      if(context_model_nuc(mut_model_cat) == CONTEXT_MODEL_NUC::FACTORED_TRIPLET){
        const int i1 = get_nuc_context_mut_rate_param_index(n0,n1x,n2,tox,mut_model_cat,JOINT1);
        if(i1>=0) _param[i1] = ( NUC::is_transition(n1x,tox) ? std::sqrt(ratio) : 1. );

        const int i2 = get_nuc_context_mut_rate_param_index(n0,n1x,n2,tox,mut_model_cat,JOINT2);
        if(i2>=0) _param[i2] = ( NUC::is_transition(n1x,tox) ? std::sqrt(ratio) : 1. );

      } else {
        const int i = get_nuc_context_mut_rate_param_index(n0,n1x,n2,tox,mut_model_cat);
        if(i>=0) _param[i] = ( NUC::is_transition(n1x,tox) ? ratio : 1. );
      }
    }
  }
}


void
rate_gtor_nuc_base::
set_ts_tv_ratio(const unsigned mut_model_cat,
                const smlfloat ratio){

  const unsigned pst(paramset_mutation_model_cat_start(mut_model_cat));
  const unsigned ps(paramset_mutation_model_cat_size(mut_model_cat));
  for(unsigned i(pst);i<(pst+ps);++i)  _param[i] = 1;

  using namespace CONTEXT_GROUP;
  const index_t cg(get_context_model_group(context_model_nuc(mut_model_cat)));

  if       (cg == NONE){
    set_ts_tv_ratio_nhood(mut_model_cat,ratio,NUC::N,NUC::N);
  } else if(cg == ONLY_PRE){
    for(unsigned n0(0);n0<NUC::SIZE;++n0){
      set_ts_tv_ratio_nhood(mut_model_cat,ratio,
                            static_cast<NUC::index_t>(n0),
                            NUC::N);
    }
  } else if(cg == ONLY_POST){
    for(unsigned n2(0);n2<NUC::SIZE;++n2){
      set_ts_tv_ratio_nhood(mut_model_cat,ratio,
                            NUC::N,
                            static_cast<NUC::index_t>(n2));
    }
  } else if(cg == TWOWAY){
    for(unsigned n0(0);n0<NUC::SIZE;++n0){
      for(unsigned n2(0);n2<NUC::SIZE;++n2){
        set_ts_tv_ratio_nhood(mut_model_cat,ratio,
                              static_cast<NUC::index_t>(n0),
                              static_cast<NUC::index_t>(n2));
      }
    }
  } else { die("Unsupported nuc context model"); }
}


/// non-reversible param coding:
///    A  C  G  T
/// A  -  0  1  2
/// C  3  -  4  5
/// G  6  7  -  8
/// T  9 10 11  -
///
/// reversible param coding:
///    A  C  G  T
/// A  -  0  1  2
/// C  0  -  3  4
/// G  1  3  -  5
/// T  2  4  5  -
///


//          i2
//        0 1 2
//      +------
//    0 | - 0 1
// i1 1 | - - 2
//    2 | - - -
//
static
unsigned
symm_matrix_nodiag_index(unsigned i1,
                             unsigned i2,
                             const unsigned size){

  assert(i1!=i2 && i1<size && i2<size);
  if(i2<i1) std::swap(i2,i1);
  return (i2-1-i1)+(size-1)*i1-(i1*(i1-1))/2;
}
#if 0
static
void
symm_matrix_nodiag_index_inverse(const unsigned i,
                                 const unsigned size,
                                 unsigned& i1,
                                 unsigned& i2){

  assert(i<(off_diag_size(size)/2));
  abort(); /// \todo finish this
}
#endif



//          i2
//        0 1 2
//      +------
//    0 | - 0 1
// i1 1 | 2 - 3
//    2 | 4 5 -
//
static
unsigned
square_matrix_nodiag_index(const unsigned i1,
                           const unsigned i2,
                           const unsigned size){

  assert(i1!=i2 && i1<size && i2<size);
  return i2+(i2<i1?0:-1)+(size-1)*i1;
}

static
void
square_matrix_nodiag_index_inverse(const unsigned i,
                                   const unsigned size,
                                   unsigned& i1,
                                   unsigned& i2){

  assert(i<off_diag_size(size));
  i1=i/(size-1);
  i2=i%(size-1);
  if(i1<=i2) i2 += 1;
}



static
void
get_nuc_param_index_name_nocontext(const RATE_MODEL_NUC::index_t rmn,
                                   const unsigned param_index,
                                   std::ostream& os) {

  {
    using namespace RATE_MODEL_NUC;

    assert(param_index<get_rate_model_param_size(rmn));

    if       (rmn == JC69 || rmn == F81){
      os << "nuc_mut";
    } else if(rmn == K80 || rmn == HKY85){
      if(param_index==1){ os << "transition"; }
      else              { os << "transversion"; }
    } else if(rmn == REV){
      os << "X1<->X2";
    } else if(rmn == NONREV){
      unsigned n1,n2;
      square_matrix_nodiag_index_inverse(param_index,NUC::SIZE,n1,n2);
      os << NUC::syms[n1] << "->" << NUC::syms[n2];
    } else {
      die("Unsupported nuc rate model");
    }
    os << "_rate";
  }
}




// produce a human readable name for the nucleotide parameters
//
void
rate_gtor_nuc_base::
get_nuc_param_index_name(const unsigned mut_model_cat_no,
                         const unsigned param_index,
                         std::ostream& os) const {

  os << "nuc_mutation_model_" << mut_model_cat_no << "_param_";

  const CONTEXT_MODEL_NUC::index_t cmn(context_model_nuc(mut_model_cat_no));
  const RATE_MODEL_NUC::index_t rmn(rate_model_nuc(mut_model_cat_no));

  assert(param_index<get_rate_context_model_param_size(rmn,cmn));

  {
    static const char cpg_tag[] = "CpG";

    using namespace CONTEXT_MODEL_NUC;

    if       (cmn == INDY){
      get_nuc_param_index_name_nocontext(rmn,param_index,os);
      os << "_indy_context";
    } else if(cmn == PRE_DOUBLET){
      get_nuc_param_index_name_nocontext(rmn,param_index/NUC::SIZE,os);
      const unsigned from0(param_index%NUC::SIZE);
      os << "_" << NUC::syms[from0] << "N_context";
    } else if(cmn == POST_DOUBLET){
      get_nuc_param_index_name_nocontext(rmn,param_index/NUC::SIZE,os);
      const unsigned from2(param_index%NUC::SIZE);
      os << "_N" << NUC::syms[from2] << "_context";
    } else if(cmn == FACTORED_TRIPLET){
      get_nuc_param_index_name_nocontext(rmn,param_index/(NUC::SIZE+NUC::SIZE),os);
      os << "_xx_context";
    } else if(cmn == TRIPLET){
      get_nuc_param_index_name_nocontext(rmn,param_index/(NUC::SIZE*NUC::SIZE),os);
      const unsigned tmpu(param_index%(NUC::SIZE*NUC::SIZE));
      const unsigned from0(tmpu%NUC::SIZE);
      const unsigned from2(tmpu/NUC::SIZE);
      os << "_" << NUC::syms[from0] << "N" << NUC::syms[from2] <<"_context";
    } else if(cmn == CPG_ONLY || cmn == TPA_ONLY){
      static const char tpa_tag[] = "TpA";

      const char* mut_tag(cmn == CPG_ONLY ? cpg_tag : tpa_tag);

      const unsigned psn(paramset_mutation_model_cat_size(mut_model_cat_no));
      if     (param_index == psn-2){ os << mut_tag << "_transversion_rate"; }
      else if(param_index == psn-1){ os << mut_tag << "_transition_rate"; }
      else { get_nuc_param_index_name_nocontext(rmn,param_index,os); }
    } else if(cmn == CPG_1TI){

      const unsigned psn(paramset_mutation_model_cat_size(mut_model_cat_no));
      if(param_index == psn-1){ os << cpg_tag << "_transition_rate"; }
      else { get_nuc_param_index_name_nocontext(rmn,param_index,os); }
    } else if(cmn == CPG_2TI){

      const unsigned psn(paramset_mutation_model_cat_size(mut_model_cat_no));
      if     (param_index == psn-2){ os << cpg_tag << "_C->T_transition_rate"; }
      else if(param_index == psn-1){ os << cpg_tag << "_G->A_transition_rate"; }
      else { get_nuc_param_index_name_nocontext(rmn,param_index,os); }
    } else if(cmn == CPG_NONREV){
      const unsigned psn(paramset_mutation_model_cat_size(mut_model_cat_no));
      os << cpg_tag;
      if     (param_index == psn-1){ os << "_G->A_rate"; }
      else if(param_index == psn-2){ os << "_G->C_rate"; }
      else if(param_index == psn-3){ os << "_G->T_rate"; }
      else if(param_index == psn-4){ os << "_C->A_rate"; }
      else if(param_index == psn-5){ os << "_C->G_rate"; }
      else if(param_index == psn-6){ os << "_C->T_rate"; }

      else { get_nuc_param_index_name_nocontext(rmn,param_index,os); }



    } else { die("Unsupported nuc context model"); }
  }

  os << "_" << param_index;
}




static
unsigned
get_rate_model_nuc_mut_rate_param_index(const RATE_MODEL_NUC::index_t i,
                                        const NUC::index_t from,
                                        const NUC::index_t to){

  using namespace RATE_MODEL_NUC;

  switch(i){
  case JC69:
  case F81:    return 0;
  case K80:
  case HKY85:
    if( NUC::is_transition(from,to) ) return 1;
    else                              return 0;
  case REV:    return symm_matrix_nodiag_index(from,to,NUC::SIZE);
  case NONREV: return square_matrix_nodiag_index(from,to,NUC::SIZE);
  default: die("Unsupported nuc rate model");
  }
}




// designed for context-dependent case, positions
// are numbered 012 and nucleotide change occurs
// at position 1
//
unsigned
rate_gtor_nuc_base::
get_nuc_context_mut_rate_param_index(const NUC::index_t from0,
                                     const NUC::index_t from1,
                                     const NUC::index_t from2,
                                     const NUC::index_t to,
                                     const unsigned mut_model_cat,
                                     const joint_param_t jp) const {

  if( from1 == to || from1 == NUC::N || to == NUC::N) {
    die("Invalid parameters for rate model");
  }

  const unsigned psn(paramset_mutation_model_cat_size(mut_model_cat));
  unsigned param_index(get_rate_model_nuc_mut_rate_param_index(rate_model_nuc(mut_model_cat),from1,to));

  { // modify param_index according to context:

    using namespace CONTEXT_MODEL_NUC;
    const index_t i(context_model_nuc(mut_model_cat));

    if((from0 == NUC::N) &&
       ((i == PRE_DOUBLET) ||
        (i == FACTORED_TRIPLET && jp == JOINT1) ||
        (i == TRIPLET))) die("get_nuc_mut_rate_param_index: invalid from0 nuc");

    if((from2 == NUC::N) &&
       ((i == POST_DOUBLET) ||
        (i == FACTORED_TRIPLET && jp == JOINT2) ||
        (i == TRIPLET))) die("get_nuc_mut_rate_param_index: invalid from2 nuc");

    if       (i == INDY){
      // pass
    } else if(i == PRE_DOUBLET){
      param_index = param_index*(NUC::SIZE)+from0;
    } else if(i == POST_DOUBLET){
      param_index = param_index*(NUC::SIZE)+from2;
    } else if(i == FACTORED_TRIPLET){
      if(jp == JOINT1) { param_index = param_index*(NUC::SIZE+NUC::SIZE)+from0; }
      else             { param_index = param_index*(NUC::SIZE+NUC::SIZE)+(NUC::SIZE+from2); }
    } else if(i == TRIPLET){
      param_index = param_index*(NUC::SIZE*NUC::SIZE)+(from2*NUC::SIZE+from0);
    } else if(i == CPG_ONLY){
      if((from0 == NUC::C && from1 == NUC::G) ||
         (from1 == NUC::C && from2 == NUC::G)) {
        if (NUC::is_transition(from1,to)) { param_index = psn-1; }
        else                              { param_index = psn-2; }
      }
    } else if(i == TPA_ONLY){
      if((from0 == NUC::T && from1 == NUC::A) ||
         (from1 == NUC::T && from2 == NUC::A)) {
        if (NUC::is_transition(from1,to)) { param_index = psn-1; }
        else                              { param_index = psn-2; }
      }
    } else if(i == CPG_1TI) {
      if((from0 == NUC::C && from1 == NUC::G) ||
         (from1 == NUC::C && from2 == NUC::G)) {
        if(NUC::is_transition(from1,to)){ param_index = psn-1; }
      }
    } else if(i == CPG_2TI) {
      if(NUC::is_transition(from1,to)){
        if     (from0 == NUC::C && from1 == NUC::G){ param_index = psn-1; }
        else if(from1 == NUC::C && from2 == NUC::G){ param_index = psn-2; }
      }
    } else if(i == CPG_NONREV) {
        if       (from0 == NUC::C && from1 == NUC::G){
          if     (to == NUC::A){ param_index = psn-1; }
          else if(to == NUC::C){ param_index = psn-2; }
          else if(to == NUC::T){ param_index = psn-3; }
        } else if(from1 == NUC::C && from2 == NUC::G){
          if     (to == NUC::A){ param_index = psn-4; }
          else if(to == NUC::G){ param_index = psn-5; }
          else if(to == NUC::T){ param_index = psn-6; }
        }

    } else { die("Unsupported nuc context model"); }
  }

  assert(param_index < psn);
  return param_index+paramset_mutation_model_cat_start(mut_model_cat);
}



// get nucleotide parameter from NUC codes... this function layer is
// mainly here to handle the factored parameter case:
//
irv_t<smlfloat>
rate_gtor_nuc_base::
get_nuc_context_mut_rate_param(const NUC::index_t from0,
                               const NUC::index_t from1,
                               const NUC::index_t from2,
                               const NUC::index_t to,
                               const unsigned mut_model_cat,
                               const trainable_mode::index_t tm) const {

  const CONTEXT_GROUP::index_t cg(get_context_model_group(context_model_nuc(mut_model_cat)));

  if(((cg == CONTEXT_GROUP::ONLY_PRE  || cg == CONTEXT_GROUP::TWOWAY) && (from0 == NUC::N)) ||
     ((cg == CONTEXT_GROUP::ONLY_POST || cg == CONTEXT_GROUP::TWOWAY) && (from2 == NUC::N))){
    die("function requires known context");
  }

  if(context_model_nuc(mut_model_cat) == CONTEXT_MODEL_NUC::FACTORED_TRIPLET){
    const unsigned param_index1(get_nuc_context_mut_rate_param_index(from0,from1,from2,to,mut_model_cat,JOINT1));
    const irv_t<smlfloat> param1(_param[param_index1],param_variance(param_index1));

    const unsigned param_index2(get_nuc_context_mut_rate_param_index(from0,from1,from2,to,mut_model_cat,JOINT2));
    const irv_t<smlfloat> param2(_param[param_index2],param_variance(param_index2));

    const bool is_full_joint(_is_train_param[param_index1] && _is_train_param[param_index2]);
    const bool is_half_joint((!is_full_joint) &&
                             (_is_train_param[param_index1] || _is_train_param[param_index2]));

    const bool is_skip((tm == trainable_mode::HALF_JOINT && (! is_half_joint)) ||
                       (tm == trainable_mode::FULL && (! is_full_joint)));

    if(is_skip){
      return irv_t<smlfloat>(0.);
    } else {
      return param1*param2;
    }

  } else {
    const unsigned param_index(get_nuc_context_mut_rate_param_index(from0,from1,from2,to,mut_model_cat));
    const bool is_skip((tm == trainable_mode::HALF_JOINT) ||
                       (tm == trainable_mode::FULL && (! _is_train_param[param_index])));

    if(is_skip){
      return irv_t<smlfloat>(0.);
    } else {
      return irv_t<smlfloat>(_param[param_index],param_variance(param_index));
    }
  }
}



///
//
irv_t<smlfloat>
rate_gtor_nuc_base::
get_nuc_context_mut_rate(const NUC::index_t from0,
                         const NUC::index_t from1,
                         const NUC::index_t from2,
                         const NUC::index_t to,
                         const unsigned mut_model_cat,
                         const prob_t* nuc_bg0,
                         const prob_t* nuc_bg2,
                         const trainable_mode::index_t tm) const {

  const CONTEXT_GROUP::index_t cg(get_context_model_group(context_model_nuc(mut_model_cat)));

  if       ((cg == CONTEXT_GROUP::ONLY_PRE || cg == CONTEXT_GROUP::TWOWAY) && (from0 == NUC::N)){
    if(nuc_bg0==0) die("unspecified edge nuc distro");
    irv_t<smlfloat> param(0);
    for(unsigned i(0);i<NUC::SIZE;++i){
      const irv_t<smlfloat> ptmp = get_nuc_context_mut_rate(static_cast<NUC::index_t>(i),from1,from2,to,mut_model_cat,nuc_bg0,nuc_bg2,tm);
      param += ptmp*nuc_bg0[i];
    }
    return param;

  } else if((cg == CONTEXT_GROUP::ONLY_POST || cg == CONTEXT_GROUP::TWOWAY) && (from2 == NUC::N)){
    if(nuc_bg2==0) die("unspecified edge nuc distro");
    irv_t<smlfloat> param(0);
    for(unsigned i(0);i<NUC::SIZE;++i){
      const irv_t<smlfloat> ptmp = get_nuc_context_mut_rate(from0,from1,static_cast<NUC::index_t>(i),to,mut_model_cat,nuc_bg0,nuc_bg2,tm);
      param += ptmp*nuc_bg2[i];
    }
    return param;

  } else {
    irv_t<smlfloat> param(get_nuc_context_mut_rate_param(from0,from1,from2,to,mut_model_cat,tm));

    {
      using namespace RATE_MODEL_NUC;
      const index_t rmn(rate_model_nuc(mut_model_cat));
      if(rmn == F81 || rmn == HKY85 || rmn == REV){
        return param*bg_pdistro_nuc()[to];
      } else {
        return param;
      }
    }
  }
}



smlfloat
rate_gtor_nuc_base::
get_category_nuc_context_mut_rate(const NUC::index_t from0,
                                  const NUC::index_t from1,
                                  const NUC::index_t from2,
                                  const NUC::index_t to,
                                  const unsigned site_category,
                                  const unsigned group_category,
                                  const prob_t* nuc_bg0,
                                  const prob_t* nuc_bg2) const {

  /// \todo deal with -1 site cat by averaging all mut_models:
  unsigned mmc(0);
  mmc=get_typed_cat_from_site_cat(SITE_CAT_TYPE::MUT_MODEL,site_category);

  smlfloat val(get_nuc_context_mut_rate(from0,from1,from2,to,mmc,nuc_bg0,nuc_bg2));

  val *= get_paramset_member(PARAMSET_SITE_RATE_CATS,
                             get_typed_cat_from_site_cat(SITE_CAT_TYPE::MUT_RATE,site_category));
  val *= get_paramset_member(PARAMSET_GROUP_RATE_CATS,
                             parse_group_rate_category(group_category));

  return val;
}



/// \todo eventually, this method should never touch locked parameters,
/// they should be initialized together with init_param_locks
///
void
rate_gtor_nuc_base::
reset_param(const DENOVO_MODEL::index_t dm,
            const bool is_protected_call) {

  base_t::reset_param(dm,true);

  if(dm == DENOVO_MODEL::RANDOM){
    static const smlfloat rate_cats_min = 0.1;
    static const smlfloat rate_cats_range = 2.;
    static const smlfloat rcprob_unorm_min = 1e-3;

    {
      const unsigned st(paramset_start(PARAMSET_SITE_RATE_CATS));
      const unsigned s(paramset_size(PARAMSET_SITE_RATE_CATS));
      for(unsigned i(st);i<(st+s);++i){
        if(_is_train_param[i]) { _param[i] = random_uniform()*rate_cats_range+rate_cats_min; }
        else                   { _param[i] = 1.; }
      }
    }
    {
      const unsigned st(paramset_start(PARAMSET_SITE_RATE_CATS_PROB));
      const unsigned s(paramset_size(PARAMSET_SITE_RATE_CATS_PROB));
      for(unsigned i(st);i<(st+s);++i){
        if(_is_train_param[i]) { _param[i] = random_uniform()+rcprob_unorm_min; }
        else                   { _param[i] = 1./static_cast<smlfloat>(category_size_site_rate()); }
      }
    }
    {
      const unsigned st(paramset_start(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB));
      const unsigned s(paramset_size(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB));
      for(unsigned i(st);i<(st+s);++i){
        if(_is_train_param[i]) { _param[i] = random_uniform()+rcprob_unorm_min; }
        else                   { _param[i] = 1./static_cast<smlfloat>(category_size_site_mutation_model()); }
      }
    }
    {
      const unsigned st(paramset_start(PARAMSET_GROUP_RATE_CATS));
      const unsigned s(paramset_size(PARAMSET_GROUP_RATE_CATS));
      for(unsigned i(st);i<(st+s);++i){
        if(_is_train_param[i]) { _param[i] = random_uniform()*rate_cats_range+rate_cats_min; }
        else                   { _param[i] = 1.; }
      }
    }
    {
      const unsigned st(paramset_start(PARAMSET_GROUP_RATE_CATS_PROB));
      const unsigned s(paramset_size(PARAMSET_GROUP_RATE_CATS_PROB));
      for(unsigned i(st);i<(st+s);++i){
        if(_is_train_param[i]) { _param[i] = random_uniform()+rcprob_unorm_min; }
        else                   { _param[i] = 1./static_cast<smlfloat>(category_size_group_rate()); }
      }
    }

  } else if(dm == DENOVO_MODEL::START){
    static const smlfloat start_ratio(3.);

    for(unsigned i(0);i<category_size_site_mutation_model();++i){
      set_ts_tv_ratio(i,start_ratio);
    }
    set_paramset_pdistro_unif(PARAMSET_SITE_RATE_CATS_PROB);
    set_paramset_pdistro_unif(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB);
    set_paramset_slope(PARAMSET_GROUP_RATE_CATS);
    set_paramset_pdistro_unif(PARAMSET_GROUP_RATE_CATS_PROB);

    // don't setup initial category parameters on a slope if they are
    // tied to assigned data
    if(category_size_assigned() == 0){
      set_paramset_slope(PARAMSET_SITE_RATE_CATS);
      set_mutation_model_slope();
    }
  }

  {  // for any denovo case...
    const unsigned st(paramset_start(PARAMSET_EDGE_STRENGTH));
    const unsigned s(paramset_size(PARAMSET_EDGE_STRENGTH));
    for(unsigned i(st);i<(st+s);++i) if(_is_train_param[i]) _param[i]=0.5;
  }

  if(! is_protected_call) {
    self_t::fix_input_param();
  }
}



static
void
bg_pdistro_nuc_conditioned_update(const SITE_MODEL::index_t sm,
                                  const prob_t* bg_pdistro,
                                  prob_t* bg_pdistro_nuc_conditioned_on_3p,
                                  prob_t* bg_pdistro_nuc_conditioned_on_5p){

  get_nuc_distro_conditioned_on_3p_from_site_model_distro(sm,
                                                          bg_pdistro,
                                                          bg_pdistro_nuc_conditioned_on_3p);

  get_nuc_distro_conditioned_on_5p_from_site_model_distro(sm,
                                                          bg_pdistro,
                                                          bg_pdistro_nuc_conditioned_on_5p);
}



void
rate_gtor_nuc_base::
bg_pdistro_nuc_conditioned_update(){
  const SITE_MODEL::index_t sm(site_model());

 ::bg_pdistro_nuc_conditioned_update(sm,bg_pdistro(),
                                     _data.bg_pdistro_nuc_conditioned_on_3p.val,
                                     _data.bg_pdistro_nuc_conditioned_on_5p.val);

  const unsigned acn(category_size_assigned());
  if(acn){
    _data.bg_pdistro_nuc_conditioned_on_3p_cat.init(acn,NUC::SIZE*NUC::SIZE);
    _data.bg_pdistro_nuc_conditioned_on_5p_cat.init(acn,NUC::SIZE*NUC::SIZE);
    for(unsigned c(0);c<acn;++c){
      ::bg_pdistro_nuc_conditioned_update(sm,bg_pdistro_cat(c),
                                          _data.bg_pdistro_nuc_conditioned_on_3p_cat.val[c],
                                          _data.bg_pdistro_nuc_conditioned_on_5p_cat.val[c]);
    }
  }
}



/// helper func to report():
void
rate_gtor_nuc_base::
report_rate_set(const NUC::index_t n0,
                const NUC::index_t n2,
                const unsigned mut_model_cat,
                const char* label,
                std::ostream& os) const {
  std::string s(label);
  s += " : MUT RATE";
  const char sn0(NUC::syms[n0]);
  const char sn2(NUC::syms[n2]);
  if(context_model_nuc(mut_model_cat) != CONTEXT_MODEL_NUC::INDY){
    s = s+" : context= "+sn0+"X"+sn2+"->Z\n";
  }
  irv_t<smlfloat> rts[NUC::SIZE*NUC::SIZE];
  for(unsigned i(0);i<NUC::SIZE;++i){
    for(unsigned j(0);j<NUC::SIZE;++j){
      if(i==j) {
        rts[j+i*NUC::SIZE] = irv_t<smlfloat>(0.);
        continue;
      }
      const NUC::index_t n1(static_cast<NUC::index_t>(i));
      const NUC::index_t n1x(static_cast<NUC::index_t>(j));

#if 0
      rts[j+i*NUC::SIZE] = get_nuc_context_mut_rate(n0, n1,
                                                    n2, n1x,
                                                    mut_model_cat,
                                                    bg_pdistro_nuc(),bg_pdistro_nuc());
#else
      rts[j+i*NUC::SIZE] = get_nuc_context_mut_rate(n0, n1,
                                                    n2, n1x,
                                                    mut_model_cat,
                                                    bg_pdistro_nuc_conditioned_on_3p(n1),
                                                    bg_pdistro_nuc_conditioned_on_5p(n1));

#endif
    }
  }

  const smlfloat rate_scale(1./rate_time_bg_scale());
  matrix_scale(rts,NUC::SIZE,rate_scale);
  os << s << "\n";

  unsigned cell_width(9);
  if(rts[1].v != 0.){ cell_width = 22; }
  static const unsigned prec(3);

  matrix_report(rts,NUC::SIZE,NUC::syms,os,cell_width,prec);


  // print out a higher precision list in easily parsed form:
  static const unsigned lprec(6);
  os << std::setprecision(lprec);
  for(unsigned i(0);i<NUC::SIZE;++i){
    const char sni(NUC::syms[i]);
    for(unsigned j(0);j<NUC::SIZE;++j){
      if(i==j) continue;
      const char snj(NUC::syms[j]);
      os << "CDM_RATE: " <<  sn0 << sni << sn2 << "->" << snj << " " << rts[j+i*NUC::SIZE] << "\n";
    }
  }

  // get transi/tranv:
  //
//  {
//    irv_t<smlfloat> transi,transv;
//    get_si_sv(transi,transv,rts,bg_pdistro_nuc());
//    os << "CT_TI: " << transi << "\n";
//    os << "CT_TV: " << transv << "\n";
//  }
  os << "\n";
}




void
rate_gtor_nuc_base::
report(const char* label,
       std::ostream& os) const {

  using namespace CONTEXT_MODEL_NUC;

  smlfloat stat_pdistro[NUC::SIZE];
  stat_pdistro_nuc(stat_pdistro);
  for(unsigned i(0);i<NUC::SIZE;++i)
    os << "stat_distro " << NUC::syms[i] << " : " << stat_pdistro[i] << "\n";
  os << "\n";


  for(unsigned i(0);i<NUC::SIZE;++i)
    os << "bg_distro " << NUC::syms[i] << " : " << bg_pdistro_nuc()[i] << "\n";
  os << "\n";

  {  // dump site rate categories
    const unsigned site_cat_size(paramset_size(PARAMSET_SITE_RATE_CATS));

    std::vector<std::pair<irv_t<smlfloat>,std::pair<irv_t<smlfloat>,std::string> > > tmp_rate_cats(site_cat_size);
    for(unsigned i(0);i<site_cat_size;++i) {
      tmp_rate_cats[i] =
        std::make_pair(get_paramset_member(PARAMSET_SITE_RATE_CATS,i),
                       std::make_pair(get_paramset_pdistro_member(PARAMSET_SITE_RATE_CATS_PROB,i),
                                      _data.site_rate_cats_labels[i]));
    }

    std::sort(tmp_rate_cats.begin(),tmp_rate_cats.end());

    os << "SITE RATE CATEGORIES (" << site_cat_size << "):\n";
    os.unsetf(std::ios::fixed);
    os << std::setprecision(7);
    for(unsigned i(0);i<site_cat_size;++i){
      os << "[ " << tmp_rate_cats[i].second.second << " ]  : "
         << tmp_rate_cats[i].first << " prior_prob=(" << tmp_rate_cats[i].second.first << ")\n";
    }
    os << "\n";
  }

  {  // mutation model category info
    const unsigned site_cat_size(paramset_size(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB));

    std::vector<std::pair<irv_t<smlfloat>,std::pair<irv_t<smlfloat>,std::string> > > tmp_rate_cats(site_cat_size);
    for(unsigned i(0);i<site_cat_size;++i) {
      tmp_rate_cats[i] =
        std::make_pair(rates_nuc_norm(i),
                       std::make_pair(get_paramset_pdistro_member(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB,i),
                                      _data.site_mutation_model_cats_labels[i]));
    }

    std::sort(tmp_rate_cats.begin(),tmp_rate_cats.end());

    os << "SITE MUTATION MODEL CATEGORIES (" << site_cat_size << "):\n";
    os.unsetf(std::ios::fixed);
    os << std::setprecision(7);
    for(unsigned i(0);i<site_cat_size;++i){
      os << "[ " << tmp_rate_cats[i].second.second << " ]  : "
         << "scale=" << tmp_rate_cats[i].first << " prior_prob=(" << tmp_rate_cats[i].second.first << ")\n";
    }
    os << "\n";

  }

  { // dump group rate categories (normalized)
    const unsigned group_cat_size(paramset_size(PARAMSET_GROUP_RATE_CATS));

    std::vector<std::pair<smlfloat,std::pair<prob_t,std::string> > > tmp_rate_cats(group_cat_size);

    for(unsigned i(0);i<group_cat_size;++i) {
      tmp_rate_cats[i] =
        std::make_pair(get_paramset_member(PARAMSET_GROUP_RATE_CATS,i),
                       std::make_pair(get_paramset_pdistro_member(PARAMSET_GROUP_RATE_CATS_PROB,i),
                                      _data.group_rate_cats_labels[i]));
    }

    std::sort(tmp_rate_cats.begin(),tmp_rate_cats.end());

    os << "GROUP RATE CATEGORIES (" << group_cat_size << "):\n";
    os.unsetf(std::ios::fixed);
    os << std::setprecision(7);
    for(unsigned i(0);i<group_cat_size;++i){
      os << "[ " << tmp_rate_cats[i].second.second << " ]  : "
         << tmp_rate_cats[i].first << " prior_prob=(" << tmp_rate_cats[i].second.first << ")\n";
    }
    os << "\n";
  }


  for(unsigned i(0);i<category_size_site_mutation_model();++i){
    const unsigned mut_model_cat(i);

    os << "\n";
    os << "MUTATION MODEL CAT: " << mut_model_cat
       << " [ " <<  _data.site_mutation_model_cats_labels[mut_model_cat] << " ]\n\n";

    const CONTEXT_GROUP::index_t cg(get_context_model_group(context_model_nuc(mut_model_cat)));

    if(cg == CONTEXT_GROUP::TWOWAY){
      for(unsigned n0(0);n0<NUC::SIZE;++n0){
        for(unsigned n2(0);n2<NUC::SIZE;++n2){
          report_rate_set(static_cast<NUC::index_t>(n0),
                          static_cast<NUC::index_t>(n2),mut_model_cat,label,os);
        }
      }
      for(unsigned n0(0);n0<NUC::SIZE;++n0){
        NUC::index_t n2 = NUC::N;
        report_rate_set(static_cast<NUC::index_t>(n0),n2,mut_model_cat,label,os);
      }
      for(unsigned n2(0);n2<NUC::SIZE;++n2){
        NUC::index_t n0 = NUC::N;
        report_rate_set(n0,static_cast<NUC::index_t>(n2),mut_model_cat,label,os);
      }

    } else if(cg == CONTEXT_GROUP::ONLY_PRE){
      for(unsigned n0(0);n0<NUC::SIZE;++n0){
        NUC::index_t n2 = NUC::N;
        report_rate_set(static_cast<NUC::index_t>(n0),n2,mut_model_cat,label,os);
    }

    } else if(cg == CONTEXT_GROUP::ONLY_POST){
      for(unsigned n2(0);n2<NUC::SIZE;++n2){
        NUC::index_t n0 = NUC::N;
        report_rate_set(n0,static_cast<NUC::index_t>(n2),mut_model_cat,label,os);
      }
    }

    report_rate_set(NUC::N,NUC::N,mut_model_cat,label,os);


    if(model_type()==RATE_GTOR_MODEL::C4POST && category_size_group()==1 && category_size_site()==1) {

      // get CpG/Non-CpG transition rate: start out by getting the state neutral flux matrix
      // (last minute hack!! -- fix me later)
      //
      const unsigned n_states(state_size());
      const unsigned N2(n_states*n_states);
      simple_array<smlfloat> flux(N2);
      std::fill(flux.val,flux.val+N2,0.);

      rates(flux.val,rates_func_options(0,0,true));
      rates_to_flux_inplace(n_states,bg_pdistro(),flux.val);

      smlfloat cpgti_flux(0.);
      smlfloat cpgti_p(0.);
      smlfloat noncpgti_flux(0.);
      smlfloat noncpgti_p(0.);
      smlfloat noncpgti_cg_flux(0.);
      smlfloat noncpgti_cg_p(0.);


      const SITE_MODEL::index_t sm(site_model());
      const unsigned base_size(SITE_MODEL::base_size(sm));

      NUC::index_t s1n[SITE_MODEL::MAX_BASE_SIZE];
      NUC::index_t s2n[SITE_MODEL::MAX_BASE_SIZE];

      for(unsigned s1(0);s1<n_states;++s1){
        SITE_MODEL::decode_nuc(sm,s1,s1n);

        for(unsigned s2(0);s2<n_states;++s2){
          SITE_MODEL::decode_nuc(sm,s2,s2n);

          // 1. is single nuc diff?
          unsigned nsubs(0);
          unsigned mut_p(0);
          for(unsigned p(0);p<base_size;++p){
            if( s1n[p] != s2n[p] ) {
              mut_p=p;
              nsubs++;
            }
          }

          if(nsubs != 1) continue;

          // 2. is it a transition?:
          if(! NUC::is_transition(s1n[mut_p],s2n[mut_p])) continue;

          // 3. is it a CpG,non-Cpg,or indeterminate?

          // don't use position 3 for anything else but the NGN->A
          if       (mut_p==3){
            if(s1n[mut_p] == NUC::G){
              if(s1n[mut_p-1] == NUC::C){
                cpgti_flux += flux.val[s2+s1*n_states];
                cpgti_p += bg_pdistro()[s1];
              } else {
                noncpgti_flux += flux.val[s2+s1*n_states];
                noncpgti_p += bg_pdistro()[s1];
                noncpgti_cg_flux += flux.val[s2+s1*n_states];
                noncpgti_cg_p += bg_pdistro()[s1];
              }
            } else {
              //skip
            }

          // and dont read any G from pos 0
          } else if(mut_p==0 && s1n[mut_p] == NUC::G){
            //skip
          } else{
            if((s1n[mut_p] == NUC::C && s1n[mut_p+1] == NUC::G) ||
               (s1n[mut_p] == NUC::G && s1n[mut_p-1] == NUC::C)){
              cpgti_flux += flux.val[s2+s1*n_states];
              cpgti_p += bg_pdistro()[s1];
            } else {
              noncpgti_flux += flux.val[s2+s1*n_states];
              noncpgti_p += bg_pdistro()[s1];
              if(s1n[mut_p] == NUC::C || s1n[mut_p] == NUC::G){
                noncpgti_cg_flux += flux.val[s2+s1*n_states];
                noncpgti_cg_p += bg_pdistro()[s1];
              }
            }
          }
        }
      }

      const smlfloat rate_scale(1./rate_time_bg_scale());
      const smlfloat summ_cpgti(rate_scale*cpgti_flux/cpgti_p);
      const smlfloat summ_noncpgti(rate_scale*noncpgti_flux/noncpgti_p);
      const smlfloat summ_noncpg_cg_ti(rate_scale*noncpgti_cg_flux/noncpgti_cg_p);

      os << "SUMM_CPGTI: " << summ_cpgti << "\n";
      os << "SUMM_NONCPGTI: " << summ_noncpgti << "\n";
      os << "SUMM_CPGTI_RATIO: " << summ_cpgti/summ_noncpgti << "\n";
      os << "SUMM_NONCPGCGTI: " << summ_noncpg_cg_ti << "\n";
      os << "SUMM_CPGCGTI_RATIO: " << summ_cpgti/summ_noncpg_cg_ti << "\n";
      os << "\n";
//       os << "TI: " << noncpgti_flux/noncpgti_p << "\n";
//       os << "TV: " << noncpgti_flux/noncpgti_p << "\n";

    }
  }

#if 0
  { /// \todo  update these calls to average all mut models:
    smlfloat rts[NUC::SIZE*NUC::SIZE];
    rates_nuc(rts);
    smlfloat sis1_wsr = get_ws_ratio(rts,bg_pdistro_nuc());

    os << label << " : RATE W->S/S->W ::  "
       << std::setprecision(6)
       << sis1_wsr << "\n";
    os << "\n";

    smlfloat transi1,transv1;
    get_si_sv(transi1,transv1,rts,bg_pdistro_nuc());

    os << label << " : RATE KAPPA ::  "
       << std::setprecision(6)
       << transi1/transv1 << "\n";
    os << "\n";
  }
  os << "\n";
#endif

}




/// keep things consistent
static
void
format_root_header(const unsigned root_cat_no,
                   const char* root_label,
                   std::ostream& os){

  os << root_label << " prior prob: cat_no " << static_cast<int>(root_cat_no);
}




void
rate_gtor_nuc_base::
report_root(const root_gtor& rg,
            const char* label,
            std::ostream& os) const {

  const SITE_MODEL::index_t sm(site_model());

  for(unsigned i(0);i<rg.root_cat_size();++i){
    const prob_t* root_pdf(rg.root_cat_distro(i));

    format_root_header(i,label,os);
    os << "\n";
    print_site_model_distro(root_pdf,sm,os);

    prob_t nuc_pdf[NUC::SIZE];
    const BIO_PDISTRO::index_t bp(BIO_PDISTRO::convert_from_site_model(sm));
    if(bp != BIO_PDISTRO::NUC){
      bio_pdistro_convert(bp,BIO_PDISTRO::NUC,root_pdf,nuc_pdf);

      format_root_header(i,label,os);
      os << "\n";
      print_site_model_distro(nuc_pdf,SITE_MODEL::NUC,os);

    }
  }
}


// i686 gcc 4.1 has lots of (harmless) link warnings when this is left inline:
//
void
rate_gtor_nuc_base::
bg_pdistro_update(){
  base_t::bg_pdistro_update();
  bg_pdistro_nuc_update();
  bg_pdistro_nuc_conditioned_update();
}
