// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor_nuc_base.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __RATE_GTOR_NUC_BASE_H
#define __RATE_GTOR_NUC_BASE_H

#include "rate_gtor.h"
#include "rate_gtor_nuc_options.h"
#include "simple_util.h"
#include "util/bio/bioseq_util.h"
#include "util/math/indy_random_var.h"

#include <iosfwd>


/// \brief base rate generator for all models using nuc mutation
/// (coding or not)
///
struct rate_gtor_nuc_base : public rate_gtor_base {
  typedef rate_gtor_base base_t;
  typedef rate_gtor_nuc_base self_t;

  rate_gtor_nuc_base(const rate_gtor_nuc_options& nopt,
                     const rate_gtor_sml_share& smls)
    : base_t(smls), _data() {
    _data.nopt = nopt;
    init_class_storage();
    init_param_locks(nopt.lopt);
    init_cat_labels();
  }

  rate_gtor_nuc_base(const self_t& s,
                     const rate_gtor_sml_share& smls)
    :  base_t(s,smls), _data(s._data) {}

  virtual
  void rates(smlfloat* rates,
             const rates_func_options& opt = rates_func_options()) const;

  virtual
  void store_state(std::ostream& os) const;

  virtual
  void load_state(std::istream& is);

  virtual
  void report(const char* label,
              std::ostream& os) const;

  virtual
  void reset_param(const DENOVO_MODEL::index_t dm,
                   const bool is_protected_call = false);

  smlfloat
  get_category_nuc_context_mut_rate(const NUC::index_t from0,
                                    const NUC::index_t from1,
                                    const NUC::index_t from2,
                                    const NUC::index_t to,
                                    const unsigned site_category,
                                    const unsigned group_category,
                                    const prob_t* nuc_bg0,
                                    const prob_t* nuc_bg2) const;

  const prob_t* bg_pdistro_nuc() const { return _data.bg_pdistro_nuc.val; }

  const prob_t* bg_pdistro_nuc_cat(const unsigned site_cat) const {
    const unsigned acn(_data.bg_pdistro_nuc_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn){ return _data.bg_pdistro_nuc_cat.val[site_cat]; }
    else   { return bg_pdistro_nuc(); }
  }

  const prob_t*
  bg_pdistro_nuc_conditioned_on_5p(const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const {
    return _data.bg_pdistro_nuc_conditioned_on_5p.val+dep_n*NUC::SIZE;
  }

  const prob_t*
  bg_pdistro_nuc_conditioned_on_5p_cat(const unsigned site_cat,
                                       const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const{
    const unsigned acn(_data.bg_pdistro_nuc_conditioned_on_5p_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn) { return _data.bg_pdistro_nuc_conditioned_on_5p_cat.val[site_cat]+dep_n*NUC::SIZE; }
    else    { return bg_pdistro_nuc_conditioned_on_5p(dep_n); }
  }

  const prob_t*
  bg_pdistro_nuc_conditioned_on_3p(const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const {
    return _data.bg_pdistro_nuc_conditioned_on_3p.val+dep_n*NUC::SIZE;
  }

  const prob_t*
  bg_pdistro_nuc_conditioned_on_3p_cat(const unsigned site_cat,
                                       const NUC::index_t dep_n = static_cast<NUC::index_t>(0)) const{
    const unsigned acn(_data.bg_pdistro_nuc_conditioned_on_3p_cat.dim1());
    assert(acn==0 || site_cat<acn);
    if(acn) { return _data.bg_pdistro_nuc_conditioned_on_3p_cat.val[site_cat]+dep_n*NUC::SIZE; }
    else    { return bg_pdistro_nuc_conditioned_on_3p(dep_n); }
  }

private:
  prob_t* bg_pdistro_nuc() { return _data.bg_pdistro_nuc.val; }

  prob_t* bg_pdistro_nuc_cat(const unsigned site_cat) { return _data.bg_pdistro_nuc_cat.val[site_cat]; }

public:

  CONTEXT_MODEL_NUC::index_t
  context_model_nuc(const unsigned i) const { return _data.nopt.context_model(i); }

  CONTEXT_MODEL_NUC::index_t
  rates_func_context_model_nuc(const rates_func_options& opt) const {
    const int i(opt.site_category);
    unsigned mmc(0);
    if(i>=0) mmc=get_typed_cat_from_site_cat(SITE_CAT_TYPE::MUT_MODEL,i);

    return context_model_nuc(mmc);
  }

  void
  set_site_rate_cats_prob(const smlfloat* x){
    const unsigned st(paramset_start(PARAMSET_SITE_RATE_CATS_PROB));
    const unsigned s(paramset_size(PARAMSET_SITE_RATE_CATS_PROB));
    for(unsigned i(0);i<(s);++i)  _param[i+st] = x[i];
  }

  void
  set_site_mutation_model_cats_prob(const smlfloat* x){
    const unsigned st(paramset_start(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB));
    const unsigned s(paramset_size(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB));
    for(unsigned i(0);i<(s);++i)  _param[i+st] = x[i];
  }

  /// \todo naming conventions for the rate matrix scaling functions
  /// need work. clarify how each one deals with 1.cats 2.context
  /// 3.the site scale. also consider simplifying the difference
  /// between the two normalization schemes, one internal and one
  /// reported
  ///
  virtual
  smlfloat rate_time_bg_scale() const {
    return rate_time_context_bg_scale(NUC::N,NUC::N);
  }

  virtual
  smlfloat rate_time_context_avg_bg_scale() const;

  bool
  is_use_edge_correction() const { return _data.nopt.is_use_edge_correction; }

  smlfloat
  edge_strength_pre() const { return edge_strength(0); }

  smlfloat
  edge_strength_post() const { return edge_strength(1); }
#if 1
  smlfloat
  edge_strength_pre2() const { return edge_strength(0); }

  smlfloat
  edge_strength_post2() const { return edge_strength(1); }
#endif
  std::vector<std::string>& site_rate_cats_labels() { return _data.site_rate_cats_labels; }

  std::vector<std::string>& site_mutation_model_cats_labels() { return _data.site_mutation_model_cats_labels; }

  void
  report_root(const root_gtor& rg,
              const char* label,
              std::ostream& os) const;

  void
  lock_full_edge_correction() {
    const unsigned st(paramset_start(PARAMSET_EDGE_STRENGTH));
    const unsigned s(paramset_size(PARAMSET_EDGE_STRENGTH));
    for(unsigned i(st);i<(st+s);++i){
      _param[i] = 1.;
      _is_train_param[i]=false;
    }
  }

  virtual
  void bg_pdistro_update();

  // public versions of these methods to set ci subset (temporary!!)
  unsigned
  param_size_mut() const {
    return paramset_size(PARAMSET_NUC);
  }

  unsigned
  param_start_mut() const {
    return paramset_start(PARAMSET_NUC);
  }

  ///////////////////////////////////////////////////////////
protected:

  /// \todo fix this to account for mutation model categories:
  void stat_pdistro_nuc(prob_t* nuc_sd) const;

  smlfloat
  edge_strength(const unsigned i) const {
    if(! is_use_edge_correction()) return 0.;
    assert(i<paramset_size(PARAMSET_EDGE_STRENGTH));
    return _param[paramset_start(PARAMSET_EDGE_STRENGTH)+i];
  }

  unsigned
  parse_group_rate_category(const unsigned group_category) const {
    return group_category%category_size_group_rate();
  }

  unsigned
  param_end() const { return paramset_start(PARAMSET_END); }

  // cat sizes:
  unsigned
  category_size_site_rate() const { return get_typed_site_cat_size(SITE_CAT_TYPE::MUT_RATE); }

  unsigned
  category_size_site_mutation_model() const { return get_typed_site_cat_size(SITE_CAT_TYPE::MUT_MODEL); }

  unsigned
  category_size_group_rate() const { return _data.nopt.n_group_rate_cats; }

  virtual
  prob_t
  get_site_category_proportional_prob(const int site_category) const {
    if(site_category<0) return 1.;

    const unsigned rc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::MUT_RATE,site_category));
    const prob_t rc_p(get_paramset_pdistro_member(PARAMSET_SITE_RATE_CATS_PROB,rc));

    const unsigned mmc(get_typed_cat_from_site_cat(SITE_CAT_TYPE::MUT_MODEL,site_category));
    const prob_t mmc_p(get_paramset_pdistro_member(PARAMSET_SITE_MUTATION_MODEL_CATS_PROB,mmc));

    return rc_p*mmc_p;
  }

  virtual
  prob_t
  get_group_category_prob(const int group_category) const {
    return get_paramset_pdistro_member(PARAMSET_GROUP_RATE_CATS_PROB,group_category);
  }

private:
  void
  rates_nuc(smlfloat rates[NUC::SIZE*NUC::SIZE],
            const unsigned site_category,
            const unsigned group_category,
            const NUC::index_t pre_n = NUC::N,
            const NUC::index_t post_n = NUC::N) const;

protected:
  virtual
  void fix_input_param();

  virtual
  void state_pdistro_to_nuc_pdistro(const prob_t* state_pdistro,
                                    prob_t* nuc_pdistro) const = 0;

private:
  void
  bg_pdistro_nuc_update(){
    state_pdistro_to_nuc_pdistro(bg_pdistro(),bg_pdistro_nuc());
    const unsigned acn(category_size_assigned());
    if(acn){
      _data.bg_pdistro_nuc_cat.init(acn,NUC::SIZE);
      for(unsigned c(0);c<acn;++c){
        state_pdistro_to_nuc_pdistro(bg_pdistro_cat(c),bg_pdistro_nuc_cat(c));
      }
    }
  }

  void bg_pdistro_nuc_conditioned_update();

protected:
  virtual
  smlfloat
  trainable_param_transform(const unsigned i) const {
    if(i<base_t::param_end()) {
      return base_t::trainable_param_transform(i);
    }
    const unsigned st(paramset_start(PARAMSET_EDGE_STRENGTH));
    const unsigned s(paramset_size(PARAMSET_EDGE_STRENGTH));
    if(i>=st && i<(st+s)){
      return std::acos((_param[i]*2.)-1.);
    } else {
      return _param[i];
    }
  }

  virtual
  void
  trainable_param_untransform(const unsigned i,
                              const smlfloat p) {
    if(i<base_t::param_end()) {
      return base_t::trainable_param_untransform(i,p);
    }
    const unsigned st(paramset_start(PARAMSET_EDGE_STRENGTH));
    const unsigned s(paramset_size(PARAMSET_EDGE_STRENGTH));
    if(i>=st && i<(st+s)){
      _param[i] = (std::cos(p)+1.)/2.;
    } else {
      _param[i] = p;
    }
  }

  // parameter arrangement:
  enum paramset_t {
    PARAMSET_SITE_RATE_CATS,
    PARAMSET_SITE_RATE_CATS_PROB,
    PARAMSET_SITE_MUTATION_MODEL_CATS_PROB,
    PARAMSET_GROUP_RATE_CATS,
    PARAMSET_GROUP_RATE_CATS_PROB,
    PARAMSET_NUC,
    PARAMSET_EDGE_STRENGTH,
    PARAMSET_END
  };

  virtual
  unsigned
  paramset_size(const int p) const {
    switch(static_cast<const paramset_t>(p)){
    case PARAMSET_SITE_RATE_CATS :
      return category_size_site_rate() ;
    case PARAMSET_SITE_RATE_CATS_PROB :
      return category_size_site_rate();
    case PARAMSET_SITE_MUTATION_MODEL_CATS_PROB :
      return category_size_site_mutation_model();
    case PARAMSET_GROUP_RATE_CATS :
      return category_size_group_rate();
    case PARAMSET_GROUP_RATE_CATS_PROB :
      return category_size_group_rate();
    case PARAMSET_NUC :
      {
        unsigned nsize(0);
        for(unsigned i(0);i<category_size_site_mutation_model();++i){
          nsize += paramset_mutation_model_cat_size(i);
        }
        return nsize;
      }
    case PARAMSET_EDGE_STRENGTH :
      if(is_use_edge_correction()) {return 2;}
      else                         {return 0;}
    default : return 0;
    }
  }

  unsigned
  paramset_start(const int p) const {
    unsigned s(base_t::param_end());
    for(int i(0);i<p;++i) s+=paramset_size(i);
    return s;
  }

  void
  set_paramset_dependent_block(const int p){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    if(s){
      _is_free_block_start[st] = true;
      _is_free_block_stop[st+s-1] = true;
    }
  }

  void
  set_paramset_val(const int p,
                   const smlfloat val){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    for(unsigned i(st);i<(st+s);++i)  _param[i] = val;
  }

  /// set paramset values in even increments from 2. to 0.  this is
  /// intended to minimally pull initial values apart so that each has
  /// a distinct gradient during initial minimization
  ///
  void
  set_paramset_slope(const int p){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    const smlfloat increment(2./(s+1));
    for(unsigned i(0);i<(s);++i) _param[i+st] = (i+1)*increment;
  }

  void
  set_paramset_pdistro_unif(const int p){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    pdistro_unif(_param+st,s);
  }

  void
  set_paramset_pdistro_norm(const int p){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    if(s) pdistro_norm(_param+st,_param+st+s);
  }

  void
  set_paramset_pdistro_norm_free_param(const int p){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    if(s) pdistro_norm_free_param(_param+st,_param+st+s,_is_train_param.begin()+st);
  }

  void
  set_paramset_train(const int p,
                     bool b){
    const unsigned st(paramset_start(p));
    const unsigned s(paramset_size(p));
    for(unsigned i(st);i<(st+s);++i) _is_train_param[i]=b;
  }

  irv_t<smlfloat>
  get_paramset_member(const int p,
                      const unsigned i) const {
    assert(i<paramset_size(p));
    const unsigned pi(paramset_start(p)+i);
    return irv_t<smlfloat>(_param[pi],param_variance(pi));
  }

  irv_t<smlfloat>
  get_paramset_pdistro_member(const int p,
                              const int i) const {
    assert(i<static_cast<int>(paramset_size(p)));
    if(i<0) return irv_t<smlfloat>(1.);
    const unsigned pi(paramset_start(p)+i);
    return irv_t<smlfloat>(_param[pi],param_variance(pi));
  }


  ///////////////////////////////////////////////////////////////////
private:
  void init_class_storage();

  void init_param_locks(const rate_gtor_nuc_lock_options& lopt);

  void init_cat_labels();

  void
  paramset_label(const paramset_t p,
                 unsigned n,
                 std::ostream& os) const;

  RATE_MODEL_NUC::index_t
  rate_model_nuc(const unsigned i) const { return _data.nopt.rate_model(i); }

  void
  scale_nuc_mut_rate_trainable_param(const unsigned mut_model_cat,
                                     smlfloat x){
    const unsigned st(paramset_mutation_model_cat_start(mut_model_cat));
    const unsigned s(paramset_mutation_model_cat_size(mut_model_cat));
    if(context_model_nuc(mut_model_cat) == CONTEXT_MODEL_NUC::FACTORED_TRIPLET) x = std::sqrt(x);
    for(unsigned i(st);i<(st+s);++i) if(_is_train_param[i]) _param[i] *= x;
  }

  smlfloat
  rate_time_context_bg_scale(const NUC::index_t pre_n,
                             const NUC::index_t post_n) const;

  struct trainable_mode {
    enum index_t { NONE, FULL, HALF_JOINT };
  private:
    trainable_mode() {}
  };


  // report the time scale of the rate matrix under a uniform state
  // distribution, used to normalize the rate matrix for model
  // comparison.
  //
  // trainable mode facilitates normalization when some rate
  // parameters are locked
  //
  smlfloat rates_nuc_norm(const unsigned mut_model_cat,
                          const trainable_mode::index_t tm = trainable_mode::NONE) const;

  enum joint_param_t { JOINT1, JOINT2 };

  void
  get_nuc_param_index_name(const unsigned mut_model_param_no,
                           const unsigned param_index,
                           std::ostream& os) const;

  unsigned
  get_nuc_context_mut_rate_param_index(const NUC::index_t from0,
                                       NUC::index_t from1,
                                       const NUC::index_t from2,
                                       NUC::index_t to,
                                       const unsigned mut_model_cat,
                                       const joint_param_t jp = JOINT1) const;

  irv_t<smlfloat>
  get_nuc_context_mut_rate_param(const NUC::index_t from0,
                                 const NUC::index_t from1,
                                 const NUC::index_t from2,
                                 const NUC::index_t to,
                                 const unsigned mut_model_cat,
                                 const trainable_mode::index_t tm) const;

  irv_t<smlfloat>
  get_nuc_context_mut_rate(const NUC::index_t from0,
                           const NUC::index_t from1,
                           const NUC::index_t from2,
                           const NUC::index_t to,
                           const unsigned mut_model_cat,
                           const prob_t* nuc_bg0,
                           const prob_t* nuc_bg2,
                           const trainable_mode::index_t tm = trainable_mode::NONE) const;


  // init to a fixed transition/transversion ratio
  //
  void
  set_ts_tv_ratio(const unsigned mut_model_cat,
                  const smlfloat ratio);


  // helper func to set_ts_tv_ratio:
  void
  set_ts_tv_ratio_nhood(const unsigned mut_model_cat,
                        const smlfloat ratio,
                        const NUC::index_t n0,
                        const NUC::index_t n2);

  // helper func to report
  void
  report_rate_set(NUC::index_t n0,
                  NUC::index_t n2,
                  const unsigned mut_model_cat,
                  const char* label,
                  std::ostream& os) const;

  void
  fix_input_param_cat_paramset(const paramset_t cats,
                               const paramset_t cats_prob);

  static
  unsigned
  get_rate_context_model_param_size(const RATE_MODEL_NUC::index_t rm,
                                    const CONTEXT_MODEL_NUC::index_t cm);

  unsigned
  paramset_mutation_model_cat_size(const unsigned cat_no) const {
    return get_rate_context_model_param_size(rate_model_nuc(cat_no),
                                             context_model_nuc(cat_no));
  }

  unsigned
  paramset_mutation_model_cat_start(const unsigned cat_no) const {
    unsigned val(paramset_start(PARAMSET_NUC));

    for(unsigned i(0);i<cat_no;++i){
      val += paramset_mutation_model_cat_size(i);
    }

    return val;
  }

  void
  set_mutation_model_slope(){
    const unsigned nmmc(category_size_site_mutation_model());
    const smlfloat increment(2./(nmmc+1));
    for(unsigned i(0);i<nmmc;++i){
      const unsigned st(paramset_mutation_model_cat_start(i));
      const unsigned s(paramset_mutation_model_cat_size(i));
      const smlfloat val((i+1)*increment);
      for(unsigned j(st);j<(st+s);++j) _param[j] *= val;
    }
  }

  /// struct to simplify class copy constructor:
  struct rate_gtor_nuc_base_autocopy_data {

    rate_gtor_nuc_base_autocopy_data()
      : bg_pdistro_nuc(NUC::SIZE),
        bg_pdistro_nuc_conditioned_on_5p(NUC::SIZE*NUC::SIZE),
        bg_pdistro_nuc_conditioned_on_3p(NUC::SIZE*NUC::SIZE) {}

    rate_gtor_nuc_options_base nopt;
    std::vector<std::string> site_rate_cats_labels;
    std::vector<std::string> site_mutation_model_cats_labels;
    std::vector<std::string> group_rate_cats_labels;

    simple_array<prob_t> bg_pdistro_nuc;
    simple_array<prob_t> bg_pdistro_nuc_conditioned_on_5p;
    simple_array<prob_t> bg_pdistro_nuc_conditioned_on_3p;

    simple_init_matrix<prob_t> bg_pdistro_nuc_cat;
    simple_init_matrix<prob_t> bg_pdistro_nuc_conditioned_on_5p_cat;
    simple_init_matrix<prob_t> bg_pdistro_nuc_conditioned_on_3p_cat;
  };

  ///////////////////////////data:
  rate_gtor_nuc_base_autocopy_data _data;
};


#endif
