// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: root_gtor.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "root_gtor.h"
#include "util/bio/bioseq_util_pdf_conversions.h"
#include "util/general/die.h"
#include "util/general/metatags.h"
#include "util/math/random_util.h"

#include <algorithm>
#include <iostream>


void
root_gtor::
reset_param(const DENOVO_MODEL::index_t dm){
  for(unsigned i(0);i<root_cat_size();++i){
    ptor_t& p(root_cat_ptor(i));

    if       ( dm == DENOVO_MODEL::RANDOM ){
      static const smlfloat unorm_min(1.e-1);
      const unsigned ps(p.param_size());
      std::vector<smlfloat> param(ps);
      for(unsigned j(0);j<ps;++j){ param[j] = random_uniform()+unorm_min; }
      p.set_param_state(param.begin());
    } else if( dm == DENOVO_MODEL::START ){
      const unsigned ss(p.state_size());
      smlfloat* distro(new smlfloat[ss]);
      pdistro_unif(distro,ss);
      p.set_distro(distro);
      delete [] distro;
    } else {
      die("unknown dm type");
    }
  }
}




static const char* const section_id = "root_gtor";
static const char* const end_label = "END";




void
root_gtor::
store_state(std::ostream& os) const {

  os << section_id << " root_gtor_model " << static_cast<int>(_rogm) << "\n";

  std::vector<smlfloat> param;
  std::vector<bool> itp;
  for(unsigned i(0);i<root_cat_size();++i){
    const ptor_t& p(root_cat_ptor(i));
    const unsigned ps(p.param_size());
    param.resize(ps);
    itp.resize(ps);
    p.param_state(param.begin());
    p.is_train_param(itp.begin());
    for(unsigned j(0);j<ps;++j){
      os << section_id << " root_cat_" << i << "_param_" << j << " "
         << itp[j] << " " << param[j] << "\n";
    }
  }
  os << section_id << " " << end_label << "\n";
}




static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid model file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  std::cerr << is.rdbuf();
  abort();
}


static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(tmpstr != section_id) load_die(is,section_id,tmpstr);
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}




void
root_gtor::
load_state(std::istream& is) {

  std::string tmpstr;

  {
    int tmpi;
    load_advance(is,"root_gtor_model");
    is >> tmpi;
    _rogm=static_cast<ROOT_GTOR_MODEL::index_t>(tmpi);

  }

  init();

  std::vector<smlfloat> param;
  std::vector<bool> itp;

  for(unsigned i(0);i<root_cat_size();++i){
    ptor_t& p(root_cat_ptor(i));
    const unsigned ps(p.param_size());
    param.resize(ps);
    itp.resize(ps);
    for(unsigned j(0);j<ps;++j){
      load_advance(is);
      bool btmp;
      is >> btmp >> param[j];
      itp[j] = btmp;
    }
    p.set_param_state(param.begin());
    p.set_is_train_param(itp.begin());
  }
  load_advance(is,end_label);
}



smlfloat
root_gtor::
param_variance(unsigned n) const {
  for(unsigned i(0);i<root_cat_size();++i){
    const ptor_t& p(root_cat_ptor(i));
    if(n < p.param_size()){
      return p.param_variance(n);
    } else {
      n -= p.param_size();
    }
  }
  die("root_gtor: invalid param_variance call\n");
}


#include "prob_gtor_state_conversion.h"
#include "prob_gtor_simple.h"
#include "prob_gtor_obs_derived.h"


#ifdef BRANCH_SPECIFIC_OBS
const prob_t* const *
root_gtor::
root_cat_node_state_prob(const unsigned root_cat) const {
  if(_rogm != ROOT_GTOR_MODEL::LSPROB) return 0;
  return static_cast<const prob_gtor_root_lsprob&>(root_cat_ptor(root_cat)).node_state_prob();
}
#endif


root_gtor::ptor_t*
root_gtor::
prob_gtor_factory(const unsigned root_cat_no) const {

  const unsigned n_states(state_size());
  const RATE_GTOR_MODEL::index_t ragm(_smls.rate_gtor_model());

  const bool is_c4_site_model(ragm == RATE_GTOR_MODEL::C4PRE ||
                              ragm == RATE_GTOR_MODEL::C4POST);

  const bool is_c4_pre(ragm==RATE_GTOR_MODEL::C4PRE);

  if       (_rogm == ROOT_GTOR_MODEL::CODON){
    if       (is_c4_site_model){
      return new prob_gtor_nsc4_from_nscodon<smlfloat>(is_c4_pre);
    } else if(ragm==RATE_GTOR_MODEL::C5){
      return new prob_gtor_nsc5_from_nscodon<smlfloat>();
    } else {
      std::cerr << "ERROR:: root codon model only valid with c4-pre/c4-post/c5 site models\n";
    }
  } else if(_rogm == ROOT_GTOR_MODEL::CODON_DINUC){
    if       (is_c4_site_model){
      return new prob_gtor_nsc4_from_nscodon_dinuc<smlfloat>(is_c4_pre);
    } else if(ragm==RATE_GTOR_MODEL::C5){
      return new prob_gtor_nsc5_from_nscodon_dinuc<smlfloat>();
    } else {
      std::cerr << "ERROR:: root codon-dinuc model only valid with c4-pre/c4-post/c5 site models\n";
    }
  } else if(_rogm == ROOT_GTOR_MODEL::NUC){
    if       (ragm==RATE_GTOR_MODEL::DINUC){
      return new prob_gtor_dinuc_from_nuc<smlfloat>();
    } else if(ragm==RATE_GTOR_MODEL::TRINUC){
      return new prob_gtor_trinuc_from_nuc<smlfloat>();
    } else if(ragm==RATE_GTOR_MODEL::CODON){
      return new prob_gtor_nscodon_from_nuc<smlfloat>();
    } else if(is_c4_site_model){
      return new prob_gtor_nsc4_from_nuc<smlfloat>();
    } else {
      std::cerr << "ERROR:: invalid site model for root nuc model\n";
    }
  } else if(_rogm == ROOT_GTOR_MODEL::NUC_POS){
    if(ragm==RATE_GTOR_MODEL::CODON){
      return new prob_gtor_nscodon_from_nuc_pos<smlfloat>();
    } else if(is_c4_site_model){
      return new prob_gtor_nsc4_from_nuc_pos<smlfloat>(is_c4_pre);
    } else {
      std::cerr << "ERROR:: invalid site model root nuc-pos model\n";
    }
  } else if(_rogm == ROOT_GTOR_MODEL::OBS_TIME_AVG){
    return new prob_gtor_obs_state_time_avg(_smls,root_cat_no);
  } else if(_rogm == ROOT_GTOR_MODEL::PPROB){
    return new prob_gtor_root_lsprob(_smls,false,root_cat_no);
  } else if(_rogm == ROOT_GTOR_MODEL::LSPROB){
    return new prob_gtor_root_lsprob(_smls,true,root_cat_no);
  } else {
    return new prob_gtor_simple<smlfloat>(n_states);
  }

  exit(EXIT_FAILURE);

  return static_cast<ptor_t*>(0);
}
