// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: sim_discrete.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "cat_info.h"
#include "report_util.h"
#include "sim_discrete.h"
#include "sim_util.h"
#include "simple_util.h"
#include "site_code.h"
#include "tree_probs.h"
#include "util/math/random_util.h"

#include <iostream>
#include <iomanip>
#include <vector>

using namespace std;



static
void
rate_to_discrete_time_anysub_prob(const smlfloat* rates,
                                  prob_t* anysub_prob,
                                  const smlfloat time_unit,
                                  const unsigned n_states){

  for(unsigned i(0);i<n_states;++i) anysub_prob[i] = -time_unit*rates[i*(1+n_states)];
}



static
void
rate_to_discrete_time_anysub_cdf(const smlfloat* rates,
                                 prob_t* anysub_cdf,
                                 const unsigned n_states){

  for(unsigned i(0);i<n_states;++i){
    for(unsigned j(0);j<n_states;++j){
      if(j==i) continue;
      anysub_cdf[j+(j<i?0:-1)+i*(n_states-1)] = rates[j+i*n_states];
    }
    pdistro_norm(anysub_cdf+i*(n_states-1),anysub_cdf+(i+1)*(n_states-1));
    pdistro_to_cdf_inplace(anysub_cdf+i*(n_states-1),n_states-1);
  }
}



static
void
setup_anysub_rates(prob_t** anysub_prob,
                   prob_t** anysub_cdf,
                   const smlfloat time_unit,
                   const subs_ml_model& mdl){

  const unsigned n_states = mdl.state_size();
  const unsigned n_site_cats = mdl.get_rate_gtor().category_size_site();

  smlfloat* rates = new smlfloat[n_states*n_states];

  for(unsigned r(0);r<n_site_cats;++r){
    mdl.get_rate_gtor().rates(rates,rates_func_options(r,0));

    rate_to_discrete_time_anysub_prob(rates,anysub_prob[r],time_unit,n_states);
    rate_to_discrete_time_anysub_cdf(rates,anysub_cdf[r],n_states);
  }
  delete [] rates;
}



static
void
rates_diag_mutonly(const rate_gtor_base& r,
                   const int site_category,
                   smlfloat* rates_diag){

  const unsigned n_states(r.state_size());

  smlfloat* rates(new smlfloat[n_states*n_states]);
  r.rates(rates,rates_func_options(site_category,
                                   0,
                                   true,true));

  for(unsigned i(0);i<n_states;++i) rates_diag[i] = rates[i*(1+n_states)];
  delete [] rates;
}



static
void
setup_anysub_neutral_rates(prob_t** anysub_prob_neutral,
                           const smlfloat time_unit,
                           const subs_ml_model& mdl){

  const unsigned n_states = mdl.state_size();
  const unsigned n_site_cats = mdl.get_rate_gtor().category_size_site();

  smlfloat* rates_diag(new smlfloat[n_states]);

  for(unsigned r(0);r<n_site_cats;++r){
    rates_diag_mutonly(mdl.get_rate_gtor(),r,rates_diag);

    for(unsigned i(0);i<n_states;++i) anysub_prob_neutral[r][i] = -time_unit*rates_diag[i];
  }
  delete [] rates_diag;
}


struct sim_info_t {
  vector<unsigned> site_cats_seq;
  simple_init_matrix<prob_t> anysub_prob; ///< prob of any substitution during the discrete time unit
  simple_init_matrix<prob_t> anysub_cdf;  ///< cdf of substitution in discrete time unit
  smlfloat time_unit_normed;              ///< discrete time unit
  unsigned n_states;
  unsigned n_site_cats;
};



static
unsigned
simulate_single_step(vector<unsigned>& endseq,
                     const sim_info_t& si,
                     const bool is_report_sim_time,
                     const prob_t* const * anysub_prob_neutral,
                     const smlfloat fracstep = 1.){

  unsigned n_mutation(0);


  const unsigned lsize = endseq.size();
  for(unsigned i(0);i<lsize;++i){
    // lookup the site rate categories:
    const unsigned sc=si.site_cats_seq[i];

    const double r = random_uniform();
    const unsigned tmpl = endseq[i];

    // record number of attempted (neutral) mutations
    if(is_report_sim_time && r<(anysub_prob_neutral[sc][tmpl]*fracstep)) n_mutation++;

    if(r<(si.anysub_prob.val[sc][tmpl]*fracstep)){
      unsigned tmpi = random_cdf_variate(si.anysub_cdf.val[sc]+tmpl*(si.n_states-1),si.n_states-1);
      if(tmpl<=tmpi) tmpi += 1;
      endseq[i] = tmpi;
    }
  }
  return n_mutation;
}


/// \param anysub_prob_neutral prob of any substitution during the
/// discrete time unit, absent selection
///
static
void
simulate_discrete_time_branch(const vector<unsigned>& startseq,
                              vector<unsigned>& endseq,
                              const smlfloat time,
                              const sim_info_t& si,
                              const bool is_report_sim_time,
                              const prob_t* const * anysub_prob_neutral,
                              smlfloat& sim_time){

  const smlfloat time_tmp = time/si.time_unit_normed;
  const unsigned nsteps = static_cast<unsigned>(time_tmp);
  const smlfloat fracstep = time_tmp-static_cast<smlfloat>(nsteps);

  std::cerr << "calc branch: steps,frac: " << nsteps << " " << fracstep << "\n";

  endseq = startseq;

  unsigned n_mutation(0);
  for(unsigned n(0);n<nsteps;++n){
    n_mutation += simulate_single_step(endseq,si,is_report_sim_time,anysub_prob_neutral);
  }

  // last step -- revert with probability 1-fracstep
  n_mutation += simulate_single_step(endseq,si,is_report_sim_time,anysub_prob_neutral,fracstep);

  sim_time = static_cast<double>(n_mutation)/static_cast<double>(endseq.size());
}



// non-context, discrete time sim
void
simulate_data_discrete(const sim_options& sim_opt,
                       const subs_ml_model& mdl,
                       site_data& sd){

  static const bool is_site_pair(false);

  const unsigned n_states = mdl.state_size();
  const unsigned n_site_cats = mdl.get_rate_gtor().category_size_site();

  sim_info_t sim_info;

  sim_info.n_states = n_states;
  sim_info.n_site_cats = n_site_cats;

  const unsigned n_treenodes(mdl.tree().n_nodes());
  const unsigned n_branches(mdl.tree().n_branches());
  const unsigned n_leaves(mdl.tree().n_leaves());

  sim_site_data_init(mdl.tree(),mdl.rate_gtor_model(),sd);

  // assign sites to site categories:
  get_sim_site_categories(sim_info.site_cats_seq,sim_opt.size,mdl.get_rate_gtor());

  // get root sequence:
  vector<vector<unsigned> > treenode_seq(n_treenodes);
  get_ancestral_seq(mdl,sim_opt.size,sim_info.site_cats_seq,treenode_seq[mdl.tree().get_node_index(mdl.tree().root)]);


  // convert instantaneous rate to the prob of exchange in the discrete time unit:
  //
  //  anysub_prob - at any one site, the probability of changing state
  //  in 1 time unit
  //
  //  anusub_cdf - at any one site, the cdf for the probability of a
  //  different state, given that a change occurs
  //
  const unsigned n_cats(n_site_cats);
  sim_info.time_unit_normed = SIMULATE_DISCRETE_TIME_UNIT/mdl.get_rate_gtor().rate_time_bg_scale();

  // setup neutral substitution count data structures:
  vector<smlfloat> branch_time_sim(n_branches);
  prob_t** anysub_prob_neutral;
  if(sim_opt.is_report_time){
    anysub_prob_neutral = new prob_t*[n_cats];
    for(unsigned i(0);i<n_cats;++i) anysub_prob_neutral[i] = new prob_t[n_states];

    setup_anysub_neutral_rates(anysub_prob_neutral,sim_info.time_unit_normed,mdl);
  }

  {
    sim_info.anysub_prob.init(n_cats,n_states);
    sim_info.anysub_cdf.init(n_cats,n_states*(n_states-1));

    setup_anysub_rates(sim_info.anysub_prob.val,sim_info.anysub_cdf.val,sim_info.time_unit_normed,mdl);

    // 3. propagate down branches to produce present day sequences:
    for(int i(1);i<static_cast<int>(n_treenodes);++i){
      simulate_discrete_time_branch(treenode_seq[mdl.tree().get_parent_index(i)],treenode_seq[i],
                                    mdl.get_time_gtor().branch_time(i-1),
                                    sim_info,sim_opt.is_report_time,anysub_prob_neutral,branch_time_sim[i-1]);
    }
  }

  if(sim_opt.is_report_time){
    for(unsigned i(0);i<n_cats;++i) delete [] anysub_prob_neutral[i];
    delete [] anysub_prob_neutral;

    const unsigned time_unit_per_state(SITE_MODEL::base_size(mdl.get_rate_gtor().site_model()));
    const smlfloat time_scale(1./static_cast<smlfloat>(time_unit_per_state));
    for(unsigned i(0);i<n_branches;++i) branch_time_sim[i] *= time_scale;

    std::cerr << "\n";
    report_time_instance("sim_time",branch_time_sim,mdl.tree(),std::cerr);
  }

  const SITE_MODEL::index_t smodel(RATE_GTOR_MODEL::convert_to_site_model(mdl.rate_gtor_model()));
  const unsigned ambig_state(SITE_MODEL::ambig_state(smodel));

  site_code sc(n_leaves);
  site_code prev_sc(n_leaves);
  unsigned prev_cat_no(UNASSIGNED_CAT_ID);

  // chop up present day sequences into site_data data format:
  for(int i(0);i<sim_opt.size;++i){

    for(unsigned t(0);t<n_leaves;++t){
      sc.set_taxid(t,treenode_seq[mdl.tree().get_node_index_from_leaf_index(t)][i]);
    }

    if(i==0) {
      for(unsigned t(0);t<n_leaves;++t){
        prev_sc.set_taxid(t,ambig_state);
      }
      prev_cat_no = UNASSIGNED_CAT_ID;
    }

    unsigned cat_no(UNASSIGNED_CAT_ID);
    if(sim_opt.assigned_cat_prob>0.){
      const prob_t r = random_uniform();
      if(r<sim_opt.assigned_cat_prob){
        cat_no=sim_info.site_cats_seq[i]+1;
      }
    }

    sd.site_count[sc][count_data_code(cat_no)]++;

    if(is_site_pair){
      sd.pair_count[make_pair(make_pair(prev_sc,sc),make_pair(prev_cat_no,cat_no))]++;

      prev_sc = sc;
      prev_cat_no = cat_no;
    }
  }
}
