// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: sim_iss.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "lhood_site.h"
#include "sim_iss.h"
#include "sim_util.h"
#include "site_code.h"
#include "subs_ml_ptol.h"
#include "tree_probs.h"
#include "util/general/die.h"
#include "util/general/log.h"
#include "util/math/random_util.h"

#include <ostream>
#include <iomanip>
#include <vector>


/// \todo switch to an mcmc sampling scheme
///
void
simulate_data_indy_site_sampler(const sim_options& sim_opts,
                                const subs_ml_model& mdl,
                                site_data& sd){

  static const unsigned MAX_SITES(1000000);

  const unsigned n_states(mdl.state_size());
  const unsigned n_site_cats(mdl.get_rate_gtor().category_size_site());

  std::vector<tree_probs> tprob(n_site_cats);

  for(unsigned sc(0);sc<n_site_cats;++sc){
    tprob[sc].init_probs(mdl,sc);
  }

  std::vector<prob_t> site_cat_prior_pdistro(n_site_cats);
  mdl.get_rate_gtor().get_site_category_pdistro(site_cat_prior_pdistro.begin());

  sim_site_data_init(mdl.tree(),mdl.rate_gtor_model(),sd);

  const unsigned n_leaves(mdl.tree().n_leaves());

  unsigned n_sites(1);
  for(unsigned t(0);t<n_leaves;++t){
    /// \todo add overflow safeguards here
    n_sites *= n_states;
  }

  prob_t* site_prob_cdf(new prob_t[n_sites*n_site_cats]);

  const unsigned ping_size(n_sites/100);
  unsigned ping_perc(0);
  bool is_sim_ping(false);
  if(n_sites >= MAX_SITES) {
    is_sim_ping=true;
    log_os << "Starting sim_distro calculation:\n";
  }

  prob_t sum(0.);
  site_code::index_type* leaf_id(new site_code::index_type[n_leaves]);
  for(unsigned sc(0);sc<n_site_cats;++sc){
    for(unsigned i(0);i<n_sites;++i){
      for(unsigned t(0),ii(i);t<n_leaves;++t){
        leaf_id[t] = ii%n_states;
        ii /= n_states;
      }

      const prob_t site_prob = get_site_prob_single_cat(leaf_id,
                                                        mdl.tree(),
                                                        tprob[sc],
                                                        n_states);

      sum += site_prob*site_cat_prior_pdistro[sc];
      site_prob_cdf[sc*n_sites+i] = sum;

      if(is_sim_ping){
        if(((i+1)%ping_size)==0) {
          log_os << "site_cat " << sc << " sim_distro: " << (++ping_perc) << "% complete\n";
        }
      }
    }
  }

  delete [] leaf_id;


  log_os << "SITE PROB SUM= " << sum << "\n";
#ifdef DEBUG
  if(std::fabs(sum-1.)>SUBS_ML_PTOL){
    log_os << "Simulator prob distro sum deviates from unity: "
           << std::setprecision(12) << sum << "\n";
    abort();
  }
#endif


  // sample from site_prob_cdf to produce simulated data, store
  // sim data in sd:
  for(int i(0);i<sim_opts.size;++i){
    const unsigned ranindex(random_cdf_variate(site_prob_cdf,n_sites*n_site_cats));
    const unsigned site_id(ranindex%n_sites);
    const unsigned cat_no(ranindex/n_sites);

    site_code sc(n_leaves);
    for(unsigned t(0),ii(site_id);t<n_leaves;++t){
      sc.set_taxid(t,ii%n_states);
      ii /= n_states;
    }

    unsigned observed_cat_no(0);

    if(sim_opts.assigned_cat_prob>0.){
      const prob_t r(random_uniform());
      if(r<sim_opts.assigned_cat_prob){
        observed_cat_no=cat_no+1;
      }
    }

    sd.site_count[sc][count_data_code(observed_cat_no)]++;
  }

  delete [] site_prob_cdf;
}
