// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: simple_util.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __SIMPLE_UTIL_H
#define __SIMPLE_UTIL_H


inline
unsigned
off_diag_size(const unsigned a){
  return (a*(a-1));
}

#if 0
inline
unsigned
encode_nmer(const unsigned a[],
            const unsigned len,
            const unsigned N){
  unsigned x = a[0];
  for(unsigned i(1);i<len;++i) x = x*N+a[i];
  return x;
}


inline
void
decode_nmer(const unsigned site_id,
            const unsigned len,
            const unsigned N,
            unsigned a[]){
  unsigned x = site_id;
  for(unsigned i(len-1);i;--i){
    a[i] = x%N;
    x /= N;
  }
  a[0] = x;
}
#endif


inline
unsigned
periodic_decrement(const unsigned val,
                   const unsigned period){
  //assert(period>0);
  return (val+(period-1))%period;
}


inline
unsigned
periodic_increment(const unsigned val,
                   const unsigned period){
  //assert(period>0);
  return (val+1)%period;
}



template <typename T>
struct simple_array {
  simple_array(const unsigned dim1)
    : val(0), _dim1(dim1) { init(); }

  simple_array(const simple_array<T>& s)
    : val(0), _dim1(s._dim1) {
    init();
    for(unsigned i(0);i<_dim1;++i) val[i] = s.val[i];
  }

  ~simple_array(){ delete [] val; }

private:
  simple_array& operator=(const simple_array<T>&);

  void
  init(){ val = new T[_dim1]; }

public:
  T* val;
private:
  unsigned _dim1;
};



template <typename T>
struct simple_init_array {

  simple_init_array() : val(0), _dim1(0) {}

  simple_init_array(const simple_init_array<T>& s) : val(0), _dim1(0) { deep_copy(s); }

  ~simple_init_array() { dealloc(); }

  simple_init_array& operator=(const simple_init_array<T>& rhs) {
    if(&rhs == this) return *this;
    deep_copy(rhs);
    return *this;
  }

  void
  init(const unsigned d1) {
    if(val && _dim1 == d1) return;
    dealloc();
    _dim1 = d1;
    alloc();
  }

  unsigned dim1() const { return _dim1; }
  unsigned size() const { return dim1(); }

private:
  void
  alloc(){ val = new T[_dim1]; }

  void
  dealloc() {
    if(val) {
      delete [] val;
      val = 0;
      _dim1 = 0;
    }
  }

  void
  deep_copy(const simple_init_array<T>& s){
    init(s._dim1);
    for(unsigned i(0);i<_dim1;++i) val[i] = s.val[i];
  }

public:
  T* val;
private:
  unsigned _dim1;
};


template <typename T>
struct simple_matrix {
  simple_matrix(const unsigned dim1,
                const unsigned dim2)
    : val(0), _dim1(dim1), _dim2(dim2) { init(); }

  simple_matrix(const simple_matrix<T>& s)
    : val(0), _dim1(s._dim1), _dim2(s._dim2) {
    init();
    for(unsigned i(0);i<_dim1*_dim2;++i) val[0][i] = s.val[0][i];
  }

  ~simple_matrix(){
    if(_dim1>0) delete [] val[0];
    delete [] val;
  }

private:
  simple_matrix& operator=(const simple_matrix<T>&);

  void
  init(){
    val = new T*[_dim1];
    if(_dim1>0) val[0] = new T[_dim1*_dim2];
    for(unsigned i(1);i<_dim1;++i) val[i] = val[0]+i*_dim2;
  }

public:
  T** val;
private:
  unsigned _dim1;
  unsigned _dim2;
};


template <typename T>
struct simple_init_matrix {

  simple_init_matrix() : val(0), _dim1(0),_dim2(0) {}

  simple_init_matrix(const simple_init_matrix<T>& s) : val(0), _dim1(0), _dim2(0) { deep_copy(s); }

  ~simple_init_matrix() { dealloc(); }

  simple_init_matrix& operator=(const simple_init_matrix<T>& rhs) {
    if(&rhs == this) return *this;
    deep_copy(rhs);
    return *this;
  }

  unsigned dim1() const { return _dim1; }
  unsigned dim2() const { return _dim2; }

  void
  init(const unsigned d1,
       const unsigned d2) {
    if(val && _dim1 == d1 && _dim2 == d2) return;
    dealloc();
    _dim1 = d1;
    _dim2 = d2;
    alloc();
  }

private:
  void
  alloc(){
    val    = new T*[_dim1];
    if(_dim1>0) val[0] = new T[_dim1*_dim2];
    for(unsigned i(1);i<_dim1;++i) val[i] = val[0]+i*_dim2;
  }

  void
  dealloc() {
    if(val) {
      if(_dim1>0) delete [] val[0];
      delete [] val;
      val = 0;
      _dim1 = 0;
      _dim2 = 0;
    }
  }

  void
  deep_copy(const simple_init_matrix<T>& s){
    init(s._dim1,s._dim2);
    for(unsigned i(0);i<_dim1*_dim2;++i) val[0][i] = s.val[0][i];
  }

public:
  T** val;
private:
  unsigned _dim1;
  unsigned _dim2;
};


template <typename T>
struct simple_matrix3d {
  simple_matrix3d(const unsigned dim1,
                  const unsigned dim2,
                  const unsigned dim3)
    : val(0), _dim1(dim1), _dim2(dim2), _dim3(dim3) { init(); }

  simple_matrix3d(const simple_matrix3d<T>& s)
    : val(0), _dim1(s._dim1), _dim2(s._dim2), _dim3(s._dim3) {
    init();
    for(unsigned i(0);i<_dim1*_dim2*_dim3;++i) val[0][0][i] = s.val[0][0][i];
  }

  ~simple_matrix3d(){
    if(_dim1){
      if(_dim2){
        delete [] val[0][0];
      }
      delete [] val[0];
    }
    delete [] val;
  }

private:
  simple_matrix3d& operator=(const simple_matrix3d<T>&);

  void
  init(){
    val = new T**[_dim1];
    if(_dim1){
      val[0] = new T*[_dim1*_dim2];
      if(_dim2){
        val[0][0] = new T[_dim1*_dim2*_dim3];
      }
    }
    for(unsigned i(0);i<_dim1;++i){
      val[i] = val[0]+i*_dim2;
      for(unsigned j(0);j<_dim2;++j) {
        val[i][j] = val[0][0] + (i*_dim2+j)*_dim3;
      }
    }
  }

public:
  T*** val;
private:
  unsigned _dim1;
  unsigned _dim2;
  unsigned _dim3;
};


template <typename T>
struct simple_init_matrix3d {
  simple_init_matrix3d() : val(0), _dim1(0), _dim2(0), _dim3(0) {}


  simple_init_matrix3d(const simple_init_matrix3d<T>& s)
    : val(0), _dim1(0), _dim2(0), _dim3(0) { deep_copy(s); }

  ~simple_init_matrix3d(){ dealloc(); }

  void
  init(const unsigned d1,
       const unsigned d2,
       const unsigned d3) {
    if(val && _dim1 == d1 && _dim2 == d2 && _dim3 == d3) return;
    dealloc();
    _dim1 = d1;
    _dim2 = d2;
    _dim3 = d3;
    alloc();
  }

  unsigned dim1() const { return _dim1; }
  unsigned dim2() const { return _dim2; }
  unsigned dim3() const { return _dim3; }

private:
  simple_init_matrix3d& operator=(const simple_init_matrix3d<T>&);

  void
  alloc(){
    val = new T**[_dim1];
    if(_dim1){
      val[0] = new T*[_dim1*_dim2];
      if(_dim2){
        val[0][0] = new T[_dim1*_dim2*_dim3];
      }
    }
    for(unsigned i(0);i<_dim1;++i){
      val[i] = val[0]+i*_dim2;
      for(unsigned j(0);j<_dim2;++j) {
        val[i][j] = val[0][0] + (i*_dim2+j)*_dim3;
      }
    }
  }

  void
  dealloc(){
    if(_dim1){
      if(_dim2){
        delete [] val[0][0];
      }
      delete [] val[0];
    }
    delete [] val;
    val = 0;
    _dim1 = 0;
    _dim2 = 0;
    _dim3 = 0;
  }

  void
  deep_copy(const simple_init_matrix3d<T>& s){
    init(s._dim1,s._dim2,s._dim3);
    for(unsigned i(0);i<_dim1*_dim2*_dim3;++i) val[0][0][i] = s.val[0][0][i];
  }

public:
  T*** val;
private:
  unsigned _dim1;
  unsigned _dim2;
  unsigned _dim3;
};

#endif
