// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: site_data.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "site_data.h"
#include "util/bio/bioseq_util.h"
#include "util/general/die.h"
#include "util/general/io_util.h"
#include "util/general/metatags.h"

#include <cassert>
#include <cstdlib>

#include <iostream>
#include <sstream>
#include <vector>
#include <utility>

using namespace std;


const unsigned DATA_FILE_VERSION(1);


void
site_data::
peek(std::ostream& os) const {
  static const unsigned peek_len(10);

  site_count_map::const_iterator i,i_end=site_count.end();

  const unsigned len(site_count.size());
  os << "seq_t:\n";
  os << "len: " << len << "\n";

  unsigned j(0);
  for(i=site_count.begin();i!=i_end;++i){
    if(j++>=peek_len) break;
    os << i->first << "\n";
  }
}


unsigned
site_data::
max_cat() const {
  count_data_code::cat_type mc(0);

  site_count_map::const_iterator i,i_end=site_count.end();

  for(i=site_count.begin();i!=i_end;++i){

    site_count_data_type::const_iterator j,j_end=i->second.end();

    for(j=i->second.begin();j!=j_end;++j){
      if(j->first.cat_no>mc) mc = j->first.cat_no;
    }
  }

  return static_cast<unsigned>(mc);
}




static
void
finish_cat_line(const std::vector<unsigned>& cat_list,
                const unsigned last_cat,
                std::ostream& os){
  for(unsigned c(0),cat_sum(0);c<last_cat;++c){
    // check for and skip printing of final "1" if it's the only count
    // in this line:
    if((c+1)==last_cat && cat_list[c]==1 && cat_sum == 0) continue;
    os << " " << cat_list[c];
    cat_sum += cat_list[c];
  }
  os << "\n";
}



static const char* org_signal("org:");
static const char* cat_signal("cat:");
static const char site_signal('S');
static const char pair_signal1('P');
static const char pair_signal2('Q');




void
site_data::
store_state(std::ostream& os,
            const char* const cmdline,
            long int seed) const {

  os << "site_data_file_version " << DATA_FILE_VERSION << "\n";
  os << "cmdline ";
  if(cmdline) os << cmdline;
  os << "\n";
  os << "random_seed " << seed << "\n";
  os << "site_model " << sm << "\n";

  const unsigned n_org(taxid.size());

  os << org_signal;
  for(unsigned t(0);t<n_org;++t){
    os << " " << taxid.getstr(t);
  }
  os << "\n";

  const unsigned mc(max_cat());

  os << cat_signal;
  {
    for(unsigned i(0);i<(mc+1);++i){
      os << " ";
      if( id_2_catlabel.find(i) != id_2_catlabel.end() ){
        os << id_2_catlabel.find(i)->second;
      } else {
        os << i;
      }
    }
  }
  os << "\n";

  std::vector<unsigned> cat_list;

  site_count_map::const_iterator i,i_end=site_count.end();

  for(i=site_count.begin();i!=i_end;++i){
    os << site_signal << i->first;

    bool is_new_g(true);
    unsigned last_cat(0);

    site_count_data_type::const_iterator j,j_end=i->second.end();

    for(j=i->second.begin();j!=j_end;++j) {
      const count_data_code::cat_type& cat(j->first.cat_no+1);
      const unsigned& count(j->second);

      if(is_new_g) {
        finish_cat_line(cat_list,last_cat,os);
        os << 0;
        last_cat=0;
        is_new_g=false;
      }

      if(cat>cat_list.size()) cat_list.resize(cat);

      for(;last_cat<cat;++last_cat) {
        if((last_cat+1)==cat) cat_list[last_cat] = count;
        else                  cat_list[last_cat] = 0;
      }
    }
    finish_cat_line(cat_list,last_cat,os);
  }


  // write out pair-info:
  pair_count_map::const_iterator pi,pi_end=pair_count.end();

  cat_list.clear();
  unsigned last_cat1(0);
  unsigned last_cat2(0);

  site_code last_p1;
  site_code last_p2;
  bool is_new_cat1(true);

  for(pi=pair_count.begin();pi!=pi_end;++pi){
    const site_pair_type sp(pi->first.first);

    if       (! (sp.first == last_p1)){
      last_p1 = sp.first;
      last_p2 = sp.second;
      if(! is_new_cat1) finish_cat_line(cat_list,last_cat2,os);
      is_new_cat1=true;
      os << pair_signal1 << sp.first << " " << sp.second <<"\n";
    } else if(! (sp.second == last_p2)){
      last_p2 = sp.second;
      if(! is_new_cat1) finish_cat_line(cat_list,last_cat2,os);
      is_new_cat1=true;
      os << pair_signal2 << sp.second << "\n";
    }

    const cat_pair_type cp(pi->first.second);

    const unsigned cat1(cp.first);
    const unsigned cat2(cp.second+1);
    const unsigned count(pi->second);

    if(cat1 != last_cat1) {
      if(! is_new_cat1) finish_cat_line(cat_list,last_cat2,os);
      is_new_cat1=true;
      last_cat1=cat1;
    }

    if(is_new_cat1){
      os << cat1;
      last_cat2=0;
      is_new_cat1=false;
    }

    if(cat2>cat_list.size()) cat_list.resize(cat2);

    for(;last_cat2<cat2;++last_cat2) {
      if((last_cat2+1)==cat2) cat_list[last_cat2] = count;
      else                    cat_list[last_cat2] = 0;
    }
  }
  if(! is_new_cat1) finish_cat_line(cat_list,last_cat2,os);
}



static
void
get_site_code(const unsigned nseq,
              site_code& sc,
              std::istream& is){

  int c;
  unsigned sci(0);
  site_code::index_type it;

  while( (c=is.get()) != '\n' && is && sci < nseq){
    if(c==' ' || c=='\t') continue;
    is.unget();
    is >> it;
    sc.set_taxid(sci,it);
    sci++;
  }
}

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream&,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid site data file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  //  std::cerr << is.rdbuf();
  abort();
}




static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}




void
site_data::
load_state(std::istream& is) {

  {  // check file version
    unsigned tmpu;
    load_advance(is,"site_data_file_version");
    is >> tmpu;
    if(tmpu != DATA_FILE_VERSION){
      std::cerr << "ERROR:: bad site data file version number. expected: " << DATA_FILE_VERSION << "\n";
      exit(EXIT_FAILURE);
    }
  }

  // throw away old cmd-line,random_seed
  load_advance(is,"cmdline");
  clear_line(is);
  load_advance(is,"random_seed");
  clear_line(is);

  {
    int tmpi;
    std::string dummy;
    is >> dummy >> tmpi;
    sm = static_cast<SITE_MODEL::index_t>(tmpi);
  }
  clear_line(is);

  taxid.clear();
  catlabel_2_id.clear();
  id_2_catlabel.clear();
  {
    std::string buf;

    while(is){
      if(is.peek()==site_signal) break;

      if( ! getline(is,buf) ) break;

      std::istringstream sbuf(buf);
      std::string tmpstr;

      sbuf >> tmpstr;
      if(tmpstr==org_signal){
        while(sbuf >> tmpstr){ taxid.assignid(tmpstr); }
      }
      if(tmpstr==cat_signal){
        unsigned cat_no(0);
        while(sbuf >> tmpstr){
          catlabel_2_id[tmpstr] = cat_no;
          id_2_catlabel[cat_no] = tmpstr;
          cat_no++;
        }
      }
    }
  }

  site_count.clear();

  const unsigned nseq(taxid.size());
  if(nseq==0) die("no sequence labels in site data file");

  site_code sc(nseq);
  site_code prev_sc(nseq);

  int c;

  unsigned tmp_uint;
  unsigned cat_no;
  bool is_zero_line;
  bool is_prev_sc_set(false);

  unsigned cat1;

  enum section_type {NONE,SINGLE,PAIR};

  section_type sec(NONE);


  site_count_data_type* site_map(0);

  while(is){
    c=is.get();
    if(c==' ' || c=='\t' || c== '\n' || ! is) continue;
    if(c==site_signal){
      sec=SINGLE;

      get_site_code(nseq,sc,is);

      site_map = &(site_count[sc]);

    } else if(c==pair_signal1 || c==pair_signal2){
      sec=PAIR;

      if(c==pair_signal1){
        get_site_code(nseq,prev_sc,is);
        is_prev_sc_set=true;
      } else {
        if(! is_prev_sc_set ) die("bad pair site_data parse");
      }

      get_site_code(nseq,sc,is);

    } else {
      if(sec==SINGLE){
        is.unget();
        is >> tmp_uint; // g
        cat_no=0;
        is_zero_line=true;
        while( (c=is.get()) != '\n' && is){
          if(c==' ' || c=='\t') continue;
          is.unget();
          is >> tmp_uint; // count

          if(tmp_uint != 0) {
            site_map->operator[](count_data_code(cat_no)) += tmp_uint;
            is_zero_line=false;
          }
          cat_no++;
        }
        if(is_zero_line) { site_map->operator[](count_data_code(cat_no)) += 1; }
      } else if(sec==PAIR){
        is.unget();
        is >> cat1;
        cat_no=0;
        is_zero_line=true;
        while( (c=is.get()) != '\n' && is){
          if(c==' ' || c=='\t') continue;
          is.unget();
          is >> tmp_uint; // count

          if(tmp_uint != 0) {
            pair_count.operator[](make_pair(make_pair(prev_sc,sc),make_pair(cat1,cat_no))) = tmp_uint;
            is_zero_line=false;
          }
          cat_no++;
        }
        if(is_zero_line) { pair_count.operator[](make_pair(make_pair(prev_sc,sc),make_pair(cat1,cat_no))) = 1; }
      } else {
        die("bad site_data parse");
      }
    }
  }
}
