// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "estep_dat_t.h"
#include "lhood_model.h"
#include "lhood_model_em.h"
#include "lhood_model_em_baz.h"
#include "lhood_model_em_pprob.h"
#include "subs_ml.h"
#include "subs_ml_minfunc.h"
#include "subs_ml_minimize.h"
#include "subs_ml_print_util.h"
#include "subs_ml_util.h"
#include "tree_probs.h"
#include "util/general/log.h"
#include "util/math/array_util.h"

#include <cstdlib>
#include <cmath>

#include <iomanip>
#include <ostream>


#if 0
///
static
void
sml_estep(const bi_tree_sml_type& tree,
          const tree_probs& tprob,
          const site_data_fastlup& sdf,
          estep_dat_t& edat,
          const unsigned N){

  update_etrans(tree,tprob,sdf,N,edat);

//   std::cerr << "eroot 1,2: " << edat.root[0] << " " << edat.root[1] << "\n";
//   std::cerr << "etrans[0-4]:\n";
//   matrix_dump(edat.trans[0],N,std::cerr);
//   matrix_dump(edat.trans[1],N,std::cerr);
//   matrix_dump(edat.trans[2],N,std::cerr);
//   matrix_dump(edat.trans[3],N,std::cerr);
}
#endif




#ifdef DEBUG
static
void
check_edat(const estep_dat_t& edat,
           const smlfloat expect){

  smlfloat sum(0.);
  sum = array_sum(edat.root,edat.n_states);
  if(std::fabs(sum-expect)>0.001){
    log_os << "bad edat root sum:" << sum << "\n";
    abort();
  }

  for(unsigned i(0);i<edat.n_branches;++i){
    sum = array_sum(edat.trans[i],edat.n_states*edat.n_states);
    log_os << "i,sum,exp: " << i << " " << sum << " " << expect << "\n";
    if(std::fabs(sum-expect)>0.001){
      log_os << "bad edat branch sum: " << sum << " branch: " << i << "\n";
      abort();
    }
  }
}
#endif



static
void
sml_em_min(const site_data_fastlup& sdf,
           subs_ml_model& mdl,
           const smlfloat start_lnp_norm){

  const unsigned n_states(mdl.state_size());
  const smlfloat norm(get_lnp_norm(mdl,sdf));
  const smlfloat ntol(mdl.opt().converge_value/norm);

  const unsigned n_branches(mdl.tree().n_branches());

  estep_dat_t edat(n_states,n_branches);
  unsigned iter(0);

  smlfloat iter_lnp_norm(start_lnp_norm);

  while(true){
    iter++;
    {
      tree_probs tprob;
      tprob.init_probs(mdl);

      // control execution of temporary baz code:
      static const bool is_baz_mode(false);

      if(is_baz_mode){
        update_etrans_baz(mdl,mdl.tree(),tprob,sdf,n_states);
        break;
      } else {
        update_etrans(mdl.tree(),tprob,sdf,n_states,edat);
      }

      smlfloat mstep_lnp,mstep_lnp_norm;
      get_lnprob_from_param(mdl,edat,mstep_lnp,mstep_lnp_norm);
      log_os << "GREPGREP regular mstep lnp: " << std::setprecision(SUBS_ML_PRINT_LNP_PRECISION) << mstep_lnp << " " << mstep_lnp_norm << "\n";

#ifdef DEBUG
      check_edat(edat,sdf.total_count);
#endif

#if 0
      mdl.rate_gtor().cf().adjust_edat(sdf);
#endif
      //lnp = get_lnprob_from_param(mdl,mdl.rate_gtor().cf().edat2);
      //std::cerr << "GREPGREP conditioned mstep lnp: " << std::setprecision(12) << lnp << "\n";
      //check_edat(mdl.rate_gtor().cf().edat2,n_states);
    }

    sml_mstep_min(edat,mdl);

    smlfloat new_lnp,new_lnp_norm;
    get_lnprob_from_param(mdl,sdf,new_lnp,new_lnp_norm);

    log_os << "EM ITER " << iter << " LNP: " << new_lnp << " " << new_lnp_norm << "\n";

    if( mdl.opt().max_steps ){
      if( (min_call_count
           +dmin_call_count) > mdl.opt().max_steps ) {
        break;
      }
    }

    if (2.0*std::fabs(iter_lnp_norm-new_lnp_norm) <= ntol*(std::fabs(iter_lnp_norm)+std::fabs(new_lnp_norm))) {
      break;
    }

    iter_lnp_norm = new_lnp_norm;
  }
}




void
score_subs_ml_model(const site_data_fastlup& sdf,
                    const subs_ml_model& mdl,
                    smlfloat& lnp,
                    smlfloat& lnp_norm,
                    const bool is_cat_post_prob_mode){

  get_lnprob_from_param(mdl,sdf,lnp,lnp_norm,is_cat_post_prob_mode);
}




void
train_subs_ml_model(const site_data_fastlup& sdf,
                    subs_ml_model& mdl,
                    bool is_refine){

  // get initial lnp
  smlfloat start_lnp,start_lnp_norm;
  get_lnprob_from_param(mdl,sdf,start_lnp,start_lnp_norm);

#ifdef DEBUG
  mdl.dump(log_os);
#endif

  const unsigned site_count(sdf.total_count);
  log_os << "Initial lnp: ";
  lnp_line_report(start_lnp,start_lnp_norm,site_count,log_os);

  // full max lhood minimization:
  //
  mdl.opt().is_extra_warnings=false; //make things quieter during iterative run
  if(mdl.opt().min == MIN::EM){
    sml_em_min(sdf,mdl,start_lnp_norm);
  } else {
    sml_fullstep_min(sdf,mdl,!(is_refine));
  }

  mdl.opt().is_extra_warnings=true;
  smlfloat final_lnp,final_lnp_norm;
  get_lnprob_from_param(mdl,sdf,final_lnp,final_lnp_norm);

  //  mdl.dump(log_os);
  log_os << "Final lnp  ";
  lnp_line_report(final_lnp,final_lnp_norm,site_count,log_os);
}
