// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml_minfunc.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "substk_exception.h"
#include "subs_ml_minfunc.h"
#include "subs_ml_print_util.h"
#include "util/general/log.h"
#include "util/math/test_float.h"

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <ostream>


#ifdef DEBUG
#define MINFUNC_DEBUG
#endif


unsigned min_call_count(0);
unsigned dmin_call_count(0);
bool write_ping(true);

double last_lnp(std::log(0.)); ///< used during profiling



// occasional tasks during minimization
static
void
xminfunc_count_check(const smlfloat lnp,
                     const smlfloat,
                     const subs_ml_model& mdl,
                     const bool is_mstep = false){

  // return neg lnp for minimization
  static const unsigned iter_report_period(50);
  const unsigned counter(min_call_count+dmin_call_count);
  if ( counter%iter_report_period == 0 ){
    log_os << "\n";
    if(is_mstep) log_os << "mstep ";
    log_os << "lnp: "
              << std::setprecision(SUBS_ML_PRINT_LNP_PRECISION)
              << -lnp << "\n";
  }

  // backup model state
  const unsigned ping_pong_rate(180000/mdl.state_size());
  const unsigned ping_pong_first_jump(5*ping_pong_rate/6);
  if ( mdl.opt().is_pingpong &&
       (counter+ping_pong_first_jump)%ping_pong_rate == 0 ){
    std::string nfile(mdl.opt().pingpongfile);
    if( write_ping ) nfile += ".ping";
    else             nfile += ".pong";

    std::ofstream fos(nfile.c_str());
    mdl.store_state(fos);
    write_ping = ! write_ping;
  }
}




smlfloat
subs_ml_minfunc_base::
dval(const smlfloat* p,smlfloat* df){

  const unsigned ndim(dim());
  const bool xfv(is_inner_step());

  // used during profiling
  if( mdl.opt().max_steps ){
    if( (min_call_count+dmin_call_count) > mdl.opt().max_steps ) {
      for(unsigned i(0);i<ndim;++i) df[i] = 0.;
      return -last_lnp;
    }
  }

  // load minimizer parameters:
  min_p_to_param(p);

  // get current lnp
  smlfloat lnp,lnp_norm;
  get_lnprob(mdl,lnp,lnp_norm);

  if( mdl.opt().max_steps ) last_lnp = lnp_norm;

  min_call_count++;
  xminfunc_count_check(lnp,lnp_norm,mdl,xfv);

  // estimate deriv at low resolution
  for(unsigned i(0);i<ndim;++i){
    df[i] = get_deriv_estimate(p,lnp_norm,i);
    min_call_count++;
    xminfunc_count_check(lnp,lnp_norm,mdl,xfv);
  }

  { // get norm of deriv:
    static const smlfloat norm_thresh(1.e-5);
    const smlfloat norm(std::sqrt(array_dot(df,df,ndim)));
    if(norm<norm_thresh){
      // estimate centered deriv
      for(unsigned i(0);i<ndim;++i){
        df[i] = (df[i] + get_deriv_estimate(p,lnp_norm,i,false))/2.;
        min_call_count++;
        xminfunc_count_check(lnp,lnp_norm,mdl,xfv);
      }
    }
  }

  for(unsigned i(0);i<ndim;++i){ df[i] *= -1.; }

  dmin_call_count++;
  xminfunc_count_check(lnp,lnp_norm,mdl,xfv);

#ifdef MINFUNC_DEBUG
  for(unsigned i(0);i<ndim;++i) { log_os << "i,p[i]: " << i << " " << p[i] << "\n"; }
  for(unsigned i(0);i<ndim;++i) { log_os << "i,df[i]: " << i << " " << df[i] << "\n"; }
  log_os << "return lnp/lnp_norm " << -lnp << " " << -lnp_norm << "\n";
#endif

  return -lnp_norm;
}




smlfloat
subs_ml_minfunc_base::
val(const smlfloat* p) {

  if( mdl.opt().max_steps ){
    if( (min_call_count+dmin_call_count) > mdl.opt().max_steps ) {
      return -last_lnp;
    }
  }

#ifdef MINFUNC_DEBUG
  {
    const unsigned n(dim());
    for(unsigned i(0);i<n;++i){
      if(is_float_invalid(p[i])){
        log_os << "minfunc.val(): invalid minimizer input i,p: " << i << " " << p[i] << "\n";
        for(unsigned j(0);j<n;++j){
          log_os << "j,p: " << j << " " << p[j] << "\n";
        }
        throw substk_exception("bad minimizer eval");
      }
    }
  }
#endif

  min_p_to_param(p);

  // recalculate probability based on model parameters
  smlfloat lnp,lnp_norm;
  get_lnprob(mdl,lnp,lnp_norm);

#ifdef MINFUNC_DEBUG
  {
    log_os << "minfunc:\n";
    const unsigned n(dim());
    for(unsigned i(0);i<n;++i) { log_os << "i,p[i]: " << i << " " << p[i] << "\n"; }
    log_os << "lnp/lnp_norm: " << -lnp << " " << -lnp_norm << "\n";
  }
#endif

  if( mdl.opt().max_steps ) last_lnp = lnp_norm;
  min_call_count++;
  xminfunc_count_check(lnp,lnp_norm,mdl,is_inner_step());

  return -lnp_norm;
}




smlfloat
subs_ml_minfunc_base::
get_deriv_estimate(const smlfloat* p,
                   const smlfloat lnp1_norm,
                   const unsigned param_no,
                   const bool is_plus) const {

  static const smlfloat delta(1e-5);
  static const smlfloat deltamin(1e-7);

  const unsigned n(dim());
  smlfloat* pcopy(new smlfloat[n]);
  std::copy(p,p+n,pcopy);

  const smlfloat delchunk(std::max(std::fabs(pcopy[param_no]*delta),deltamin));
  const smlfloat start_param(pcopy[param_no]);

  if(is_plus) pcopy[param_no] = start_param+delchunk;
  else        pcopy[param_no] = start_param-delchunk;

  //  log_os << "i,delchunk,pstart,pdel: " << param_no << " " << delchunk << " " << start_param << " " << pcopy[param_no] << "\n";

  subs_ml_model mdl_copy(mdl);
  min_p_to_param_x(pcopy,mdl_copy);
  delete [] pcopy;

  //  mdl_copy.param_state(pcopy,pv);

  //  log_os << "pdel_actual: " << pcopy[param_no] << "\n";

  smlfloat lnp2,lnp2_norm;
  get_lnprob(mdl_copy,lnp2,lnp2_norm);

  //  log_os << "lnp1_norm,lnp2_norm: " << lnp1_norm << " " << lnp2_norm << "\n";
  //  log_os << "returning: " << (lnp2_norm-lnp1_norm)/delchunk << "\n";

  if(is_plus) return (lnp2_norm-lnp1_norm)/delchunk;
  else        return (lnp1_norm-lnp2_norm)/delchunk;
}
