// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml_minimize.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "subs_ml_minfunc.h"
#include "subs_ml_minimize.h"
#include "util/math/minimize_conj_direction.h"
#include "util/math/minimize_conj_gradient.h"
#ifdef USE_PRAXIS
#include "util/math/minimize_praxis.h"
#endif

#include "util/general/log.h"
#include <ostream>


static
void
run_conj_dir_minimizer(subs_ml_minfunc_base& mf,
                       const smlfloat start_ntol,
                       const smlfloat ntol){

  static const smlfloat line_tol(1e-7);
  static const smlfloat start_ratio(0.05);
  static const smlfloat min_start_dist(1e-6);

  const unsigned ndim(mf.dim());
  const unsigned reset_iter(std::max(ndim/4,static_cast<unsigned>(10)));

  smlfloat* p(new smlfloat[ndim]);
  smlfloat* conj_dir(new smlfloat[ndim*ndim]);

  smlfloat iter_ntol(start_ntol);
  unsigned total_iter(0);

  while(true){
    unsigned iter(0);
    smlfloat fmin;

    mf.param_to_min_p(p);
    std::fill(conj_dir,conj_dir+ndim*ndim,0.);
    for(unsigned i(0);i<ndim;++i) {
      const smlfloat start_dist( std::max(std::fabs(p[i]*start_ratio),min_start_dist) );
      conj_dir[i*(ndim+1)] = start_dist;
    }

    minimize_conj_direction(p,conj_dir,mf,iter_ntol,ntol,line_tol,fmin,iter,reset_iter);
    mf.min_p_to_param(p);

    total_iter += std::min(iter+1,reset_iter);
    log_os << "CONJ DIR MINIMIZER ITERATIONS: " << std::min(iter+1,reset_iter) << "\n";

    if(reset_iter==0 || iter < reset_iter) break;
  }

  log_os << "CONJ DIR MINIMIZER TOTAL ITERATIONS: " << total_iter << "\n";

  delete [] p;
  delete [] conj_dir;
}



static
void
run_conj_grad_minimizer(subs_ml_minfunc_base& mf,
                        smlfloat ntol,
                        unsigned reset_iter=0,
                        const unsigned max_iter=1000){

  if(reset_iter==0) reset_iter=max_iter;

  const unsigned ndim(mf.dim());
  smlfloat* p(new smlfloat[ndim]);

  unsigned total_iter(0);

  while(true){
    unsigned iter;
    smlfloat fmin;

    if(total_iter+reset_iter>max_iter) reset_iter=max_iter-total_iter;

    mf.param_to_min_p(p);
    minimize_conj_gradient(p,mf,ntol,fmin,iter,reset_iter);
    mf.min_p_to_param(p);

    total_iter += std::min(iter+1,reset_iter);

    log_os << "CONJ GRAD MINIMIZER CYCLE/TOTAL ITERATIONS: " << std::min(iter+1,reset_iter) << " " << total_iter << "\n";

    if(iter < reset_iter) break;
    if(total_iter >= max_iter) {
      log_os << "MAX CONJ GRAD TOTAL ITERATIONS: " << max_iter << "\n";
      break;
    }
  }

  delete [] p;
}





static
void
sml_fullstep_min_tol(const site_data_fastlup& sdf,
                     subs_ml_model& mdl,
                     const smlfloat start_ntol,
                     const smlfloat end_ntol){

  // search for ml estimate of remaining parameters
  //
  const smlfloat norm(get_lnp_norm(mdl,sdf));
  const smlfloat end_ntol_norm(end_ntol/norm);
  const smlfloat start_ntol_norm(start_ntol/norm);

  min_call_count = 0;
  dmin_call_count = 0;

  subs_ml_minfunc mff(sdf,mdl,PARAM_VIEW::MIN_PARAM);
  subs_ml_minfunc mfg(sdf,mdl,PARAM_VIEW::INDY_MIN_PARAM);

  if (mdl.opt().min == MIN::CG_PRAXIS_HYBRID){
    // start out with a short conj gradient min
    static const unsigned max_iter(DEFAULT_CONJ_GRAD_START_ITER);
    run_conj_grad_minimizer(mfg,end_ntol_norm,max_iter,max_iter);

    const unsigned ndim(mff.dim());
    smlfloat* p(new smlfloat[ndim]);

    mff.param_to_min_p(p);
#ifdef USE_PRAXIS
    minimize_praxis(p,mff,end_ntol_norm,1.0,0.05);
#else
    pass_away("praxis minimization unavailable in this build");
#endif
    mff.min_p_to_param(p);

    delete [] p;

  } else if     (mdl.opt().min == MIN::CONJ_DIR){

    run_conj_dir_minimizer(mff,start_ntol_norm,end_ntol_norm);

  } else if (mdl.opt().min == MIN::CONJ_GRAD){

    static const unsigned reset_iter(60);
    run_conj_grad_minimizer(mfg,end_ntol_norm,reset_iter);

  } else if(mdl.opt().min == MIN::CG_CD_HYBRID){

    // start out with a short conj gradient min
    static const unsigned max_iter(DEFAULT_CONJ_GRAD_START_ITER);
    run_conj_grad_minimizer(mfg,end_ntol_norm,max_iter,max_iter);

    // then do regular conj dir min
    run_conj_dir_minimizer(mff,start_ntol_norm,end_ntol_norm);
  }

  log_os << "TOTAL MIN  CALLS: " << min_call_count <<"\n";
  log_os << "TOTAL DMIN CALLS: " << dmin_call_count <<"\n";
}


void
sml_fullstep_min(const site_data_fastlup& sdf,
                 subs_ml_model& mdl,
                 const bool is_start_tol){

  // search for ml estimate of remaining parameters
  //
  const smlfloat end_ntol(mdl.opt().converge_value);
  const smlfloat start_ntol(is_start_tol ? std::max(10.,end_ntol) : end_ntol);

  sml_fullstep_min_tol(sdf,mdl,start_ntol,end_ntol);
}




void
sml_mstep_min(const estep_dat_t& edat,
              subs_ml_model& mdl){

  const smlfloat norm(get_lnp_norm(mdl,edat));
  smlfloat ntol(mdl.opt().converge_value/norm);

  ntol *= 1e2; // b/c mstep is an "inner" iteration

  min_call_count = 0;
  dmin_call_count = 0;

  subs_ml_minfunc_em mfe(edat,mdl,PARAM_VIEW::MIN_PARAM);

  run_conj_dir_minimizer(mfe,ntol,ntol);

  log_os << "TOTAL MSTEP MIN CALLS: " << min_call_count <<"\n";
  log_os << "TOTAL MSTEP DMIN CALLS: " << dmin_call_count <<"\n";
}
