// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: tree_probs.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "subs_ml_model.h"
#include "subs_ml_ptol.h"
#ifdef BRANCH_SPECIFIC_OBS
#include "rate_edge_dependencies.h"
#include "subs_ml_model_root_from_obs.h"
#endif
#include "subs_prob_util.h"
#include "subs_prob_util_core.h"
#include "tree_probs.h"
#include "util/math/matrix_util.h"
#include "util/math/prob_util_io.h"

#ifdef PPIT_DEBUG
#include "util/math/matrix_util_io.h"
#endif

void
tree_probs::
alloc(const unsigned init_n_states,
      const unsigned init_n_branches){

  n_states=init_n_states;
  n_branches=init_n_branches;

  _root_prob.init(n_states);
  _branch_prob.init(n_branches,n_states*n_states);

  is_zero_time_branch.resize(n_branches);
  std::fill(is_zero_time_branch.begin(),is_zero_time_branch.end(),false);
}


void
tree_probs::
init_probs(const subs_ml_model& mdl,
           const unsigned site_category,
           const unsigned group_category,
           const bool is_use_submodel,
           const unsigned submodel_no,
           workspace ws[3]){

  _is_site_prob_mode=false;

  const unsigned& init_n_states(is_use_submodel ? mdl.submodel_state_size(submodel_no):mdl.state_size());
  const unsigned& init_n_branches(mdl.get_time_gtor().branch_size());

  if(n_states != init_n_states ||
     n_branches != init_n_branches){
    alloc(init_n_states,init_n_branches);
  }

  const prob_t* full_root(mdl.get_root_gtor().site_cat_distro(site_category));
  if(is_use_submodel){
    mdl.get_rate_gtor().submodel_pdistro_reduction(full_root,submodel_no,root_prob());
  } else {
    std::copy(full_root,full_root+n_states,root_prob());
  }

#ifndef NDEBUG
  pdistro_check(root_prob(),n_states,SUBS_ML_PTOL);
#endif

  rates_func_options ropt(site_category,group_category,false,false,is_use_submodel,submodel_no);

#ifdef BRANCH_SPECIFIC_OBS
  const prob_t * const * node_state_prob(mdl.get_root_gtor().site_cat_node_state_prob(site_category));

  site_model_edge_pdistros smp;
  if(node_state_prob){
    smp.ropt_hookup(ropt);
  }

  const SITE_MODEL::index_t sm(mdl.get_rate_gtor().site_model());
  const time_gtor& tgm(mdl.get_time_gtor());

  if(node_state_prob){
    const prob_t * const * zzz(branch_state_prob);
    simple_array<prob_t> state_prob_tmp(n_states);
    simple_array<prob_t> tprob_tmp(n_states*n_states);
    simple_array<prob_t> tprob_tmp2(n_states*n_states);

    for(unsigned b(0);b<n_branches;++b){
      static const unsigned n_sections(3);
      const smlfloat section_frac(1./static_cast<smlfloat>(n_sections));
      const smlfloat section_time(tgm.branch_time(b)*section_frac);

      for(unsigned j(0);j<n_sections;++j){
        const unsigned node_id(b+1);
        const unsigned parent_node_id(mdl.tree().get_parent_index(node_id));
        const smlfloat node_frac((section_frac*0.5)*(1+j*2));

        for(unsigned i(0);i<n_states;++i){
          state_prob_tmp.val[i] = zzz[node_id][i]*(node_frac)+
            zzz[parent_node_id][i]*(1.-node_frac);
        }
        set_site_model_edge_pdistros(sm,state_prob_tmp.val,smp);
        get_single_branch_subs_prob(tprob_tmp.val,section_time,mdl,ropt,ws);

        if(j==0){
          matrix_copy(tprob_tmp.val,branch_prob(b),n_states);
        } else {
          matrix_mult(tprob_tmp.val,tprob_tmp2.val,branch_prob(b),n_states);
        }
        if((j+1)!=n_sections){
          matrix_copy(branch_prob(b),tprob_tmp2.val,n_states);
        }
      }
    }

  } else {
    for(unsigned b(0);b<n_branches;++b){
      get_single_branch_subs_prob(branch_prob(b),tgm.branch_time(b),mdl,ropt,ws);
    }
  }

#else

  get_all_branch_subs_prob(_branch_prob.val,is_zero_time_branch,mdl,ropt,ws);

#endif

  for(unsigned b(0);b<n_branches;++b){
    matrix_transpose_inplace(branch_prob(b),n_states);
  }
}



void
tree_probs::
init_site_probs(const subs_ml_model& mdl,
                const unsigned site_category,
                const unsigned group_category,
                const bool is_use_submodels,
                const unsigned submodel_no,
                const pprob_cache& left_average,
                const pprob_cache& right_average,
                const unsigned n_sites,
                workspace ws[3]){

  /// \tmp for now.. sets up root_prob, n_states, and n_branches if init_probs wasn't called:
  init_probs(mdl,site_category,group_category,is_use_submodels,submodel_no,ws);

  _is_site_prob_mode=true;
  _n_sites=n_sites;

  // 0. cluster left-average right-average into N or fewer sets
  //
  // start with no clustering for now, and test on very small genes:
  //
  const unsigned n_clusters(_n_sites*n_branches);

  _site_branch_cluster_map.init(_n_sites,n_branches);
  _site_branch_timecluster_map.init(_n_sites,n_branches);
  _cluster_timecluster_start_map.resize(n_clusters);
  _cluster_branch_member.resize(n_clusters*n_branches,false);
  for(unsigned i(0),k(0);i<_n_sites;++i) {
    for(unsigned j(0);j<n_branches;++j) {
      _site_branch_cluster_map.val[i][j] = k;
      _site_branch_timecluster_map.val[i][j] = k;
      _cluster_timecluster_start_map[k] = k;
      _cluster_branch_member[k*n_branches+j] = true;
      k++;
    }
  }

  // note that nclusters here is the number of accepted nuc contexts,
  // ntimecluster includes the number of different times for each context
  //
  _exemplar_edge_distro.init(n_clusters,NUC::SIZE*2);
  for(unsigned i(0);i<_n_sites;++i){
    for(unsigned j(0);j<n_branches;++j){
      const unsigned cln(_site_branch_cluster_map.val[i][j]);
      for(unsigned n(0);n<NUC::SIZE;++n){
        _exemplar_edge_distro.val[cln][n]           = left_average.val[i][j][n];
        _exemplar_edge_distro.val[cln][NUC::SIZE+n] = right_average.val[i][j][n];
      }
    }
  }

  // note that nclusters here is the number of accepted nuc contexts,
  // ntimecluster includes the number of different times for each context
  // ...the conversion between these two sets is specified by _cluster_branch_member
  // created as a side effect of the nuc context clustering step
  //
  // ...for the toy clustering, these are the same:
  const unsigned n_timeclusters(n_clusters);

  // the brutal step
  const unsigned n2(n_states*n_states);
  _timecluster_prob.init(n_timeclusters,n2);

  // tmp vector to be loaded and passed into get_subs_prob in each
  // iteration of n_clusters loop:
  //
  std::vector<smlfloat> cluster_time_seq(n_branches,0);

  for(unsigned i(0);i<n_clusters;++i){
    unsigned cluster_bcount(0);

    // setup branch arrays for each cluster call to get subs_prob:
    //
    for(unsigned b(0);b<n_branches;++b){
      if(_cluster_branch_member[i*n_branches+b]){
        cluster_time_seq[cluster_bcount] = mdl.get_time_gtor().branch_time(b);
        is_zero_time_branch[cluster_bcount] = (cluster_time_seq[cluster_bcount]<=ZERO_BRANCH_TIME_THRESHOLD);
        cluster_bcount++;
      }
    }

    const unsigned startval(_cluster_timecluster_start_map[i]);
    rates_func_options ropt(site_category,group_category,false,false,is_use_submodels,submodel_no);
    ropt.nuc_pdistro_overedge_5p=_exemplar_edge_distro.val[i];
    ropt.nuc_pdistro_overedge_3p=_exemplar_edge_distro.val[i]+NUC::SIZE;

    get_subs_prob(_timecluster_prob.val+startval,mdl,ropt,
                  cluster_bcount,cluster_time_seq.begin(),is_zero_time_branch,ws);


#ifdef PPIT_DEBUG
    for(unsigned b(0);b<n_states;++b){
      std::cerr << "GREP_TPROB clust i left right ";
      std::cerr << i << " " << b
                << " " << _exemplar_edge_distro.val[i][b]
                << " " << _exemplar_edge_distro.val[i][b+NUC::SIZE] << "\n";
    }

    std::cerr << "GREP_TPROB cluster matrix (pre-transpose):\n";
    matrix_dump(_timecluster_prob.val[startval],n_states,std::cerr);
#endif

    for(unsigned b(0);b<cluster_bcount;++b){
      matrix_transpose_inplace(_timecluster_prob.val[startval+b],n_states);
    }

#ifdef PPIT_SITE_PROB_DEBUG
    std::cerr << "GREP_CLUSTER" << i << " " << _exemplar_edge_distro.val[i][0] << "\n";

    std::cerr << "GREP_R cluster,bcount,time1,start,prob[0][1] "
              << i << " " << cluster_bcount << " "
              << cluster_time_seq[0] << " "
              <<  _cluster_timecluster_start_map[i] << " "
              << _timecluster_prob.val[_cluster_timecluster_start_map[i]][1] << "\n";
#endif
  }
}



void
tree_probs::
self_check() const {

  pdistro_check(root_prob(),n_states,SUBS_ML_PTOL);
  for(unsigned b(0);b<n_branches;++b){
    for(unsigned i(0);i<n_states;++i){
      pdistro_check(branch_prob(b)+i,n_states,SUBS_ML_PTOL,n_states);
    }
  }
}
