// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: minimize_conj_gradient.hh 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "array_util.h"
#include "minimize_line.h"

#ifdef CG_DEBUG
#include "../general/log.h"
#include <ostream>
#endif


template <typename FloatType,
          typename MinFunc>
void
minimize_conj_gradient(FloatType* vec_current,
                       MinFunc& mf,
                       const FloatType tol,
                       FloatType& f_min,
                       unsigned& iter,
                       const unsigned max_iter){

  const unsigned n(mf.dim());

  FloatType* workspace(new FloatType[4*n]);

  FloatType* vec_deriv(workspace);
  FloatType* vec_grad(workspace+n);
  FloatType* vec_conj_grad(workspace+2*n);
  FloatType* linespace(workspace+3*n);

  FloatType gdot(0);

  for(iter=0;iter<max_iter;++iter){

    f_min = mf.dval(vec_current,vec_deriv);

    FloatType gnum(0);
    if(iter){
      for(unsigned i(0);i<n;++i){
        gnum += (vec_deriv[i]+vec_grad[i])*vec_deriv[i];
      }
      gnum /= gdot;
    }

    for(unsigned i(0);i<n;++i){
      vec_grad[i] = -vec_deriv[i];
      vec_conj_grad[i] = vec_grad[i]+gnum*vec_conj_grad[i];
    }

    const FloatType f_min_prev(f_min);

#ifdef CG_DEBUG
    try {
      minimize_line(vec_current,vec_conj_grad,f_min,mf,f_min,linespace);
    } catch (...) {
      std::ostream& log_os(log_stream());
      log_os << "minimize_conj_grad: gnum,gdot " << gnum << " " << gdot << "\n";
      for(unsigned i(0);i<n;++i){
        log_os << "i, x, -grad(x), conj_grad: " << i << "\t"
               << vec_current[i] << "\t" << vec_grad[i] << "\t" << vec_conj_grad[i] << "\n";
      }
      throw;
    }
#else
    minimize_line(vec_current,vec_conj_grad,f_min,mf,f_min,linespace);
#endif

    if(std::fabs(f_min-f_min_prev) <= tol) break;  // converging on absolute tolerance

    gdot = array_dot(vec_grad,vec_grad,n);
    if(gdot == 0.) break;
  }

  delete [] workspace;
}
